#include "disassemblerview.h"
#include <QProcess>

DisassemblerView::DisassemblerView(Type type, int cogId, MemoryMap *newMap):
    QListWidget(nullptr),
    m_newMap(newMap),
    m_type(type),
    m_cogId(cogId)
{
    setFont(QFont("Monospace",6));
    switch(type) {
        case ScratchPad:
            setWindowTitle("Scratchpad Asm");
            addEntry(AsmFunctionEntry("*",AsmFunctionEntry::FunctionStatic,0,MemoryMap::ScratchPadSize));
            break;
        case DynamicCog:
            setWindowTitle(QString("Dynamic Cog %1 Asm").arg(cogId));
            addEntry(AsmFunctionEntry("dynamicCogPeripheralsClk",AsmFunctionEntry::FunctionStatic,0,256));
            addEntry(AsmFunctionEntry("dynamicCogPreEvalPll",AsmFunctionEntry::FunctionStatic,256,256));
            addEntry(AsmFunctionEntry("dynamicCogAsyncVid",AsmFunctionEntry::FunctionStatic,512,512));
            addEntry(AsmFunctionEntry("dynamicCogPreEvalClk",AsmFunctionEntry::FunctionStatic,1024,256));
            addEntry(AsmFunctionEntry("dynamicCogCheckVid",AsmFunctionEntry::FunctionStatic,1280,256));
            addEntry(AsmFunctionEntry("dynamicCogPinOut",AsmFunctionEntry::FunctionStatic,1536,256));
            break;
    }

    doUpdate(true);
}

void DisassemblerView::addEntry(AsmFunctionEntry f) {
    Entry e;
    e.f = f;
    e.prevMem = QByteArray(f.size,char(0));
    m_entries << e;
}

void DisassemblerView::updateFunctionEntries(QList<AsmFunctionEntry> functions) {
    /*if (m_type == ScratchPad)
        return;
    m_entries.clear();
    clear();
    for (auto f:functions) {
        if (f.type == AsmFunctionEntry::FunctionDynamicCog && m_type != DynamicCog)
            continue;
        addEntry(f);
    }
    doUpdate(true);*/
}

void DisassemblerView::checkIfMemoryChanged() {
    doUpdate(false);
}

void DisassemblerView::doUpdate(bool force) {
    bool hasChanged = force;
    for (auto &e:m_entries) {
        quint8* startPtr=getBasePointerOfNewMap()+e.f.relOffset;
        if (!force && memcmp(startPtr,e.prevMem.constData(),quint32(e.f.size)) == 0)
            continue;
        hasChanged = true;
        memcpy(e.prevMem.data(), startPtr, quint32(e.f.size));
        e.prevResult = disassemble(e.prevMem, e.f);
    }
    if (hasChanged) {
        clear();
        for (auto &e:m_entries)
            addItems(e.prevResult);
    }
}

quint8* DisassemblerView::getBasePointerOfNewMap() const {
    switch(m_type) {
        case ScratchPad: return m_newMap->scratchPad();
        case DynamicCog: return m_newMap->cogDynamicAsm(m_cogId);
    }
    return nullptr;
}

QStringList DisassemblerView::disassemble(QByteArray data, AsmFunctionEntry f) {
    QProcess proc;
    proc.setProgram("ndisasm");
    proc.setArguments(QStringList()<<"-b"<<"64"<<"-");
    proc.setInputChannelMode(QProcess::ManagedInputChannel);
    proc.setProcessChannelMode(QProcess::ForwardedErrorChannel);
    proc.setReadChannel(QProcess::StandardOutput);
    proc.start();
    if (!proc.waitForStarted()) {
        qWarning()<<"unable to start ndisasm";
        return QStringList();
    }
    proc.write(data);
    proc.closeWriteChannel();
    if (!proc.waitForFinished())  {
        qWarning()<<"ndisasm did not finish";
        return QStringList();
    }
    QStringList rawLines = QString::fromUtf8(proc.readAllStandardOutput()).split('\n');
    while (!rawLines.isEmpty() && rawLines.last().trimmed().isEmpty())
        rawLines.removeLast();
    int nopLines=0;
    while (!rawLines.isEmpty() && rawLines.last().trimmed().endsWith(" nop")) {
        rawLines.removeLast();
        nopLines++;
    }

    QStringList result;
    result << f.name+":";
    for (auto l:rawLines)
        result << "    "+l;
    if (nopLines>0)
        result << "    "+QString::number(nopLines)+"x nop";
    return result;
}
