#include "mainwindow.h"
#include "ui_mainwindow.h"
#include <QMessageBox>
#include <QFile>
#include <QTimer>
#include <QtDebug>
#include <QRegExp>
#include <QSplitter>
#include <QHBoxLayout>
#include <QInputDialog>

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    m_timer(new QTimer(this)),
    ui(new Ui::MainWindow),
    m_mainSplitter(nullptr),
    m_sharedMemoryMap(nullptr),
    m_copyPrevMemoryMap(nullptr),
    m_intermediateMemory(nullptr),
    m_inMemoryUpdate(false)
{
    ui->setupUi(this);
    try {
        m_sharedMemoryMap = new MemoryMap("asm-prop-mem-1", false);
        m_copyPrevMemoryMap = new MemoryMap("",true);
        m_intermediateMemory = new quint8[MemoryMap::TotalMemoryMapSize];
    }
    catch(QString&e) {
        QMessageBox::warning(this,"Error",e);
        QTimer::singleShot(0,QCoreApplication::instance(),SLOT(quit()));
        return;
    }

    m_mainSplitter = new QSplitter(centralWidget());
    auto cl = new QHBoxLayout(ui->centralWidget);
    cl->addWidget(m_mainSplitter);

    connect(m_timer,SIGNAL(timeout()),this,SLOT(verifyMemoryChange()));
    m_timer->start(1000);
}

MainWindow::~MainWindow()
{
    delete ui;
    delete m_sharedMemoryMap;
    delete m_copyPrevMemoryMap;
    delete[] m_intermediateMemory;
}

void MainWindow::on_action_Add_Hub_Regs_triggered() {
    addView("hubregs");
}

void MainWindow::on_action_Add_Cog_Regs_triggered() {
    bool ok = false;
    int no = QInputDialog::getInt(this,"CogId","CogId",0,0,7,1,&ok);
    if (!ok)
        return;
    addView("cogregs:"+QString::number(no));
}

void MainWindow::on_action_Add_Scratchpad_Asm_triggered() {
    addView("scratchpad");
}

void MainWindow::on_action_Add_Dynamic_Cog_Asm_triggered() {
    bool ok = false;
    int no = QInputDialog::getInt(this,"CogId","CogId",0,0,7,1,&ok);
    if (!ok)
        return;
    addView("dyncogasm:"+QString::number(no));
}

void MainWindow::on_action_Add_Memory_Map_triggered() {
    addView("memorymap");
}

void MainWindow::addView(QString id) {
    qDebug()<<m_mainSplitter->count()<<"Add"<<id;
    ViewWidgetContainer *container = nullptr;
    if (id=="hubregs")
        container = new ViewWidgetContainer(new RegisterViewWidget(-1,m_sharedMemoryMap,m_copyPrevMemoryMap),m_mainSplitter,id,this);
    else if (id.startsWith("cogregs:")) {
        const int cogId = id.section(":",1,1).toInt();
        if (cogId>=0 && cogId<=7)
            container = new ViewWidgetContainer(new RegisterViewWidget(cogId,m_sharedMemoryMap,m_copyPrevMemoryMap),m_mainSplitter,id,this);
    }
    else if (id == "memorymap")
        container = new ViewWidgetContainer(new MemoryMapViewWidget(m_sharedMemoryMap),m_mainSplitter,id,this);
    else if (id == "scratchpad")
        container = new ViewWidgetContainer(new DisassemblerView(DisassemblerView::ScratchPad,-1,m_sharedMemoryMap),m_mainSplitter,id,this);
    else if (id.startsWith("dyncogasm:")) {
        const int cogId = id.section(":",1,1).toInt();
        if (cogId>=0 && cogId<=7)
            container = new ViewWidgetContainer(new DisassemblerView(DisassemblerView::DynamicCog,cogId,m_sharedMemoryMap),m_mainSplitter,id,this);
    }
    if (!container)
        return;
    m_viewWidgets << container;
    m_mainSplitter->addWidget(container);
    parseRegisterDefinition();
}

void MainWindow::parseRegisterDefinition() {
    qDebug()<<"Parse Register Definitions";
    QFile f("../propellersim/propeller-asm/asm-regs.gen.asm");
    if (!f.open(QIODevice::ReadOnly)) {
        qWarning()<<"unable to read"<<f.fileName();
        return;
    }
    QRegExp defRe("^\\s*(hub|cog)_([a-zA-Z_]+)\\s+equ\\s+([0-9]+)\\s*;(.+)$");
    QList<RegisterInfo> hubRegs;
    QList<RegisterInfo> cogRegs;
    while (!f.atEnd()) {
        const QString l = QString::fromUtf8(f.readLine()).trimmed();
        if (!defRe.exactMatch(l))
            continue;
        RegisterInfo ri;
        QString regSet = defRe.cap(1);
        ri.name = defRe.cap(2);
        ri.relativeOffset = defRe.cap(3).toInt();
        const QString regTypeStr = defRe.cap(4);
        if (regTypeStr == "bool")
            ri.type = RegisterInfo::Boolean;
        else if (regTypeStr == "quint32")
            ri.type = RegisterInfo::UInt32;
        else if (regTypeStr == "quint64")
            ri.type = RegisterInfo::UInt64;
        else if (regTypeStr == "void*")
            ri.type = RegisterInfo::Pointer;
        else if (!regTypeStr.startsWith("sizeof")) {
            qWarning()<<"invalid register type"<<regTypeStr<<ri.name;
            continue;
        }
        if (regSet == "hub")
            hubRegs << ri;
        else if (regSet == "cog")
            cogRegs << ri;
        else
            qWarning()<<"unknown reg set"<<regSet;
    }

    for (auto w:m_viewWidgets) {
        auto rw = dynamic_cast<RegisterViewWidget*>(w->viewWidget());
        if (rw)
            rw->updateRegisterDefinitions(rw->isHubRegs() ? hubRegs : cogRegs);
    }
}

void MainWindow::verifyMemoryChange() {
    if (m_inMemoryUpdate)
        return;
    m_inMemoryUpdate = true;
    m_sharedMemoryMap->redoAlignmentSearch();
    memcpy(m_intermediateMemory,m_sharedMemoryMap->fullMemoryMap(),MemoryMap::TotalMemoryMapSize);
    emit verifyMemorySignal();
    memcpy(m_copyPrevMemoryMap->fullMemoryMap(),m_intermediateMemory,MemoryMap::TotalMemoryMapSize);
    m_inMemoryUpdate = false;
}

void MainWindow::viewCloseRequested() {
    auto sel = dynamic_cast<const QWidget*>(sender());
    for (auto w:m_viewWidgets) {
        if (w == sel) {
            qDebug()<<"Close"<<w->typeId();
            w->hide();
            w->deleteLater();
            m_viewWidgets.removeAll(w);
            return;
        }
    }
}
