package cppgenerator

import AssignmentStatement
import CaseStatement
import ConstantValueExpression
import DataType
import IfStatement
import IndexAssignmentStatement
import ParsedExpression
import ParsedModule
import ParsedModuleFunction
import ParsedPureFunction
import ParsedStatement
import PureFunctionCall
import SpecifierExpression
import SpecifierIndexExpression
import SpecifierReference
import SubModuleCallStatement
import SubModuleInstance
import java.lang.Exception
import java.lang.StringBuilder
import java.math.BigInteger

fun dataTypeToCpp(dt:DataType):String = when(dt) {
    DataType.TypeBoolean -> "bool"
    DataType.TypeUInt32 -> "quint32"
    DataType.TypeUInt64 -> "quint64"
}
fun cppInitialValue(dt:DataType):String = when(dt) {
    DataType.TypeBoolean -> "false"
    DataType.TypeUInt32 -> "0U"
    DataType.TypeUInt64 -> "0ULL"
}

class CppGeneratorException(message:String) : Exception(message)

fun expressionToCpp(expr:ParsedExpression, nameResolver:(ref:SpecifierReference)->String):String = when(expr) {
    is ConstantValueExpression ->  when(expr.dataType) {
        DataType.TypeBoolean -> if (expr.value == BigInteger.ZERO) "false" else "true"
        DataType.TypeUInt32 -> "0x"+expr.value.toString(16)+"U"
        DataType.TypeUInt64 -> "0x"+expr.value.toString(16)+"ULL"
    }
    is SpecifierExpression -> nameResolver(expr.spec)
    is SpecifierIndexExpression -> nameResolver(expr.spec)+"["+ expressionToCpp(expr.index,nameResolver)+"]"
    is PureFunctionCall -> {
        val paramStr = ArrayList<String>()
        for (p in expr.parameters) {
            paramStr.add(expressionToCpp(p,nameResolver))
        }
        expr.func.toCpp(paramStr)
    }
}

fun removeBrackets(str:String):String {
    return str
    //TODO der folgende code hat folgenden fehler ((a)&(b)) wird zu (a)&(b)
    /*var result = str
    while (result.length>=2 && result[0] == '(' && result[result.length-1] == ')')
        result = result.substring(1,result.length-1)
    return result*/
}

fun ifStatementToCpp(indent:String, dest:StringBuilder, s:IfStatement, nameResolver: (ref: SpecifierReference) -> String, getSubModule: (subModuleIndex: Int) -> SubModuleInstance) {
    val subIndent = "$indent    "
    dest.append("if (${removeBrackets(expressionToCpp(s.condition,nameResolver))})")
    if (s.ifBody.isEmpty())
        dest.append("{}\n")
    else if (s.ifBody.size == 1 && s.ifBody.first() is AssignmentStatement) {
        dest.append("\n")
        statementToCpp(subIndent, dest, s.ifBody, nameResolver, getSubModule)
    }
    else {
        dest.append(" {\n")
        statementToCpp(subIndent, dest, s.ifBody, nameResolver, getSubModule)
        dest.append("${indent}}\n")
    }
    if (s.elseBody.isEmpty())
        return
    if (s.elseBody.size == 1) {
        val elseStm = s.elseBody.first()
        if (elseStm is IfStatement) {
            dest.append("${indent}else ")
            ifStatementToCpp(indent, dest, elseStm, nameResolver, getSubModule)
            return
        }
        if (elseStm is AssignmentStatement) {
            dest.append("${indent}else\n")
            statementToCpp(subIndent, dest, s.elseBody, nameResolver, getSubModule)
            return
        }
    }
    dest.append("${indent}else {\n")
    statementToCpp(subIndent, dest, s.elseBody, nameResolver, getSubModule)
    dest.append("${indent}}\n")
}

fun statementToCpp(indent: String, dest:StringBuilder, stmLst: List<ParsedStatement>, nameResolver:(ref:SpecifierReference)->String, getSubModule:(subModuleIndex:Int)->SubModuleInstance) {
    val subIndent = "$indent    "
    for (s in stmLst) {
        when(s) {
            is IfStatement -> {
                dest.append(indent)
                ifStatementToCpp(indent,dest, s,nameResolver, getSubModule)
            }
            is CaseStatement -> {
                dest.append("${indent}switch (${expressionToCpp(s.expresssion, nameResolver)}) {\n")
                for (i in s.items) {
                    dest.append(indent)
                    for (c in i.values)
                        dest.append("case ${expressionToCpp(c, nameResolver)}: ")
                    dest.append("{\n")
                    statementToCpp(subIndent, dest, i.body, nameResolver, getSubModule)
                    dest.append("${subIndent}break;\n")
                    dest.append("${indent}}\n")
                }
                dest.append("${indent}}\n")
            }
            is AssignmentStatement -> {
                dest.append("${indent}${nameResolver(s.dest)} = ${expressionToCpp(s.expr, nameResolver)};\n")
            }
            is IndexAssignmentStatement -> {
                dest.append("${indent}${nameResolver(s.dest)}[${expressionToCpp(s.index, nameResolver)}] = ${expressionToCpp(s.value, nameResolver)};\n")
            }
            is SubModuleCallStatement -> {
                val subMod = getSubModule(s.subModuleIndex)
                val args = ArrayList<String>()
                for (inp in s.inputs)
                    args.add(expressionToCpp(inp,nameResolver))
                for (outp in subMod.type.moduleFunctions[s.evaluationStep].output) {
                    val fullName = "out_${subMod.name}_${outp.name}"
                    dest.append(indent+dataTypeToCpp(outp.dataType)+" "+fullName+" = "+cppInitialValue(outp.dataType)+";\n")
                    args.add(fullName)
                }

                dest.append("${indent}mod_${subMod.name}.evaluate${s.evaluationStep}(${args.joinToString(",")});\n")
            }
        }
    }
}

fun pureFunctionToCpp(indent: String, dest:StringBuilder, pf:ParsedPureFunction) {
    val args = ArrayList<String>()
    for (p in pf.inputParams)
        args.add("const "+dataTypeToCpp(p.dataType)+" "+p.name)
    dest.append("${indent}static ${dataTypeToCpp(pf.returnValue)} ${pf.name} (${args.joinToString(",")}) {\n")

    for (l in pf.locals)
        dest.append("$indent    ${dataTypeToCpp(l.dataType)} ${l.name} = ${cppInitialValue(l.dataType)};\n")
    dest.append("$indent    ${dataTypeToCpp(pf.returnValue)} returnValue = ${cppInitialValue(pf.returnValue)};\n")

    statementToCpp("$indent    ", dest ,pf.body, {
        //Name Lookup
        specifierReference: SpecifierReference ->
        if (specifierReference.type == SpecifierReference.Type.Input)
            pf.inputParams[specifierReference.listIndex].name
        else if (specifierReference.type == SpecifierReference.Type.Local)
            pf.locals[specifierReference.listIndex].name
        else if (specifierReference.type == SpecifierReference.Type.Output && specifierReference.listIndex == 0)
            "returnValue"
        else
            throw CppGeneratorException("invalid pure function variable type ${specifierReference.type}")
    },{
        throw CppGeneratorException("no submodules allowed in pure functions")
    })
    dest.append("${indent}    return returnValue;\n")
    dest.append("${indent}}\n\n")
}

fun moduleFunctionToCpp(indent: String, dest:StringBuilder, mod:ParsedModule, f:ParsedModuleFunction, no:Int) {
    val args = ArrayList<String>()
    for (p in f.input)
        args.add("const "+dataTypeToCpp(p.dataType)+" "+p.name)
    for (p in f.output)
        args.add(dataTypeToCpp(p.dataType)+"& "+p.name)

    dest.append("${indent}__attribute__((always_inline)) void evaluate$no (${args.joinToString(",")}) {\n")
    for (l in f.locals)
        dest.append("$indent    ${dataTypeToCpp(l.dataType)} ${l.name} = ${cppInitialValue(l.dataType)};\n")
    for (o in f.output)
        dest.append("$indent    ${o.name} = ${cppInitialValue(o.dataType)};\n")
    statementToCpp("$indent    ", dest ,f.body, {
        //Name Lookup
        specifierReference: SpecifierReference ->
        if (specifierReference.subModuleIndex<0) { //eigenes modul
            val lst = if (specifierReference.evaluationStepIndex == -1 && specifierReference.type == SpecifierReference.Type.Register)
                mod.registers
            else {
                if (specifierReference.evaluationStepIndex != no)
                    throw CppGeneratorException("internal error: name resolve failed, evaluationStepIndex mismatch ${specifierReference.evaluationStepIndex} $no ${specifierReference.type}")
                when (specifierReference.type) {
                    SpecifierReference.Type.Input -> f.input
                    SpecifierReference.Type.Output -> f.output
                    SpecifierReference.Type.Local -> f.locals
                    else -> throw CppGeneratorException("internal error: name resolve failed, type mismatch: ${specifierReference.type} ${specifierReference.evaluationStepIndex}")
                }
            }
            lst[specifierReference.listIndex].name
        }
        else {
            if (specifierReference.type != SpecifierReference.Type.FromSubModule)
                throw CppGeneratorException("internal error: specifier reference subModule invalid type: ${specifierReference.type}")
            val subMod = mod.subModules[specifierReference.subModuleIndex]
            "out_"+subMod.name+"_"+subMod.type.moduleFunctions[specifierReference.evaluationStepIndex].output[specifierReference.listIndex].name
        }
    },{
        subModuleIndex -> mod.subModules[subModuleIndex]
    })
    dest.append("$indent}\n\n")
}

fun generateCppForParsedModule(mainModule:ParsedModule, destStr:StringBuilder, indent: String, generatedSubModules:HashSet<ParsedModule>) {
    if (generatedSubModules.contains(mainModule) || mainModule.isExternModule)
        return
    generatedSubModules.add(mainModule)
    destStr.append("${indent}struct module_${mainModule.name} {\n")
    for (sm in mainModule.subModules) {
        generateCppForParsedModule(sm.type, destStr, "$indent    ", generatedSubModules)
        destStr.append("${indent}    module_${sm.type.name} mod_${sm.name};\n")
    }
    for (reg in mainModule.registers) {
        if (reg.arraySize>=0)
            destStr.append("$indent    ${dataTypeToCpp(reg.dataType)} ${reg.name}[${reg.arraySize}];\n")
        else
            destStr.append("$indent    ${dataTypeToCpp(reg.dataType)} ${reg.name};\n")
    }

    if (mainModule.registers.isNotEmpty()) {
        destStr.append("$indent    module_${mainModule.name}():")
        for ((i,reg) in mainModule.registers.withIndex()) {
            if (reg.arraySize>=0)
                continue
            if (i>0)
                destStr.append(", ")
            destStr.append("${reg.name}(${cppInitialValue(reg.dataType)})")
        }
        destStr.append("{\n")
        for ((i,reg) in mainModule.registers.withIndex()) {
            if (reg.arraySize<0)
                continue
            destStr.append("$indent        for(int arridx=0; arridx<${reg.arraySize}; ++arridx)\n")
            destStr.append("$indent            ${reg.name}[arridx] = ${cppInitialValue(reg.dataType)};\n")
        }
        destStr.append("$indent    }\n")
    }

    for (pf in mainModule.pureFunctions)
        pureFunctionToCpp("$indent    ", destStr, pf)
    for ((i,f) in mainModule.moduleFunctions.withIndex())
        moduleFunctionToCpp("$indent    ",destStr, mainModule, f, i)

    destStr.append("$indent    void serialize(QDataStream& stream) const {\n")
    for (sub in mainModule.subModules)
        destStr.append("$indent    mod_${sub.name}.serialize(stream);\n")
    for (r in mainModule.registers) {
        if (r.arraySize<0)
            destStr.append("$indent        stream << ${r.name};\n")
        else {
            destStr.append("$indent        for (int arridx=0; arridx<${r.arraySize}; ++arridx)\n")
            destStr.append("$indent            stream << ${r.name}[arridx];\n")
        }
    }
    destStr.append("$indent    }\n")

    destStr.append("$indent    void deserialize(QDataStream& stream) {\n")
    for (sub in mainModule.subModules)
        destStr.append("$indent        mod_${sub.name}.deserialize(stream);\n")
    for (r in mainModule.registers) {
        if (r.arraySize<0)
            destStr.append("$indent        stream >> ${r.name};\n")
        else {
            destStr.append("$indent        for (int arridx=0; arridx<${r.arraySize}; ++arridx)\n")
            destStr.append("$indent            stream >> ${r.name}[arridx];\n")
        }
    }
    destStr.append("$indent    }\n")

    destStr.append("$indent    QList<QPair<QString,QString> > listInnerVars(QString prefix) {\n")
    destStr.append("$indent        QList<QPair<QString,QString> > result;\n")
    for (sub in mainModule.subModules)
        destStr.append("$indent        result += mod_${sub.name}.listInnerVars(prefix+\"${sub.name}.\");\n")
    for (r in mainModule.registers) {
        if (r.arraySize<0) {
            if (r.dataType == DataType.TypeBoolean)
                destStr.append("$indent        result << QPair<QString,QString>(prefix+\"${r.name}\",${r.name} ? \"true\" : \"false\");\n")
            else
                destStr.append("$indent        result << QPair<QString,QString>(prefix+\"${r.name}\",\"0x\"+QString::number(${r.name},16));\n")
        }
        else { } //TODO arrays auch zurückgeben?
    }
    destStr.append("$indent        return result;\n")
    destStr.append("$indent    }\n")

    destStr.append("$indent};\n")
}
