package flatten

import AssignmentStatement
import CaseStatement
import IfStatement
import IndexAssignmentStatement
import ParsedExpression
import ParsedModule
import ParsedModuleFunction
import ParsedStatement
import Specifier
import SpecifierExpression
import SpecifierIndexExpression
import SpecifierReference
import SubModuleCallStatement
import java.lang.Exception

class FlattenException(message:String):Exception(message)

fun flattenModule(module:ParsedModule):ParsedModule {
    val resultRegisters = ArrayList(module.registers)
    val subModules = module.subModules
    val startOfSubModuleRegisters = ArrayList<Int>()
    val resultPureFunctions = ArrayList(module.pureFunctions)
    //suchen, dass alle untermodule intern sind, und register hinzufügen
    for (sm in subModules) {
        if (sm.type.isExternModule)
            throw FlattenException("extern modules are not supported in flatten")
        resultPureFunctions.addAll(sm.type.pureFunctions)
        startOfSubModuleRegisters.add(resultRegisters.size)
        for (reg in sm.type.registers)
            resultRegisters.add(Specifier("reg_"+sm.name+"_"+reg.name,reg.dataType,reg.arraySize))
    }

    fun replaceStatement(rootStmList:List<ParsedStatement>, subModuleCb:(stm:SubModuleCallStatement)->ArrayList<ParsedStatement>, specifierReplace:(spec:SpecifierReference)->SpecifierReference):ArrayList<ParsedStatement> {
        fun exprReplace(baseExpr:ParsedExpression) = baseExpr.iterateReplace {
            expr ->
            if (expr is SpecifierExpression)
                SpecifierExpression(specifierReplace(expr.spec),expr.dataType)
            else if (expr is SpecifierIndexExpression)
                SpecifierIndexExpression(specifierReplace(expr.spec),expr.index,expr.dataType)
            else
                expr
        }

        fun replStm(stmList:List<ParsedStatement>):ArrayList<ParsedStatement> {
            val result = ArrayList<ParsedStatement>()
            for (stm in stmList) {
                result.addAll(when (stm) {
                    is AssignmentStatement -> {
                        arrayListOf(AssignmentStatement(specifierReplace(stm.dest), exprReplace(stm.expr)))
                    }
                    is IndexAssignmentStatement -> arrayListOf(IndexAssignmentStatement(specifierReplace(stm.dest), exprReplace(stm.index), exprReplace(stm.value)))
                    is IfStatement -> arrayListOf(IfStatement(exprReplace(stm.condition), replStm(stm.ifBody), replStm(stm.elseBody)))
                    is CaseStatement -> arrayListOf(CaseStatement(exprReplace(stm.expresssion), stm.items.map {
                        CaseStatement.Item(it.values, replStm(it.body))
                    }))
                    is SubModuleCallStatement -> {
                        val newInputs = ArrayList<ParsedExpression>()
                        for (inp in stm.inputs)
                            newInputs.add(exprReplace(inp))
                        subModuleCb(SubModuleCallStatement(stm.subModuleIndex,stm.evaluationStep,newInputs))
                    }
                })
            }
            return result
        }
        return replStm(rootStmList)
    }



    fun flattenEvaluationStep(thisEvalStepIdx:Int, inFunc:ParsedModuleFunction):ParsedModuleFunction {
        val resultLocals = ArrayList(inFunc.locals)
        val startOfOutputsForSm = Array<Array<Int?>>(subModules.size) {Array<Int?>(subModules[it].type.moduleFunctions.size) {null} }


        val expressionReplaceSubModuleVars = fun(spec:SpecifierReference):SpecifierReference = when(spec.type) {
                SpecifierReference.Type.Local,SpecifierReference.Type.Input,SpecifierReference.Type.Output,SpecifierReference.Type.Register -> {
                    if (spec.type != SpecifierReference.Type.Register && spec.evaluationStepIndex != thisEvalStepIdx)
                        throw FlattenException("evaluations step index mismatch $thisEvalStepIdx ${spec.evaluationStepIndex} ${spec.subModuleIndex} ${spec.type}")
                    spec
                }
                SpecifierReference.Type.FromSubModule ->
                    SpecifierReference(-1,thisEvalStepIdx,
                            ((startOfOutputsForSm[spec.subModuleIndex][spec.evaluationStepIndex] ?: throw FlattenException("used output before call"))+ spec.listIndex),
                            SpecifierReference.Type.Local)
            }
        val replaceSubModuleCallStatement = fun(call:SubModuleCallStatement):ArrayList<ParsedStatement> {
            val resultStm = ArrayList<ParsedStatement>()
            val sm = subModules[call.subModuleIndex]
            val smFunc = sm.type.moduleFunctions[call.evaluationStep]
            val startOfInputs = resultLocals.size
            for ((i,inp) in smFunc.input.withIndex()) {
                val localIdx = resultLocals.size
                resultLocals.add(Specifier("in_"+sm.name+"_"+inp.name,inp.dataType,inp.arraySize))
                //input expressions wurden schon vorher ersetzt
                resultStm.add(AssignmentStatement(SpecifierReference(-1,thisEvalStepIdx,localIdx,SpecifierReference.Type.Local),call.inputs[i]))
            }
            val startOfLocals = resultLocals.size
            for (l in smFunc.locals)
                resultLocals.add(Specifier("loc_"+sm.name+"_"+l.name,l.dataType,l.arraySize))
            val startOfOutputs = resultLocals.size
            if (startOfOutputsForSm[call.subModuleIndex][call.evaluationStep] != null)
                throw FlattenException("submodule called twice")
            startOfOutputsForSm[call.subModuleIndex][call.evaluationStep] = startOfOutputs
            for (outp in smFunc.output)
                resultLocals.add(Specifier("outp_"+sm.name+"_"+outp.name,outp.dataType,outp.arraySize))

            resultStm.addAll(replaceStatement(smFunc.body,{ arrayListOf(it) /*expressions wurden schon ersetzt*/},{
                spec ->
                when(spec.type) {
                    SpecifierReference.Type.Local -> SpecifierReference(-1,thisEvalStepIdx,startOfLocals+spec.listIndex,SpecifierReference.Type.Local)
                    SpecifierReference.Type.Input -> SpecifierReference(-1,thisEvalStepIdx,startOfInputs+spec.listIndex,SpecifierReference.Type.Local)
                    SpecifierReference.Type.Output -> SpecifierReference(-1,thisEvalStepIdx,startOfOutputs+spec.listIndex,SpecifierReference.Type.Local)
                    SpecifierReference.Type.Register -> SpecifierReference(-1,-1,startOfSubModuleRegisters[call.subModuleIndex]+spec.listIndex,SpecifierReference.Type.Register)
                    SpecifierReference.Type.FromSubModule -> throw FlattenException("recursive submodules are not supported")
                }
            }))
            return resultStm
        }
        val resultBody = replaceStatement(inFunc.body,replaceSubModuleCallStatement,expressionReplaceSubModuleVars)
        return ParsedModuleFunction(inFunc.input,inFunc.output,resultLocals,resultBody)
    }
    val resultModFunctions = ArrayList<ParsedModuleFunction>()
    for ((i,modFun) in module.moduleFunctions.withIndex())
        resultModFunctions.add(flattenEvaluationStep(i,modFun))
    return ParsedModule(module.name,false,resultRegisters,resultModFunctions,resultPureFunctions,ArrayList())
}