package ilanggenerator

import AssignmentStatement
import CaseStatement
import ConstantValueExpression
import DataType
import IfStatement
import IndexAssignmentStatement
import ParsedExpression
import ParsedModule
import ParsedModuleFunction
import ParsedPureFunction
import ParsedStatement
import PureFunctionCall
import SpecifierExpression
import SpecifierIndexExpression
import SpecifierReference
import SubModuleCallStatement
import java.lang.Exception
import java.lang.StringBuilder
import java.math.BigInteger

fun dataTypeToILang(dt:DataType):String = when(dt) {
    DataType.TypeBoolean -> "B"
    DataType.TypeUInt32 -> "I"
    DataType.TypeUInt64 -> "L"
}

class ILangGeneratorException(message:String) : Exception(message)

fun expressionToILang(expr:ParsedExpression, nameResolver:(ref:SpecifierReference)->String):String = when(expr) {
    is ConstantValueExpression ->  when(expr.dataType) {
        DataType.TypeBoolean -> if (expr.value == BigInteger.ZERO) "false" else "true"
        DataType.TypeUInt32 -> "0I"+expr.value.toString(16)
        DataType.TypeUInt64 -> "0L"+expr.value.toString(16)
    }
    is SpecifierExpression -> nameResolver(expr.spec)
    is SpecifierIndexExpression -> "(arrayAccess "+nameResolver(expr.spec)+" "+ expressionToILang(expr.index,nameResolver)+")"
    is PureFunctionCall -> {
        var result = "("+expr.func.toILang()
        for (p in expr.parameters)
            result += " "+expressionToILang(p,nameResolver)
        result+")"
    }
}

fun statementToILang(indent: String, dest:StringBuilder, stmLst: List<ParsedStatement>, nameResolver:(ref:SpecifierReference)->String) {
    val subIndent = "$indent    "
    for (s in stmLst) {
        dest.append(indent)
        when(s) {
            is IfStatement -> {
                dest.append("if ${expressionToILang(s.condition, nameResolver)}:\n")
                if (s.ifBody.isNotEmpty()) {
                    dest.append("${subIndent}true:\n")
                    statementToILang("${subIndent}    ", dest, s.ifBody, nameResolver)
                }
                if (s.elseBody.isNotEmpty()) {
                    dest.append("${subIndent}false:\n")
                    statementToILang("${subIndent}    ", dest, s.elseBody, nameResolver)
                }
            }
            is CaseStatement -> {
                dest.append("switch ${expressionToILang(s.expresssion, nameResolver)}:\n")
                for (i in s.items) {
                    dest.append("${subIndent}case")
                    for (c in i.values)
                        dest.append(" ${expressionToILang(c, nameResolver)}")
                    dest.append(":\n")
                    statementToILang("${subIndent}    ", dest, i.body, nameResolver)
                }
            }
            is AssignmentStatement -> {
                dest.append("let ${nameResolver(s.dest)} ${expressionToILang(s.expr, nameResolver)}\n")
            }
            is IndexAssignmentStatement -> {
                dest.append("letarr ${nameResolver(s.dest)} ${expressionToILang(s.index, nameResolver)} ${expressionToILang(s.value, nameResolver)}\n")
            }
            is SubModuleCallStatement -> throw ILangGeneratorException("ILang does not support submodules")
        }
    }
}

fun pureFunctionToILang(indent: String, dest:StringBuilder, pf:ParsedPureFunction) {
    dest.append("${indent}function ${pf.name}:\n")
    val subIndent = "${indent}    "
    for (p in pf.inputParams)
        dest.append("${subIndent}input ${dataTypeToILang(p.dataType)} ${p.name}\n")
    dest.append("${subIndent}output ${dataTypeToILang(pf.returnValue)} returnValue\n")
    for (p in pf.locals)
        dest.append("${subIndent}local ${dataTypeToILang(p.dataType)} ${p.name}\n")
    dest.append("${subIndent}body:\n")
    statementToILang("${subIndent}    ", dest ,pf.body) {
        //Name Lookup
        specifierReference: SpecifierReference ->
        if (specifierReference.type == SpecifierReference.Type.Input)
            pf.inputParams[specifierReference.listIndex].name
        else if (specifierReference.type == SpecifierReference.Type.Local)
            pf.locals[specifierReference.listIndex].name
        else if (specifierReference.type == SpecifierReference.Type.Output && specifierReference.listIndex == 0)
            "returnValue"
        else
            throw ILangGeneratorException("invalid pure function variable type ${specifierReference.type}")
    }
}

fun iLangModuleFunctionNameLoopkup(mod:ParsedModule, no:Int, specifierReference: SpecifierReference):String {
    val f = mod.moduleFunctions[no]
    return if (specifierReference.subModuleIndex<0) { //eigenes modul
        val lst = if (specifierReference.evaluationStepIndex == -1 && specifierReference.type == SpecifierReference.Type.Register)
            mod.registers
        else {
            if (specifierReference.evaluationStepIndex != no)
                throw ILangGeneratorException("internal error: name resolve failed, evaluationStepIndex mismatch ${specifierReference.evaluationStepIndex} $no ${specifierReference.type}")
            when (specifierReference.type) {
                SpecifierReference.Type.Input -> f.input
                SpecifierReference.Type.Output -> f.output
                SpecifierReference.Type.Local -> f.locals
                else -> throw ILangGeneratorException("internal error: name resolve failed, type mismatch: ${specifierReference.type} ${specifierReference.evaluationStepIndex}")
            }
        }
        lst[specifierReference.listIndex].name
    }
    else {
        if (specifierReference.type != SpecifierReference.Type.FromSubModule)
            throw ILangGeneratorException("internal error: specifier reference subModule invalid type: ${specifierReference.type}")
        val subMod = mod.subModules[specifierReference.subModuleIndex]
        "out_"+subMod.name+"_"+subMod.type.moduleFunctions[specifierReference.evaluationStepIndex].output[specifierReference.listIndex].name
    }
}

fun moduleFunctionToILang(indent: String, dest:StringBuilder, mod:ParsedModule, f:ParsedModuleFunction, no:Int, name:String, useRegSet:List<String>) {
    dest.append("${indent}function $name:\n")
    val subIndent = "${indent}    "
    for (r in useRegSet)
        dest.append("${subIndent}useregset $r\n")
    for (p in f.input)
        dest.append("${subIndent}input ${dataTypeToILang(p.dataType)} ${p.name}\n")
    for (p in f.output)
        dest.append("${subIndent}output ${dataTypeToILang(p.dataType)} ${p.name}\n")
    for (p in f.locals)
        dest.append("${subIndent}local ${dataTypeToILang(p.dataType)} ${p.name}\n")
    dest.append("${subIndent}body:\n")
    statementToILang("$subIndent    ", dest ,f.body) { iLangModuleFunctionNameLoopkup(mod, no, it) }
}

fun regSetToILang(flattenedModule:ParsedModule, destStr: StringBuilder, indent: String, name:String) {
    destStr.append("${indent}regset $name:\n")
    for (reg in flattenedModule.registers) {
        if (reg.arraySize>=0)
            destStr.append("$indent    regarray ${dataTypeToILang(reg.dataType)} ${reg.name} 0I${reg.arraySize.toString(16)}\n")
        else
            destStr.append("$indent    reg ${dataTypeToILang(reg.dataType)} ${reg.name}\n")
    }
}

fun generateILangForParsedModule(flattenedModule:ParsedModule, destStr: StringBuilder, indent: String) {
    destStr.append("${indent}module ${flattenedModule.name}:\n")
    for (reg in flattenedModule.registers) {
        if (reg.arraySize>=0)
            destStr.append("$indent    regarray ${dataTypeToILang(reg.dataType)} ${reg.name} 0I${reg.arraySize.toString(16)}\n")
        else
            destStr.append("$indent    reg ${dataTypeToILang(reg.dataType)} ${reg.name}\n")
    }

    for (pf in flattenedModule.pureFunctions)
        pureFunctionToILang("$indent    ", destStr, pf)
    for ((i,f) in flattenedModule.moduleFunctions.withIndex())
        moduleFunctionToILang("$indent    ",destStr, flattenedModule, f, i, "evalstep$i",listOf())

}
