package simplify

import AssignmentStatement
import CaseStatement
import ConstantValueExpression
import IfStatement
import IndexAssignmentStatement
import ParsedExpression
import ParsedModule
import ParsedStatement
import SpecifierExpression
import SpecifierReference
import SubModuleCallStatement
import cppgenerator.CppGeneratorException
import cppgenerator.statementToCpp
import isFixedFalse
import isFixedTrue
import java.lang.StringBuilder
import java.math.BigInteger


fun simplify(module:ParsedModule, evalStep:Int, initialFixedValues:HashMap<String,BigInteger>):List<ParsedStatement> {
    val modFun = module.moduleFunctions[evalStep]


    fun replaceExpressions(baseExpression: ParsedExpression, varStateMap:VarStateMap<ParsedExpression?>) = simplifyExpression(baseExpression.iterateReplace {
        expr -> if (expr is SpecifierExpression) {
            val specExpr = varStateMap.find(expr.spec)
            if (specExpr !is ConstantValueExpression)
                expr
            else
                specExpr
        }
        else
            expr
    })

    val joinVarStatesCb = fun(expressions:List<ParsedExpression?>):ParsedExpression? {
        if (expressions.isEmpty())
            return null
        val exprA = expressions.first()
        for (i in 1 until expressions.size) {
            val exprB = expressions[i]
            if (exprA === exprB)
                continue
            if (exprA is ConstantValueExpression && exprB is ConstantValueExpression && exprA.value == exprB.value)
                continue
            return null
        }
        return exprA
    }

    fun replaceStatements(stmList:List<ParsedStatement>, varStateMap: VarStateMap<ParsedExpression?>):List<ParsedStatement> {
        val result = ArrayList<ParsedStatement>()
        for (stm in stmList) {
            when(stm) {
                is AssignmentStatement -> {
                    val expr = replaceExpressions(stm.expr, varStateMap)
                    result.add(AssignmentStatement(stm.dest, expr))
                    varStateMap.set(stm.dest,expr)
                }
                is IndexAssignmentStatement ->
                    result.add(IndexAssignmentStatement(stm.dest, replaceExpressions(stm.index, varStateMap), replaceExpressions(stm.value, varStateMap)))
                is IfStatement -> {
                    val cond = replaceExpressions(stm.condition, varStateMap)
                    if (isFixedTrue(cond))
                        result.addAll(replaceStatements(stm.ifBody, varStateMap))
                    else if (isFixedFalse(cond))
                        result.addAll(replaceStatements(stm.elseBody, varStateMap))
                    else {
                        val varStateMap1 = VarStateMap(varStateMap)
                        val varStateMap2 = VarStateMap(varStateMap)
                        val newIfStm = IfStatement(cond, replaceStatements(stm.ifBody, varStateMap1), replaceStatements(stm.elseBody, varStateMap2))
                        varStateMap.joinSets(arrayListOf(varStateMap1,varStateMap2),joinVarStatesCb)
                        if (newIfStm.ifBody.isNotEmpty() || newIfStm.elseBody.isNotEmpty())
                            result.add(newIfStm)
                    }
                }
                is CaseStatement -> {
                    val expr = replaceExpressions(stm.expresssion, varStateMap)
                    val newItems = ArrayList<CaseStatement.Item>()
                    val subVarStates = ArrayList<VarStateMap<ParsedExpression?>>()
                    var isNotEmpty = false
                    for (item in stm.items) {
                        val subVarStateMap = VarStateMap(varStateMap)
                        val newI = CaseStatement.Item(item.values, replaceStatements(item.body, subVarStateMap))
                        newItems.add(newI)
                        subVarStates.add(subVarStateMap)
                        if (newI.body.isNotEmpty())
                            isNotEmpty = true
                    }
                    val newCaseStm = CaseStatement(expr,newItems)
                    if (expr is ConstantValueExpression) {
                        val matchingItem = newCaseStm.getMatchingItem(expr.value) ?: throw SimplifyException("value ${expr.value} not handled in case")
                        result.addAll(matchingItem.body)
                    }
                    else if (isNotEmpty)
                        result.add(newCaseStm)
                    varStateMap.joinSets(subVarStates, joinVarStatesCb)
                }
                is SubModuleCallStatement->throw SimplifyException("submodule not supported in simplify")
            }
        }
        return result
    }

    fun doSimplifyStep(stmLst:List<ParsedStatement>):List<ParsedStatement> {
        println("Simplify...")
        return replaceStatements(stmLst, VarStateMap<ParsedExpression?>(module, evalStep) {
            spec, ref ->
            val initialValue = initialFixedValues[spec.name]
            if (initialValue != null)
                ConstantValueExpression(initialValue,spec.dataType)
            else if (ref.type == SpecifierReference.Type.Local)
                ConstantValueExpression(BigInteger.ZERO,spec.dataType) //lokale variablen sind am anfang 0
            else
                null
        })
    }

    var resultBody = modFun.body
    for (steps in 0 until 20)
        resultBody = removeUnusedAssignments(module,evalStep,doSimplifyStep(resultBody))
    return resultBody

    /*
    val destCpp = StringBuilder()
    statementToCpp("", destCpp ,resultBody, {
        //Name Lookup
        specifierReference: SpecifierReference ->
        when (specifierReference.type) {
            SpecifierReference.Type.Input -> modFun.input[specifierReference.listIndex].name
            SpecifierReference.Type.Output -> modFun.output[specifierReference.listIndex].name
            SpecifierReference.Type.Local -> modFun.locals[specifierReference.listIndex].name
            SpecifierReference.Type.Register -> module.registers[specifierReference.listIndex].name
            else -> throw SimplifyException("no submodules supported")
        }
    },{throw SimplifyException("no submodules supported") })
    println(destCpp)*/
}