import java.math.BigInteger

enum class DataType {TypeBoolean, TypeUInt32, TypeUInt64}

class Specifier(val name: String, val dataType: DataType, val arraySize:Int) {
    override fun toString() = name
}

class SpecifierReference(val subModuleIndex:Int, val evaluationStepIndex:Int, val listIndex:Int, val type:Type) {
    enum class Type {Input,Output,Local,Register,FromSubModule}
}

sealed class PureFunction(val parameters:List<DataType>, val returnValue:DataType) {
    abstract fun toCpp(params: List<String>): String
    abstract fun toILang(): String
    open fun partialEvaluate(args:List<ParsedExpression>):ParsedExpression = PureFunctionCall(this, args)
}
class CppBuildInFunction1(val ilangName:String, param0:DataType, returnValue: DataType, val toCppFunc:(p0:String)->String, val peval:(p0:ParsedExpression)->ParsedExpression?):PureFunction(
    arrayListOf(param0),returnValue) {
    override fun toCpp(params: List<String>): String = toCppFunc(params[0])
    override fun toILang(): String = ilangName
    override fun partialEvaluate(args:List<ParsedExpression>):ParsedExpression = peval(args[0]) ?: PureFunctionCall(this, args)
}
class CppBuildInFunction2(val ilangName:String, param0:DataType, param1:DataType, returnValue: DataType, val toCppFunc:(p0:String, p1:String)->String, val peval:(p0:ParsedExpression, p1:ParsedExpression)->ParsedExpression?):PureFunction(
    arrayListOf(param0, param1),returnValue) {
    override fun toCpp(params: List<String>): String = toCppFunc(params[0], params[1])
    override fun toILang(): String = ilangName
    override fun partialEvaluate(args:List<ParsedExpression>):ParsedExpression = peval(args[0], args[1]) ?: PureFunctionCall(this, args)
}
class CppBuildInFunction3(val ilangName:String, param0:DataType, param1:DataType, param2:DataType, returnValue: DataType, val toCppFunc:(p0:String, p1:String, p2:String)->String):PureFunction(
    arrayListOf(param0, param1, param2),returnValue) {
    override fun toILang(): String = ilangName
    override fun toCpp(params: List<String>): String = toCppFunc(params[0], params[1], params[2])
}
class ParsedPureFunction(val inputParams:List<Specifier>, val locals:List<Specifier>, returnValue:DataType, val name:String, val body:List<ParsedStatement>):PureFunction(inputParams.map { it.dataType }, returnValue) {
    override fun toCpp(params: List<String>): String {
        return name+"("+params.joinToString(",")+")"
    }
    override fun toILang(): String = name
}

sealed class ParsedExpression(val dataType: DataType) {
    abstract fun iterateReplace(cb:(expr:ParsedExpression)->ParsedExpression):ParsedExpression
    abstract fun iterate(cb:(expr:ParsedExpression)->Unit)
}
class ConstantValueExpression(val value:BigInteger, dataType: DataType):ParsedExpression(dataType) {
    override fun iterateReplace(cb:(expr:ParsedExpression)->ParsedExpression):ParsedExpression = cb(this)
    override fun iterate(cb:(expr:ParsedExpression)->Unit) = cb(this)
    override fun toString():String = "0x$value"
}
class SpecifierExpression(val spec:SpecifierReference, dataType: DataType):ParsedExpression(dataType) {
    override fun iterateReplace(cb:(expr:ParsedExpression)->ParsedExpression):ParsedExpression = cb(this)
    override fun iterate(cb:(expr:ParsedExpression)->Unit) = cb(this)
    override fun toString():String = "specifier(${spec.type},${spec.subModuleIndex},${spec.evaluationStepIndex},${spec.listIndex})"
}
class SpecifierIndexExpression(val spec:SpecifierReference, val index:ParsedExpression, dataType: DataType):ParsedExpression(dataType) {
    override fun iterateReplace(cb:(expr:ParsedExpression)->ParsedExpression):ParsedExpression {
        val newIdx = index.iterateReplace(cb)
        return cb(if (newIdx === index) this else SpecifierIndexExpression(spec,newIdx,dataType))
    }
    override fun iterate(cb:(expr:ParsedExpression)->Unit) {
        index.iterate(cb)
        cb(this)
    }
}
class PureFunctionCall(val func:PureFunction, val parameters:List<ParsedExpression>):ParsedExpression(func.returnValue) {
    override fun iterateReplace(cb:(expr:ParsedExpression)->ParsedExpression):ParsedExpression {
        val newArgs = ArrayList<ParsedExpression>()
        for (arg in parameters)
            newArgs.add(arg.iterateReplace(cb))
        return cb(if(newArgs == parameters) this else PureFunctionCall(func,newArgs))
    }
    override fun iterate(cb:(expr:ParsedExpression)->Unit) {
        for (a in parameters)
            a.iterate(cb)
        cb(this)
    }
}

sealed class ParsedStatement
class IfStatement(val condition:ParsedExpression, val ifBody:List<ParsedStatement>, val elseBody:List<ParsedStatement>):ParsedStatement()
class CaseStatement(val expresssion:ParsedExpression, val items:List<Item>):ParsedStatement() {
    class Item(val values:List<ConstantValueExpression>, val body:List<ParsedStatement>)
    fun getMatchingItem(value:BigInteger):Item? {
        for (i in items) {
            for (v in i.values)
                if (v.value == value)
                    return i
        }
        return null
    }
}
class AssignmentStatement(val dest:SpecifierReference, val expr:ParsedExpression):ParsedStatement()
class IndexAssignmentStatement(val dest:SpecifierReference, val index:ParsedExpression, val value:ParsedExpression):ParsedStatement()
class SubModuleCallStatement(val subModuleIndex:Int, val evaluationStep:Int, val inputs:List<ParsedExpression>):ParsedStatement()

class SubModuleInstance(val name:String, val type:ParsedModule)
class ParsedModuleFunction(val input:List<Specifier>, val output:List<Specifier>, val locals:List<Specifier>, val body:List<ParsedStatement>)
class ParsedModule(val name:String, val isExternModule:Boolean, val registers:List<Specifier>, val moduleFunctions:List<ParsedModuleFunction>, val pureFunctions:List<ParsedPureFunction>, val subModules:List<SubModuleInstance>) {
    fun findSpecifier(ref:SpecifierReference):Specifier = when(ref.type) {
        SpecifierReference.Type.Input -> moduleFunctions[ref.evaluationStepIndex].input[ref.listIndex]
        SpecifierReference.Type.Output -> moduleFunctions[ref.evaluationStepIndex].output[ref.listIndex]
        SpecifierReference.Type.Local -> moduleFunctions[ref.evaluationStepIndex].locals[ref.listIndex]
        SpecifierReference.Type.Register -> registers[ref.listIndex]
        SpecifierReference.Type.FromSubModule -> subModules[ref.subModuleIndex].type.moduleFunctions[ref.evaluationStepIndex].output[ref.listIndex]
    }
    fun getSpecifierReferenceByName(name:String, evaluationStepIndex: Int):SpecifierReference? {
        for (i in registers.indices)
            if (registers[i].name == name)
                return SpecifierReference(-1,-1,i,SpecifierReference.Type.Register)
        val modFun = moduleFunctions[evaluationStepIndex]
        for (i in modFun.input.indices)
            if (modFun.input[i].name == name)
                return SpecifierReference(-1,evaluationStepIndex,i,SpecifierReference.Type.Input)
        for (i in modFun.output.indices)
            if (modFun.output[i].name == name)
                return SpecifierReference(-1,evaluationStepIndex,i,SpecifierReference.Type.Output)
        for (i in modFun.locals.indices)
            if (modFun.locals[i].name == name)
                return SpecifierReference(-1,evaluationStepIndex,i,SpecifierReference.Type.Local)
        return null
    }
}

fun isFixedTrue(expr:ParsedExpression) = expr is ConstantValueExpression && expr.value != BigInteger.ZERO
fun isFixedFalse(expr:ParsedExpression) = expr is ConstantValueExpression && expr.value == BigInteger.ZERO
fun isFixedZero(expr:ParsedExpression) = expr is ConstantValueExpression && expr.value == BigInteger.ZERO
val constTrue = ConstantValueExpression(BigInteger.ONE,DataType.TypeBoolean)
val constFalse = ConstantValueExpression(BigInteger.ZERO,DataType.TypeBoolean)
val constZero = ConstantValueExpression(BigInteger.ZERO,DataType.TypeBoolean)

fun builtInFunctions():HashMap<String,PureFunction> {
    val mask32 = BigInteger.valueOf(0xFFFFFFFFL)
    val mask64 = BigInteger.ONE.shiftLeft(64)-BigInteger.ONE
    return hashMapOf<String,PureFunction>(
    "add32" to CppBuildInFunction2("add32", DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeUInt32, { p0,p1 -> "($p0 + $p1)" }, {p0,p1-> if (isFixedZero(p0)) p1 else if (isFixedZero(p1)) p0 else if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression((p0.value+p1.value).and(mask32),DataType.TypeUInt32) else null}),
    "sub32" to CppBuildInFunction2("sub32", DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeUInt32, { p0,p1 -> "($p0 - $p1)" }, {p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression((p0.value-p1.value).and(mask32),DataType.TypeUInt32) else null}),
    "and32" to CppBuildInFunction2("and32", DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeUInt32, { p0,p1 -> "($p0 & $p1)" }, {p0,p1-> if (isFixedZero(p0) || isFixedZero(p1)) constZero else if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression(p0.value.and(p1.value),DataType.TypeUInt32) else null}),
    "or32" to CppBuildInFunction2("or32", DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeUInt32, { p0,p1 -> "($p0 | $p1)" }, {p0,p1-> if (isFixedZero(p0)) p1 else if (isFixedZero(p1)) p0 else if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression(p0.value.or(p1.value),DataType.TypeUInt32) else null}),
    "xor32" to CppBuildInFunction2("xor32", DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeUInt32, { p0,p1 -> "($p0 ^ $p1)" }, {p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression(p0.value.xor(p1.value),DataType.TypeUInt32) else null}),
    "shr32" to CppBuildInFunction2("shr32", DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeUInt32, { p0,p1 -> "($p0 >> $p1)"}, {p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression(p0.value.shr(p1.value.toInt()).and(mask32),DataType.TypeUInt32) else null}),
    "shl32" to CppBuildInFunction2("shl32", DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeUInt32, { p0,p1 -> "($p0 << $p1)" }, {p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression(p0.value.shl(p1.value.toInt()).and(mask32),DataType.TypeUInt32) else null}),
    "testBit32" to CppBuildInFunction2("testBit32", DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeBoolean, { p0,p1 -> "(($p0 & (1U<<$p1)) != 0U)" }, {p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression(if(p0.value.testBit(p1.value.toInt())) BigInteger.ONE else BigInteger.ZERO,DataType.TypeBoolean) else null}),
    "invert32" to CppBuildInFunction1("invert32", DataType.TypeUInt32,DataType.TypeUInt32, { p0 -> "(~$p0)" }, {p0 -> if (p0 is ConstantValueExpression) ConstantValueExpression(p0.value.not().and(mask32),DataType.TypeUInt32) else null}),
    "eq32" to CppBuildInFunction2("eq32", DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeBoolean, { p0,p1 -> "($p0 == $p1)" }, {p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression(if(p0.value == p1.value) BigInteger.ONE else BigInteger.ZERO,DataType.TypeBoolean) else null}),
    "neq32" to CppBuildInFunction2("neq32", DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeBoolean, { p0,p1 -> "($p0 != $p1)" }, {p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression(if(p0.value != p1.value) BigInteger.ONE else BigInteger.ZERO,DataType.TypeBoolean) else null}),
    "sgnle32" to CppBuildInFunction2("sgnle32", DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeBoolean, { p0,p1 -> "(qint32($p0) < qint32($p1))" }, {
            p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) {
            fun convToSigned(v:BigInteger):BigInteger = if (v.testBit(31)) BigInteger.valueOf(-0x80000000L) + v.and(BigInteger.valueOf(0x7FFFFFFF)) else v.and(BigInteger.valueOf(0x7FFFFFFF))
                ConstantValueExpression(if (convToSigned(p0.value) < convToSigned(p1.value)) BigInteger.ONE else BigInteger.ZERO, DataType.TypeBoolean)
            } else null}
    ),
    "le32" to CppBuildInFunction2("le32", DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeBoolean, { p0,p1 -> "($p0 < $p1)" }, {p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression(if(p0.value < p1.value) BigInteger.ONE else BigInteger.ZERO,DataType.TypeBoolean) else null}),
    "leq32" to CppBuildInFunction2("leq32", DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeBoolean, { p0,p1 -> "($p0 <= $p1)" }, {p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression(if(p0.value <= p1.value) BigInteger.ONE else BigInteger.ZERO,DataType.TypeBoolean) else null}),
    "ge32" to CppBuildInFunction2("ge32", DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeBoolean, { p0,p1 -> "($p0 > $p1)" }, {p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression(if(p0.value > p1.value) BigInteger.ONE else BigInteger.ZERO,DataType.TypeBoolean) else null}),
    "geq32" to CppBuildInFunction2("geq32", DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeBoolean, { p0,p1 -> "($p0 >= $p1)" }, {p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression(if(p0.value >= p1.value) BigInteger.ONE else BigInteger.ZERO,DataType.TypeBoolean) else null}),
    "inRange32" to CppBuildInFunction3("inRange32", DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeBoolean) { p0,p1,p2 -> "($p0 >= $p1 && $p0 <= $p2)" },
    "mux32" to CppBuildInFunction3("mux32", DataType.TypeBoolean,DataType.TypeUInt32,DataType.TypeUInt32,DataType.TypeUInt32) { p0,p1,p2 -> "($p0 ? $p1 : $p2)" },
    "mux64" to CppBuildInFunction3("mux64", DataType.TypeBoolean,DataType.TypeUInt64,DataType.TypeUInt64,DataType.TypeUInt64) { p0,p1,p2 -> "($p0 ? $p1 : $p2)" },
    "lower32" to CppBuildInFunction1("lower32", DataType.TypeUInt64,DataType.TypeUInt32, { p0 -> "quint32($p0 & 0xFFFFFFFFULL)" }, {
        p0 -> if (p0 is ConstantValueExpression) ConstantValueExpression(p0.value.and(mask32),DataType.TypeUInt32) else null}),

    "add64" to CppBuildInFunction2("add64", DataType.TypeUInt64,DataType.TypeUInt64,DataType.TypeUInt64, { p0,p1 -> "($p0 + $p1)" },{p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression((p0.value+p1.value).and(mask64),DataType.TypeUInt64) else null}),
    "sub64" to CppBuildInFunction2("sub64", DataType.TypeUInt64,DataType.TypeUInt64,DataType.TypeUInt64, { p0,p1 -> "($p0 - $p1)" },{p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression((p0.value-p1.value).and(mask64),DataType.TypeUInt64) else null}),
    "and64" to CppBuildInFunction2("and64", DataType.TypeUInt64,DataType.TypeUInt64,DataType.TypeUInt64, { p0,p1 -> "($p0 & $p1)" },{p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression((p0.value.and(p1.value)),DataType.TypeUInt64) else null}),
    "or64" to CppBuildInFunction2("or64", DataType.TypeUInt64,DataType.TypeUInt64,DataType.TypeUInt64, { p0,p1 -> "($p0 | $p1)" },{p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression((p0.value.or(p1.value)),DataType.TypeUInt64) else null}),
    "shr64" to CppBuildInFunction2("shr64", DataType.TypeUInt64,DataType.TypeUInt32,DataType.TypeUInt64, { p0,p1 -> "($p0 >> $p1)" },{p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression((p0.value.shr(p1.value.toInt())).and(mask64),DataType.TypeUInt64) else null}),
    "shl64" to CppBuildInFunction2("shl64", DataType.TypeUInt64,DataType.TypeUInt32,DataType.TypeUInt64, { p0,p1 -> "($p0 << $p1)" },{p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression((p0.value.shl(p1.value.toInt())).and(mask64),DataType.TypeUInt64) else null}),
    "eq64" to CppBuildInFunction2("eq64", DataType.TypeUInt64,DataType.TypeUInt64,DataType.TypeBoolean, { p0,p1 -> "($p0 == $p1)" },{p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression(if(p0.value == p1.value) BigInteger.ONE else BigInteger.ZERO,DataType.TypeBoolean) else null}),
    "neq64" to CppBuildInFunction2("neq64", DataType.TypeUInt64,DataType.TypeUInt64,DataType.TypeBoolean, { p0,p1 -> "($p0 != $p1)" }, {p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression(if(p0.value != p1.value) BigInteger.ONE else BigInteger.ZERO,DataType.TypeBoolean) else null}),
    "testBit64" to CppBuildInFunction2("testBit64", DataType.TypeUInt64,DataType.TypeUInt32,DataType.TypeBoolean, { p0,p1 -> "(($p0 & (1ULL<<$p1)) != 0ULL)" }, {p0,p1-> if (p0 is ConstantValueExpression && p1 is ConstantValueExpression) ConstantValueExpression(if(p0.value.testBit(p1.value.toInt())) BigInteger.ONE else BigInteger.ZERO,DataType.TypeBoolean) else null}),
    "cast64" to CppBuildInFunction1("cast64", DataType.TypeUInt32,DataType.TypeUInt64, { p0 -> "quint64($p0)" },{p0 -> if (p0 is ConstantValueExpression) ConstantValueExpression(p0.value.and(mask64),DataType.TypeUInt64) else null}),

    "muxB" to CppBuildInFunction3("muxB", DataType.TypeBoolean,DataType.TypeBoolean,DataType.TypeBoolean,DataType.TypeBoolean) { p0,p1,p2 -> "($p0 ? $p1 : $p2)" },
    "orB" to CppBuildInFunction2("orB", DataType.TypeBoolean,DataType.TypeBoolean,DataType.TypeBoolean, { p0,p1 -> "($p0 || $p1)" }, {p0, p1 -> if (isFixedTrue(p0) || isFixedTrue(p1)) constTrue else if (isFixedFalse(p0)) p1 else if (isFixedFalse(p1)) p0 else null}),
    "andB" to CppBuildInFunction2("andB", DataType.TypeBoolean,DataType.TypeBoolean,DataType.TypeBoolean, { p0,p1 -> "($p0 && $p1)" },{p0, p1 -> if (isFixedFalse(p0) || isFixedFalse(p1)) constFalse else if (isFixedTrue(p0)) p1 else if (isFixedTrue(p1)) p0 else null}),
    "notB" to CppBuildInFunction1("notB", DataType.TypeBoolean,DataType.TypeBoolean, { p0 -> "(!$p0)" }, {p0 -> if (isFixedTrue(p0)) constFalse else if (isFixedFalse(p0)) constTrue else null}),
    "eqB" to CppBuildInFunction2("eqB", DataType.TypeBoolean,DataType.TypeBoolean,DataType.TypeBoolean, { p0,p1 -> "($p0 == $p1)" }, {p0, p1 -> if (isFixedTrue(p0)) { if (isFixedTrue(p1)) constTrue else if (isFixedFalse(p1)) constFalse else null} else if (isFixedFalse(p0)) { if (isFixedTrue(p1)) constFalse else if (isFixedFalse(p1)) constTrue else null} else null}),
    "neqB" to CppBuildInFunction2("neqB", DataType.TypeBoolean,DataType.TypeBoolean,DataType.TypeBoolean, { p0,p1 -> "($p0 != $p1)" }, {p0, p1 -> if (isFixedTrue(p0)) { if (isFixedTrue(p1)) constFalse else if (isFixedFalse(p1)) constTrue else null} else if (isFixedFalse(p0)) { if (isFixedTrue(p1)) constTrue else if (isFixedFalse(p1)) constFalse else null} else null})
)}
