package verilogparser

import AssignmentStatement
import CaseStatement
import ConstantValueExpression
import DataType
import IfStatement
import ParsedExpression
import ParsedStatement
import PureFunction
import PureFunctionCall
import Specifier
import SpecifierExpression
import SpecifierReference
import SubModuleCallStatement
import SubModuleInstance
import java.math.BigInteger

class ParsedFunction(val pureFunctionMap:HashMap<String, PureFunction>, val evaluationStepIndex:Int, val subModuleList:ArrayList<SubModuleInstance>) {
    private val specifiers = HashMap<String, ParsedSpecifier>()
    private val constants = HashMap<String, ConstantValueExpression>()
    val functionBody = ArrayList<ParsedStatement>()
    val inputs = ArrayList<Specifier>()
    val outputs = ArrayList<Specifier>()
    val locals = ArrayList<Specifier>()

    private fun specifierInUse(name:String) = specifiers.containsKey(name) || constants.containsKey(name)

    private fun forceSpecifier(pos: SourcePosition, name:String): ParsedSpecifier = specifiers[name] ?: throw ParserException(
        pos,
        "no specifier '$name' found"
    )

    private fun addSpecifier(pos:SourcePosition, specifier: ParsedSpecifier) {
        if (specifierInUse(specifier.verilogName))
            throw ParserException(pos, "duplicate specifier '${specifier.verilogName}'")
        specifiers[specifier.verilogName] = specifier
    }

    fun addInput(pos:SourcePosition, verilogName: String, dataType: DataType) {
        val spec = Specifier(verilogName, dataType,-1)
        addSpecifier(pos,ParsedSpecifier(verilogName, ParsedSpecifier.UsageType.InputPort, dataType, SpecifierReference(-1, evaluationStepIndex, inputs.size, SpecifierReference.Type.Input)))
        inputs.add(spec)
    }

    fun addOutput(pos:SourcePosition, verilogName: String, dataType: DataType) {
        val spec = Specifier(verilogName, dataType,-1)
        addSpecifier(pos,ParsedSpecifier(verilogName, ParsedSpecifier.UsageType.OutputPort, dataType, SpecifierReference(-1, evaluationStepIndex, outputs.size, SpecifierReference.Type.Output)))
        outputs.add(spec)
    }

    fun addLocal(pos:SourcePosition, verilogName: String, dataType: DataType) {
        val spec = Specifier(verilogName, dataType,-1)
        addSpecifier(pos,ParsedSpecifier(verilogName, ParsedSpecifier.UsageType.LocalVariable, dataType, SpecifierReference(-1, evaluationStepIndex, locals.size, SpecifierReference.Type.Local)))
        locals.add(spec)
    }

    fun addRegister(pos:SourcePosition, verilogName: String, dataType: DataType, readWrite:Boolean, registerIndex:Int) {
        addSpecifier(pos,ParsedSpecifier(verilogName, if (readWrite) ParsedSpecifier.UsageType.RegisterRW else ParsedSpecifier.UsageType.RegisterRO, dataType, SpecifierReference(-1, -1, registerIndex, SpecifierReference.Type.Register)))
    }

    fun addFromSubModule(pos:SourcePosition, verilogName: String, dataType: DataType, subModuleIndex:Int, stepIndex:Int, portIndex:Int) {
        addSpecifier(pos,ParsedSpecifier(verilogName, ParsedSpecifier.UsageType.FromSubModule,dataType,SpecifierReference(subModuleIndex, stepIndex, portIndex, SpecifierReference.Type.FromSubModule)))
    }

    fun addConstant(pos:SourcePosition, verilogName: String, constValue:BigInteger, dataType: DataType) {
        if (specifierInUse(verilogName))
            throw ParserException(pos, "duplicate specifier '${verilogName}'")
        constants[verilogName] = ConstantValueExpression(constValue, dataType)
    }

    fun getFilteredSpecifiers(filter:(sp:ParsedSpecifier)->Boolean):List<ParsedSpecifier> {
        return specifiers.values.filter(filter)
    }

    private fun parseFunctionExpression(it: TokenIterator, func: PureFunction, funcName:String): ParsedExpression {
        val parsedParams = ArrayList<ParsedExpression>()
        for (i in 0 until func.parameters.size) {
            if (it.atEnd())
                throw ParserException(it.pos(), "expected paramter $i of function '$funcName'")
            parsedParams.add(parseExpression(it, func.parameters[i]))
            if (i+1<func.parameters.size)
                it.forceOperator(",","expected ';'")
        }
        if (!it.atEnd())
            throw ParserException(it.pos(), "expected no more parameters in '$funcName'")
        return PureFunctionCall(func, parsedParams)
    }

    private fun parseExpression(it: TokenIterator, expectedDataType: DataType): ParsedExpression {
        val tk = it.token()
        if (tk is NumberToken) {
            it.next()
            return when (expectedDataType) {
                DataType.TypeBoolean -> {
                    if (tk.value != BigInteger.ZERO && tk.value != BigInteger.ONE)
                        throw ParserException(tk.pos, "invalid boolean value ${tk.value}")
                    ConstantValueExpression(tk.value,expectedDataType)
                }
                DataType.TypeUInt32 -> {
                    if (tk.value < BigInteger.ZERO || tk.value >= (BigInteger.ONE.shiftLeft(32)))
                        throw ParserException(tk.pos, "value ${tk.value} out of uint32 range")
                    ConstantValueExpression(tk.value,expectedDataType)
                }
                DataType.TypeUInt64 -> {
                    if (tk.value < BigInteger.ZERO || tk.value >= (BigInteger.ONE.shiftLeft(64)))
                        throw ParserException(tk.pos, "value ${tk.value} out of uint64 range")
                    ConstantValueExpression(tk.value,expectedDataType)
                }
            }
        }
        if (tk is GroupToken && tk.type == GroupType.Bracket) {
            it.next()
            return parseExpressionUntilEnd(TokenIterator(tk), expectedDataType)
        }
        if (tk !is AtomToken)
            throw ParserException(tk.pos, "invalid expression, $tk")
        it.next()
        val func = pureFunctionMap[tk.name]
        if (func != null) {
            if (func.returnValue != expectedDataType)
                throw ParserException(
                    tk.pos,
                    "expected datatype '$expectedDataType', but datatype of '${tk.name}' was '${func.returnValue}'"
                )
            return parseFunctionExpression(it.forceGroup(GroupType.Bracket, "expected argument in ()"), func, tk.name)
        }
        val const = constants[tk.name]
        if (const != null)
            return const
        val cppspec = forceSpecifier(tk.pos, tk.name)
        if (!cppspec.readAllowed())
            throw ParserException(
                tk.pos,
                "variable '${cppspec.verilogName}' of type '${cppspec.usage}' must not be used inside expressions"
            )
        if (cppspec.dataType != expectedDataType)
            throw ParserException(
                tk.pos,
                "expected datatype '$expectedDataType', but datatype of '${cppspec.verilogName}' was '${cppspec.dataType}'"
            )
        return SpecifierExpression(cppspec.ref,cppspec.dataType)
    }

    private fun parseExpressionUntilEnd(it: TokenIterator, expectedDataType: DataType): ParsedExpression {
        val res = parseExpression(it, expectedDataType)
        if (!it.atEnd())
            throw ParserException(it.pos(), "unexpected end of expression: ${it.token()}")
        return res
    }

    private fun parseIfStatement(it: TokenIterator): ParsedStatement {
        val condition = parseExpressionUntilEnd(it.forceGroup(GroupType.Bracket,"expected if condition in ()"),
            DataType.TypeBoolean
        )
        it.forceAtom("begin", "expected begin")
        val ifBody = parseStatementBlockUntilEndToken(it)
        if (!it.isAtom("else"))
            return IfStatement(condition, ifBody, ArrayList())
        it.next()
        if (it.isAtom("if")) {
            it.next()
            return IfStatement(condition, ifBody, arrayListOf(parseIfStatement(it)))
        }
        it.forceAtom("begin", "expected begin")
        return IfStatement(condition, ifBody, parseStatementBlockUntilEndToken(it))
    }

    private fun parseCaseStatement(it: TokenIterator): ParsedStatement {
        val caseExpr = parseExpressionUntilEnd(it.forceGroup(GroupType.Bracket,"expected case expression"),
            DataType.TypeUInt32
        )
        val cases = ArrayList<CaseStatement.Item>()
        while (!it.isAtom("endcase")) {
            val caseValues = ArrayList<ConstantValueExpression>()
            while(true) {
                caseValues.add(parseExpression(it, DataType.TypeUInt32) as? ConstantValueExpression ?: throw ParserException(it.pos(),"expected constant value in case"))
                if (it.isOperator(",")) {
                    it.next()
                    continue
                }
                it.forceOperator(":","expected ',' or ':' in case statement")
                break
            }
            it.forceAtom("begin","expected begin after case values")
            val body = parseStatementBlockUntilEndToken(it)
            cases.add(CaseStatement.Item(caseValues, body))
        }
        it.forceAtom("endcase","expected endcase")
        return CaseStatement(caseExpr, cases)
    }

    private fun parseStatementBlockUntilEndToken(it: TokenIterator):ArrayList<ParsedStatement> = parseStatementBlock(it) { tk -> tk is AtomToken && tk.name == "end"}

    private fun parseSubModuleCall(it: TokenIterator, declareToken:CExpToken):ParsedStatement {
        if (declareToken.parameters.size != 3)
            throw ParserException(declareToken.pos, "invalid call-module params")
        val subModuleType = declareToken.parameters[0]
        val subModuleName = declareToken.parameters[1]
        val subModuleEvaluationStep = declareToken.parameters[2].toInt()
        fun findSubModule():Int {
            for (i in subModuleList.indices)
                if (subModuleList[i].name == subModuleName)
                    return i
            throw ParserException(declareToken.pos, "no submodule '$subModuleName' found or not accessible here")
        }
        val subModuleIdx = findSubModule()
        val subModule = subModuleList[subModuleIdx]
        if (subModuleEvaluationStep<0 || subModuleEvaluationStep>=subModule.type.moduleFunctions.size)
            throw ParserException(declareToken.pos, "evaluation step out of range")
        val subModuleFun = subModule.type.moduleFunctions[subModuleEvaluationStep]
        fun findInputOfSubModule(pos:SourcePosition, portName:String):Int {
            for (i in subModuleFun.input.indices) {
                if (subModuleFun.input[i].name == portName)
                    return i
            }
            throw ParserException(pos, "no port '$portName' in module '$subModuleName' found")
        }
        val inputExpressions = Array<ParsedExpression?>(subModuleFun.input.size) { null }
        while (true) {
            if (it.atEnd())
                throw ParserException(it.pos(), "unexpected end of statement block")
            val tk = it.token()
            it.next()
            if (tk is CExpToken && tk.keyword == "call-module") {
                if (tk.parameters.size != 4)
                    throw ParserException(tk.pos, "invalid call-module params")
                if (tk.parameters[0] != subModuleType || tk.parameters[1] != subModuleName || tk.parameters[2].toInt() != subModuleEvaluationStep)
                    throw ParserException(tk.pos,"call to declare submodule mismatch")
                //val clockFwd = tk.parameters[3] == "clockfwd" //wird nicht mehr benötigt
                val inputs = ArrayList<ParsedExpression>()
                for (i in subModuleFun.input.indices) {
                    val expr = inputExpressions[i]
                    if (expr == null) {
                        val portWithSameName = specifiers[subModuleFun.input[i].name]
                        if (portWithSameName == null || portWithSameName.dataType != subModuleFun.input[i].dataType || portWithSameName.usage != ParsedSpecifier.UsageType.InputPort)
                            throw ParserException(tk.pos, "no assignment for port ${subModuleFun.input[i].name}")
                        inputs.add(SpecifierExpression(portWithSameName.ref,portWithSameName.dataType))
                        continue
                    }
                    inputs.add(expr)
                }

                return SubModuleCallStatement(subModuleIdx,subModuleEvaluationStep,inputs)
            }
            if (tk !is AtomToken)
                throw ParserException(tk.pos, "expected sub module parameter assignment, got $tk")

            val extractedPort = extractPortName(subModuleName,tk.name) ?: throw ParserException(tk.pos, "'${tk.name}' is not a port name of sub module '$subModuleName'")
            val portIdx = findInputOfSubModule(tk.pos, extractedPort)
            if (inputExpressions[portIdx] != null)
                throw ParserException(tk.pos,"duplicate port assignment")
            it.forceOperator("=","expected '=' in assignment")
            val expr = parseExpression(it, subModuleFun.input[portIdx].dataType)
            it.forceOperator(";", "expected ';' after assignment, got ${it.token()}")
            inputExpressions[portIdx] = expr
        }
    }

    private fun parseStatementBlock(it: TokenIterator, stopCallback:(tk: Token)->Boolean):ArrayList<ParsedStatement> {
        val stmList = ArrayList<ParsedStatement>()
        while (true) {
            if (it.atEnd())
                throw ParserException(it.pos(), "unexpected end of statement block")
            if (stopCallback(it.token())) {
                it.next()
                return stmList
            }
            val tk = it.token()
            it.next()
            if (tk is CExpToken && tk.keyword == "call-module")
                throw ParserException(tk.pos,"unexpected call module statement")
            if (tk is CExpToken && tk.keyword == "declare-module-input") {
                stmList.add(parseSubModuleCall(it,tk))
                continue
            }
            if (tk !is AtomToken)
                throw ParserException(tk.pos, "unexpected expression $tk")
            if (tk.name == "else" || tk.name == "end")
                throw ParserException(tk.pos, "misplaced '${tk.name}'")
            if (tk.name == "if") {
                stmList.add(parseIfStatement(it))
                continue
            }
            if (tk.name == "case") {
                stmList.add(parseCaseStatement(it))
                continue
            }
            val destSpec = forceSpecifier(tk.pos, tk.name)
            if (!destSpec.writeAllowed())
                throw ParserException(
                    tk.pos,
                    "variable '${destSpec.verilogName}' is not writable, usage was '${destSpec.usage}', submodule called ${destSpec.subModuleCalled}"
                )
            it.forceOperator("=","expected '=' in assignment")
            val expr = parseExpression(it, destSpec.dataType)
            if (it.isCExp("set-valid-value")) {
                it.next()
                if (destSpec.usage != ParsedSpecifier.UsageType.OutputPort)
                    throw ParserException(
                        it.pos(),
                        "variable '${destSpec.verilogName}' cannot be marked as valid, usage mode was '${destSpec.usage}'"
                    )
                destSpec.validValue = true
            }
            it.forceOperator(";", "expected ';' after assignment, got ${it.token()}")
            stmList.add(AssignmentStatement(destSpec.ref, expr))
        }
    }

    fun parseBody(it: TokenIterator, stopCallback:(tk: Token)->Boolean) {
        functionBody.addAll(parseStatementBlock(it,stopCallback))
    }
}