package verilogparser

import java.lang.Exception
import java.lang.StringBuilder
import java.math.BigInteger
import java.util.regex.Pattern

class SourcePosition(val fileName:String, val idx:Int, val line:Int, val col:Int) {
    override fun toString() = "$fileName:$line:$col"
}
class ParserException:Exception {
    constructor(pos1: SourcePosition, msg1: String):super("$pos1: $msg1")
    constructor(pos1: SourcePosition, msg1: String, pos2: SourcePosition, msg2: String):super("$pos1: $msg1, previous: $pos2: $msg2")
    constructor(token1: Token, msg1:String):super("${token1.pos}, $token1, $msg1")
}

sealed class Token(val pos: SourcePosition)
class AtomToken(pos: SourcePosition, val name:String): Token(pos) {
    override fun toString() = "verilogparser.AtomToken($pos $name)"
}
enum class GroupType {Root,Bracket,Curly,Index,InlineAttribute}
enum class DigitType {Decimal,Hex,Binary}
class GroupToken(pos: SourcePosition, val type: GroupType, val elements:ArrayList<Token> = ArrayList()): Token(pos) {
    override fun toString() = "verilogparser.GroupToken($pos $type ${elements.size})"
}
class NumberToken(pos: SourcePosition, val value:BigInteger, val binaryDigits:Int, val digitType: DigitType): Token(pos) {
    override fun toString() = "verilogparser.NumberToken($pos "+(if (binaryDigits<0) "" else "${binaryDigits}'")+when(digitType) {
        DigitType.Decimal -> value.toString(10)
        DigitType.Hex -> "h"+value.toString(16)
        DigitType.Binary -> "b"+value.toString(2)
    }+")"
}
class OperatorToken(pos: SourcePosition, val op:String): Token(pos) {
    override fun toString() = "verilogparser.OperatorToken($pos $op)"
}
class StringToken(pos: SourcePosition, val str:String): Token(pos) {
    override fun toString() = "verilogparser.StringToken($pos '$str')"
}
class CExpToken(pos: SourcePosition, val keyword: String, val parameters:List<String>): Token(pos) {
    override fun toString() = "verilogparser.CExpToken($pos '$keyword': '$parameters')"
}

class TokenIterator(val grp: GroupToken) {
    private var idx=0
    fun next(): Token {
        val prev = token()
        idx++
        return prev
    }
    fun getCurrentIndex() = idx
    fun setCurrentIndex(newIdx:Int) {
        idx = newIdx
    }
    fun pos(): SourcePosition = if (idx<grp.elements.size) grp.elements[idx].pos else if (grp.elements.isEmpty()) grp.pos else grp.elements.last().pos
    fun atEnd() = idx>=grp.elements.size
    fun token() = if (idx<grp.elements.size) grp.elements[idx] else throw ParserException(pos(),"end of token list reached")
    fun isAtom(keyword:String):Boolean {
        val t = token()
        return t is AtomToken && t.name == keyword
    }
    fun isCExp(keyword:String):Boolean {
        val t = token()
        return t is CExpToken && t.keyword == keyword
    }
    fun isOperator(op:String):Boolean {
        val t = token()
        return t is OperatorToken && t.op == op
    }
    fun forceAtom(errMsg:String): AtomToken {
        val t = next()
        return t as? AtomToken ?: throw ParserException(t,errMsg)
    }
    fun forceCExp(errMsg:String): CExpToken {
        val t = next()
        return t as? CExpToken ?: throw ParserException(t, errMsg)
    }
    fun forceCExp(keyword:String, errMsg:String): CExpToken {
        val t = next()
        val cexp =  t as? CExpToken ?: throw ParserException(t, errMsg)
        if (cexp.keyword != keyword)
            throw ParserException(t, errMsg)
        return cexp
    }
    fun forceAtom(atom:String, errMsg:String) {
        val t = next()
        val at =  t as? AtomToken ?: throw ParserException(t,errMsg)
        if (at.name != atom)
            throw ParserException(t,errMsg)
    }
    fun forceGroup(type: GroupType, errMsg:String): TokenIterator {
        val t = next()
        return TokenIterator(if (t is GroupToken && t.type == type) t else throw ParserException(t,errMsg))
    }
    fun forceNumber(errMsg:String): NumberToken {
        val t = next()
        return t as? NumberToken ?: throw ParserException(t,errMsg)
    }
    fun isGroup(type: GroupType):Boolean {
        val t = token()
        return t is GroupToken && t.type == type
    }
    fun forceOperator(op:String, errMsg:String): OperatorToken {
        val t = next()
        return if (t is OperatorToken && t.op == op) t else throw ParserException(t,errMsg)
    }
    fun commaSplitted(callback:(subIterator: TokenIterator)->Unit) {
        var subGrp = GroupToken(pos(),grp.type)
        while (!atEnd()) {
            val t = next()
            if (t is OperatorToken && t.op == ",") {
                callback(TokenIterator(subGrp))
                subGrp = GroupToken(pos(),grp.type)
            }
            else
                subGrp.elements.add(t)
        }
        callback(TokenIterator(subGrp))
    }
}

fun fileLexer(fileName:String, sourceCode:String): TokenIterator {
    val stopChar = '\u0000'
    val twoCharOperators = arrayListOf("||","&&","==","!=","<=",">=","<<",">>","~&","&~","|~","~|","^~","~^")
    val singleCharOperators = "+-*/%!~<>&|^=?:,;@."
    val groupStack = ArrayList<GroupToken>()

    var idx=0
    var line=1
    var col=1

    fun pos() = SourcePosition(fileName,idx,line,col)
    fun isSpecifierStartChar(chr:Char) = chr in 'A'..'Z' || chr in 'a'..'z' || chr == '_'

    fun isDigit(chr:Char, type: DigitType) = when(type) {
        DigitType.Decimal -> chr in '0'..'9'
        DigitType.Hex -> chr in '0'..'9' || chr in 'A'..'F' || chr in 'a'..'f'
        DigitType.Binary -> chr in '0'..'1'
    }

    fun isSpecifierChar(chr:Char) = isSpecifierStartChar(chr) || isDigit(chr, DigitType.Decimal)

    fun nextChar(amount:Int=1) {
        for (k in 0 until amount) {
            if (idx>=sourceCode.length)
                return
            val c = sourceCode[idx]
            idx++
            if (c == '\n') {
                line++
                col = 1
            } else
                col++
        }
    }
    fun getTopOfStack(): GroupToken {
        if (groupStack.isEmpty())
            throw ParserException(pos(),"group stack was empty, nexting error?")
        return groupStack.last()
    }
    fun addToken(token: Token) {
        getTopOfStack().elements.add(token)
    }
    fun startGroup(type: GroupType, charCount:Int=1) {
        val grp = GroupToken(pos(),type)
        addToken(grp)
        nextChar(charCount)
        groupStack.add(grp)
    }
    fun endGroup(type: GroupType, charCount:Int=1): GroupToken {
        val grp = getTopOfStack()
        if (grp.type != type) {
            throw ParserException(pos(),"invalid group nesting type $type",grp.pos,"opening type was: ${grp.type}")
        }
        nextChar(charCount)
        groupStack.removeAt(groupStack.size-1)
        return grp
    }
    fun skipSingleLineComment() {
        while(idx<sourceCode.length && sourceCode[idx] != '\n')
            nextChar()
    }
    fun skipMultilineComment() {
        val startPos = pos()
        nextChar(2)
        while(idx<sourceCode.length && sourceCode[idx] == ' ')
            nextChar()
        val isCustomToken = if (idx<sourceCode.length && sourceCode[idx] == '@') idx else -1
        while(idx+1<sourceCode.length) {
            if (sourceCode[idx] == '*' && sourceCode[idx+1] == '/') {
                nextChar(2)
                if (isCustomToken>=0) {
                    val tk = sourceCode.substring(isCustomToken, idx - 2).trim()
                    if (!tk.startsWith("@cexp:"))
                        throw ParserException(startPos, "invalid comment token $tk")
                    val params = tk.split(":")
                    addToken(CExpToken(startPos,params[1],params.subList(2,params.size)))
                }
                return
            }
            nextChar()
        }
        throw ParserException(startPos,"unexpected end of file, end of multiline comment expected")
    }
    fun skipDirective() {
        nextChar()
        val startPos = pos()
        while(idx<sourceCode.length && sourceCode[idx] != '\n' && sourceCode[idx] != '\r')
            nextChar()
        val directive = sourceCode.substring(startPos.idx,idx)
        if (!Pattern.compile("include\\s*\"[^\"]+\"\\s*(//.*)?").matcher(directive).matches())
            throw ParserException(startPos,"unknown directive '$directive'")
    }
    fun parseAtom(): AtomToken {
        val startPos = pos()
        while (idx<sourceCode.length && isSpecifierChar(sourceCode[idx]))
            nextChar()
        return AtomToken(startPos,sourceCode.substring(startPos.idx,idx))
    }
    fun parseDigitString(type: DigitType, underscoreAllowed:Boolean):String {
        val result = StringBuilder()
        while (idx<sourceCode.length) {
            if (isDigit(sourceCode[idx],type)) {
                result.append(sourceCode[idx])
                nextChar()
            }
            else if (underscoreAllowed && sourceCode[idx] == '_') //_ ist ein optisches trennzeichen und wird entfernt
                nextChar()
            else
                break
        }
        if (idx<sourceCode.length && isDigit(sourceCode[idx], DigitType.Hex))
            throw ParserException(pos(),"illegal digit '${sourceCode[idx]}'")
        return result.toString()
    }
    fun parseNumber(): NumberToken {
        val startPos = pos()
        val numberString = parseDigitString(DigitType.Decimal,false)
        if (sourceCode[idx] != '\'')
            return NumberToken(startPos, BigInteger(numberString),-1, DigitType.Decimal)
        val digitCount = numberString.toInt()
        nextChar()
        if (sourceCode[idx] == 'b') {
            nextChar()
            return NumberToken(startPos, BigInteger(parseDigitString(DigitType.Binary,true), 2), digitCount, DigitType.Binary)
        }
        if (sourceCode[idx] == 'h') {
            nextChar()
            return NumberToken(startPos, BigInteger(parseDigitString(DigitType.Hex,true), 16), digitCount, DigitType.Hex)
        }
        throw ParserException(pos(),"expected 'b' or 'h' as number type")
    }
    fun parseOperator(chars:Int): OperatorToken {
        val startPos=pos()
        val op = sourceCode.substring(idx,idx+chars)
        nextChar(chars)
        return OperatorToken(startPos,op)
    }
    fun parseString(): StringToken {
        val startPos = pos()
        nextChar()
        while(idx<sourceCode.length || sourceCode[idx] == '\n') {
            if (sourceCode[idx] == '"') {
                nextChar()
                return StringToken(startPos,sourceCode.substring(startPos.idx+1,pos().idx-1))
            }
            if (sourceCode[idx] == '\\')
                throw ParserException(pos(),"'\\' in string not yet supported")
            nextChar()
        }
        throw ParserException(pos(),"expected end of string")
    }

    groupStack.add(GroupToken(pos(), GroupType.Root))
    while (idx<sourceCode.length) {
        val c = sourceCode[idx]
        val nc = if (idx+1<sourceCode.length) sourceCode[idx+1] else stopChar

        if (c == ' ' || c == '\t' || c=='\r' || c == '\n')
            nextChar()
        else if (c=='/' && nc == '/')
            skipSingleLineComment()
        else if (c=='/' && nc == '*')
            skipMultilineComment()
        else if (c=='`')
            skipDirective()
        else if (isSpecifierStartChar(c))
            addToken(parseAtom())
        else if (c == '(' && nc == '*') {
            if (idx+2<sourceCode.length && sourceCode[idx+2] == ')') {
                addToken(OperatorToken(pos(),"(*)"))
                nextChar(3)
            }
            else
                startGroup(GroupType.InlineAttribute, 2)
        }
        else if (c == '(')
            startGroup(GroupType.Bracket)
        else if (c == '{')
            startGroup(GroupType.Curly)
        else if (c == '[')
            startGroup(GroupType.Index)
        else if (c == ')')
            endGroup(GroupType.Bracket)
        else if (c == '}')
            endGroup(GroupType.Curly)
        else if (c == ']')
            endGroup(GroupType.Index)
        else if (c == '*' && nc == ')')
            endGroup(GroupType.InlineAttribute,2)
        else if (isDigit(c, DigitType.Decimal))
            addToken(parseNumber())
        else if (twoCharOperators.contains(sourceCode.substring(idx,idx+2)))
            addToken(parseOperator(2))
        else if (singleCharOperators.contains(c))
            addToken(parseOperator(1))
        else if (c == '"')
            addToken(parseString())
        else
            throw ParserException(pos(),"invalid char '$c'")
    }
    return TokenIterator(endGroup(GroupType.Root))
}
