package verilogparser

import DataType
import ParsedModule
import ParsedModuleFunction
import ParsedPureFunction
import Specifier
import SpecifierReference
import SubModuleInstance
import builtInFunctions
import java.lang.NumberFormatException
import java.lang.StringBuilder
import java.math.BigInteger

class ParsedSpecifier(val verilogName:String, val usage:UsageType, val dataType:DataType, val ref:SpecifierReference) {
    enum class UsageType {InputPort, OutputPort, FromSubModule, RegisterRO, RegisterRW, LocalVariable}
    var validValue = false //bei true kann auf den wert zugegrifen werden, obwohl es ein output ist
    var subModuleCalled = false //bei true dürfen kein werte mehr zugewiesen werden
    fun readAllowed():Boolean = when(usage) {
        UsageType.FromSubModule,UsageType.InputPort,UsageType.RegisterRO,UsageType.RegisterRW,UsageType.LocalVariable -> true
        UsageType.OutputPort -> validValue
    }
    fun writeAllowed():Boolean = when(usage) {
        UsageType.FromSubModule -> !subModuleCalled
        UsageType.OutputPort, UsageType.RegisterRW,UsageType.LocalVariable -> true
        UsageType.InputPort, UsageType.RegisterRO -> false
    }
    override fun toString():String = verilogName
}

fun extractPortName(subModuleName:String, fullPortName:String):String? {
    val prefix = "submod_${subModuleName}_"
    if (!fullPortName.startsWith(prefix))
        return null
    if (fullPortName.endsWith("_in"))
        return fullPortName.substring(prefix.length,fullPortName.length-3)
    if (fullPortName.endsWith("_out"))
        return fullPortName.substring(prefix.length,fullPortName.length-4)
    return null
}

class UnitParser(rootIterator: TokenIterator, allModules:List<ParsedModule>) {
    private val evaluationSteps = ArrayList<ParsedFunction>()
    private val specifiersThatRequireInit = ArrayList<String>()
    private val usedSubModules = ArrayList<SubModuleInstance>()
    private val registers = ArrayList<Specifier>()
    private var codeSectionNo = 0
    private val pureFunctionMap = builtInFunctions()
    val parsedResult:ParsedModule

    init {
        val generalSettings = rootIterator.forceCExp("settings", "expected settings cexp before module")
        if (generalSettings.parameters.size != 1)
            throw ParserException(
                generalSettings.pos,
                "cexp settings, 1 parameters expected: number of evaluation steps"
            )
        for (i in 0 until generalSettings.parameters[0].toInt())
            evaluationSteps.add(ParsedFunction(pureFunctionMap,i, usedSubModules))
        if (rootIterator.forceAtom("expected module").name != "module")
            throw ParserException(rootIterator.pos(), "epxected module")
        val moduleName = rootIterator.forceAtom("expected module name")
        parseModulePorts(rootIterator.forceGroup(GroupType.Bracket, "expected ports ()"), allModules)
        rootIterator.forceOperator(";", "; expected at end of port list")


        val declaredPureFunctions = ArrayList<ParsedPureFunction>()
        while (!rootIterator.atEnd() && !rootIterator.isAtom("endmodule")) {
            if (rootIterator.isCExp("reg-dummy") || rootIterator.isCExp("reg-state") || rootIterator.isCExp("reg-local") || rootIterator.isCExp("port-input"))
                parseRegDeclaration(rootIterator)
            else if (rootIterator.isAtom("localparam"))
                parseLocalParam(rootIterator)
            else if (rootIterator.isAtom("function"))
                declaredPureFunctions.add(parseFunction(rootIterator))
            else if (rootIterator.isAtom("always"))
                parseAlwaysBlock(rootIterator)
            else
                throw ParserException(rootIterator.pos(), "unexpected token ${rootIterator.token()}")
        }

        val modFunctions = ArrayList<ParsedModuleFunction>()
        for (st in evaluationSteps) {
            modFunctions.add(ParsedModuleFunction(
                st.inputs,
                st.outputs,
                st.locals,
                st.functionBody))
        }
        parsedResult = ParsedModule(moduleName.name, false, registers, modFunctions, declaredPureFunctions, usedSubModules)
    }

    private fun parseBitSizeDeclaration(it: TokenIterator):Int {
        if (!it.isGroup(GroupType.Index))
            return 0
        val sizeGrp = it.forceGroup(GroupType.Index, "internal error")
        val msb = sizeGrp.forceNumber("expected size number")
        sizeGrp.forceOperator(":", "expected size separator")
        val lsb = sizeGrp.forceNumber("expected size number")
        if (!sizeGrp.atEnd())
            throw ParserException(sizeGrp.pos(), "expected end of size grp")
        if (lsb.value != BigInteger.ZERO)
            throw ParserException(lsb.pos, "lsb must be 0")
        val msbInt = msb.value.toInt()
        if (msbInt<0)
            throw ParserException(lsb.pos, "msb must be >=0")
        return msbInt+1
    }

    private fun bitSizeToDataType(pos: SourcePosition, bitSize:Int):DataType = when (bitSize) {
        0 -> DataType.TypeBoolean
        32 -> DataType.TypeUInt32
        64 -> DataType.TypeUInt64
        else -> throw ParserException(pos,"unsupported bit width $bitSize")
    }

    private fun parseFunction(rootIterator: TokenIterator):ParsedPureFunction {
        rootIterator.next()
        val returnType = bitSizeToDataType(rootIterator.pos(), parseBitSizeDeclaration(rootIterator))
        val funcName = rootIterator.forceAtom("expected function name")
        rootIterator.forceOperator(";","expected ';' after function name")
        val localFun = ParsedFunction(pureFunctionMap,-1,ArrayList())
        //ports und lokale variablen parsen
        val inputs = ArrayList<Specifier>()
        val locals = ArrayList<Specifier>()
        while (true) {
            val tk = rootIterator.forceAtom("expected input, reg or begin")
            if (tk.name == "begin")
                break
            if (tk.name != "input" && tk.name != "reg")
                throw ParserException(tk.pos, "expected input, reg or begin")
            val varType = bitSizeToDataType(rootIterator.pos(), parseBitSizeDeclaration(rootIterator))
            val varName = rootIterator.forceAtom("expected var name")
            rootIterator.forceOperator(";","expected';'")
            if (tk.name == "input") {
                localFun.addInput(tk.pos, varName.name, varType)
                inputs.add(Specifier(varName.name, varType,-1))
            }
            else {
                localFun.addLocal(tk.pos, varName.name, varType)
                locals.add(Specifier(varName.name, varType,-1))
            }
        }

        localFun.addOutput(funcName.pos,funcName.name,returnType)
        localFun.parseBody(rootIterator) { tk -> tk is AtomToken && tk.name == "end" }
        rootIterator.forceAtom("endfunction","expected endfunction")
        val newPureFunc = ParsedPureFunction(inputs, locals, returnType, funcName.name, localFun.functionBody)
        pureFunctionMap[funcName.name] = newPureFunc
        return newPureFunc
    }

    private fun parseAlwaysBlock(rootIterator: TokenIterator) {
        val initVars = HashSet<String>()
        rootIterator.next()
        rootIterator.forceOperator("@","expected @")
        rootIterator.forceOperator("(*)","expected (*)")
        rootIterator.forceAtom("begin", "expected begin")
        while (!rootIterator.atEnd() && !rootIterator.isCExp("always-block-end")) {
            if (rootIterator.isCExp("dummyinit-start")) {
                rootIterator.next()
                while (!rootIterator.atEnd() && !rootIterator.isCExp("dummyinit-stop")) {
                    val destVar = rootIterator.forceAtom("expected specifier in dummyinit").name
                    rootIterator.forceOperator("=","expected = in dummyinit")
                    val num = rootIterator.forceNumber("expected number 0 in dummyinit")
                    if (num.value != BigInteger.ZERO)
                        throw ParserException(num.pos, "number must be 0 in dummyinit")
                    rootIterator.forceOperator(";","expected ; in dummyinit")
                    if (!specifiersThatRequireInit.contains(destVar))
                        throw ParserException(num.pos, "init var '$destVar' not found")
                    initVars.add(destVar)
                }
            }
            else if (rootIterator.isCExp("code-start")) {
                val cexp = rootIterator.forceCExp("internal error")
                if (codeSectionNo>=evaluationSteps.size)
                    throw ParserException(cexp.pos,"number of code-start mismatch settings token $codeSectionNo vs ${evaluationSteps.size}")
                if (cexp.parameters.size != 1)
                    throw ParserException(cexp.pos,"clock struct name expected in code-start")
                val clockStructName = cexp.parameters[0]
                val currentFunction = evaluationSteps[codeSectionNo]
                currentFunction.parseBody(rootIterator) { tk -> tk is CExpToken && tk.keyword == "code-stop" }
                codeSectionNo+=1
            }
            else {
                rootIterator.next()
            }
        }
        for (spec in specifiersThatRequireInit) {
            if (!initVars.contains(spec))
                throw ParserException(rootIterator.pos(), "var '$spec' was not initialized in dummyinit-start section")
        }
        rootIterator.forceCExp("always-block-end", "expected 'always-block-end' before end")
        rootIterator.forceAtom("end","expected end")
        if (codeSectionNo!=evaluationSteps.size)
            throw ParserException(rootIterator.pos(),"number of code-start mismatch settings token $codeSectionNo vs ${evaluationSteps.size}")
    }

    private fun parseLocalParam(rootIterator: TokenIterator) {
        rootIterator.next()
        val bitSize = parseBitSizeDeclaration(rootIterator)
        if (bitSize != 32)
            throw ParserException(rootIterator.pos(), "localparam must be 32 bits wide")
        val name = rootIterator.forceAtom("expected localparam name").name
        rootIterator.forceOperator("=","expected '='")
        val num = rootIterator.token()
        rootIterator.next()
        if (num !is NumberToken)
            throw ParserException(num.pos, "expected number")
        if (num.value.compareTo(BigInteger.ZERO)<0 || num.value.compareTo(BigInteger.ONE.shiftLeft(32))>=0)
            throw ParserException(num.pos, "value ${num.value} out of uint32 range")
        rootIterator.forceOperator(";","expected ';'")
        //konstanten in allen evaluation steps hinzufügen
        for (step in evaluationSteps)
            step.addConstant(num.pos,name,num.value,DataType.TypeUInt32)
    }

    private fun parseRegDeclaration(rootIterator: TokenIterator) {
        val pos = rootIterator.pos()
        val cexpTk = rootIterator.forceCExp("internal error")
        rootIterator.forceAtom("reg", "expected reg")
        val bitSize = parseBitSizeDeclaration(rootIterator)
        val regName = rootIterator.forceAtom("expected reg name").name
        rootIterator.forceOperator(";", "expected ';' after port declaration")
        val dataType = bitSizeToDataType(pos, bitSize)
        if (cexpTk.keyword == "port-input") {
            iterateAffectedEvaluationSteps(cexpTk) {
                step,stepNo -> step.addInput(pos,regName,dataType)
            }
            return
        }
        specifiersThatRequireInit.add(regName)
        if (cexpTk.keyword == "reg-dummy") //wird nicht weiter verwendet
            return
        if (cexpTk.keyword == "reg-local") {
            if (getAffectedEvaluationStepNumbers(cexpTk).size>1)
                throw ParserException(cexpTk,"local variables are only allowed in one execution step")
            iterateAffectedEvaluationSteps(cexpTk) {
                step,stepNo -> step.addLocal(pos, regName, dataType)
            }
            return
        }
        val registerIndex = registers.size
        registers.add(Specifier(regName,dataType,-1))
        if (cexpTk.parameters.size != 2)
            throw ParserException(cexpTk.pos, "register cexp requires two parameters: step read list, step write no")
        val writeRegisterAtStep = cexpTk.parameters[1].toInt()
        iterateAffectedEvaluationSteps(cexpTk) {
                step,stepNo -> step.addRegister(pos,regName,dataType,stepNo == writeRegisterAtStep, registerIndex)
        }
    }

    private fun instanceByName(pos:SourcePosition, name:String):Int {
        for (i in 0 until usedSubModules.size)
            if (usedSubModules[i].name == name)
                return i
        throw ParserException(pos, "no instance '$name' found")
    }

    private fun parseModulePorts(it: TokenIterator, allModules: List<ParsedModule>) {
        while (!it.atEnd()) {
            val pos = it.pos()
            val cexpTk = it.forceCExp("expected @cexp port notation")
            //der erste parameter der cexp ist eine kommagetrennte liste, in welchen evaluationSteps die variable gültig ist

            if (cexpTk.keyword == "submodule") {
                if (cexpTk.parameters.size != 2)
                    throw ParserException(pos, "illegal cexp:submodule")
                val modType = cexpTk.parameters[0]
                val modName = cexpTk.parameters[1]
                try {
                    usedSubModules.add(SubModuleInstance(modName,allModules.first { it -> it.name == "cloned_$modType" }))
                }
                catch(e:NoSuchElementException) {
                    throw ParserException(pos, "no module '$modType' found, ${allModules.map { it.name }}")
                }
                continue
            }
            val portIsInput = if (it.isAtom("input")) true else if (it.isAtom("output")) false else throw ParserException(
                it.pos(),
                "expected port direction"
            )
            it.next()
            if (!portIsInput)
                it.forceAtom("reg","reg")
            val bitSize = parseBitSizeDeclaration(it)

            val portName = it.forceAtom("expected port name").name
            if (!it.atEnd())
                it.forceOperator(",", "expected ',' or end of ports")
            if (cexpTk.keyword == "port-ignore")
                continue
            if (cexpTk.keyword == "port-state") {
                //outputs verifizieren, das outputs 32/64/bool bit sind
                if (!portIsInput)
                    specifiersThatRequireInit.add(portName)
                continue
            }
            if (cexpTk.keyword == "port-clock") {
                if (bitSize != 0)
                    throw ParserException(pos, "clock port must be of type bool")
                iterateAffectedEvaluationSteps(cexpTk) { step,_ -> step.addInput(pos,portName,DataType.TypeBoolean) }
                continue
            }
            if (cexpTk.keyword == "port-submodule-dummy") {
                specifiersThatRequireInit.add(portName)
                continue
            }
            if (cexpTk.keyword != "port" && cexpTk.keyword != "port-submodule")
                throw ParserException(pos, "unknown port usage type $cexpTk")

            if (getAffectedEvaluationStepNumbers(cexpTk).size>1)
                throw ParserException(cexpTk,"ports are only allowed in one execution step")

            if (!portIsInput)
                specifiersThatRequireInit.add(portName)
            val portDataType = bitSizeToDataType(pos, bitSize)
            if (cexpTk.keyword == "port-submodule") {
                //1. parameter gibt an, in welchem substep dieser port benutzt wird
                //2. parameter gibt an, zu welchem substep der port des submoduls gehört
                if (cexpTk.parameters.size != 3)
                    throw ParserException(cexpTk.pos,"port-submodule expect 2 parameters")
                val subModuleName = cexpTk.parameters[1]
                val subModulePart = cexpTk.parameters[2].toInt()
                val subModuleIndex = instanceByName(pos, subModuleName)
                val subModule = usedSubModules[subModuleIndex].type
                val prt = extractPortName(subModuleName, portName) ?: throw ParserException(pos, "no matching submodule found for var $portName")

                if (subModulePart<0 || subModulePart>subModule.moduleFunctions.size)
                    throw ParserException(pos, "no evalute function no $subModulePart in $subModuleName")
                val subModuleFun = subModule.moduleFunctions[subModulePart]
                if (portIsInput) {
                    val portIndex = subModuleFun.output.indexOfFirst { it.name == prt }
                    if (portIndex<0)
                        throw ParserException(pos, "submodule '$subModuleName' has no output '$prt'")
                    iterateAffectedEvaluationSteps(cexpTk) { step,_ -> step.addFromSubModule(pos, portName, portDataType, subModuleIndex, subModulePart, portIndex) }
                }
                //Parameter in das Submodul werden speziell geparsed


                continue
            }
            //normaler port
            iterateAffectedEvaluationSteps(cexpTk) { step,_ ->
                if (portIsInput)
                    step.addInput(pos, portName, portDataType)
                else
                    step.addOutput(pos, portName, portDataType)
            }
        }
    }

    private fun getAffectedEvaluationStepNumbers(cexpTk: CExpToken):List<Int> {
        if (cexpTk.parameters.isEmpty())
            throw ParserException(cexpTk.pos,"first parameter of @cexp port/submodule declaration must be a comma separated evaluation list")
        return cexpTk.parameters.first().split(",").map {
            try {
                it.toInt()
            } catch(e: NumberFormatException) {
                throw ParserException(cexpTk.pos,"illegal value $it")
            }
        }
    }

    private fun iterateAffectedEvaluationSteps(cexpTk: CExpToken, cb:(ParsedFunction, stepNo:Int)->Unit) {
        for (u in getAffectedEvaluationStepNumbers(cexpTk)) {
            if (u < 0 || u >= evaluationSteps.size)
                throw ParserException(cexpTk.pos, "invalid evaluation step $u")
            cb(evaluationSteps[u],u)
        }
    }
}
