package assembler

import ilang.*
import java.lang.Exception

sealed class AsmOperand {
    abstract fun isMemory():Boolean
    abstract fun memoryIs64Bit():Boolean
}
class AsmConstantOperand(val value:ULong):AsmOperand() {
    override fun toString() = "0x${value.toString(16)}"
    override fun isMemory():Boolean = false
    override fun memoryIs64Bit():Boolean = false
}
class AsmLabelOperand(val lbl:String):AsmOperand() {
    override fun toString() = lbl
    override fun isMemory():Boolean = false
    override fun memoryIs64Bit():Boolean = false
}
class AsmLocalOperand(val name:String, val belongsToVar:ILangVariable?):AsmOperand() {
    override fun toString() = name
    override fun isMemory():Boolean = false
    override fun memoryIs64Bit():Boolean = false
}
class AsmMemOperand(val variable:ILangVariable):AsmOperand() {
    init {
        if (variable.direction != ILangVariableDirection.Register)
            throw Exception("variable must be of type register")
    }
    override fun toString() = "[${variable.name}]"
    override fun isMemory():Boolean = true
    override fun memoryIs64Bit():Boolean = variable.dataType == ILangDataType.Int64
}
class AsmArrayOperand(val arrayVariable: ILangArrayVariable, val idx:AsmLocalOperand, val idxMultiplicator:Int, val byteOffset:Int):AsmOperand() {
    override fun toString() = "[${arrayVariable.name} + $idx*$idxMultiplicator + $byteOffset]"
    override fun isMemory():Boolean = true
    override fun memoryIs64Bit():Boolean = false
}
class AsmRamHubBaseOperand(val offset:String, val idx:AsmLocalOperand?, val idxMultiplicator:Int, val dataType: ILangDataType):AsmOperand() {
    override fun toString() = "[hubBase + $offset]"
    override fun isMemory():Boolean = true
    override fun memoryIs64Bit():Boolean = false
}
fun setOfAsmLocalOperandDirectOrArraySideEffect(op:AsmOperand?):Set<AsmLocalOperand> = if (op is AsmLocalOperand) setOf(op) else if (op is AsmArrayOperand) setOf(op.idx) else setOf()
fun setOfAsmLocalOperandDirect(op:AsmOperand?):Set<AsmLocalOperand> = if (op is AsmLocalOperand) setOf(op) else setOf()
fun setOfAsmLocalOperandArraySideEffect(op:AsmOperand?):Set<AsmLocalOperand> = if (op is AsmArrayOperand) setOf(op.idx) else setOf()

sealed class AsmAbstractLookupTable(val label:AsmLabelOperand, val tableAccessSize:CpuAccessSize)
class AsmNumericLookupTable(val values:List<ULong>, label:AsmLabelOperand, tableAccessSize: CpuAccessSize):AsmAbstractLookupTable(label,tableAccessSize)
class AsmJumpLookupTable(val values:List<JumpDestinationLookupExpression.Entry>, label:AsmLabelOperand):AsmAbstractLookupTable(label,CpuAccessSize.DWord)

sealed class AsmStatement {
    //Gibt alle lokale Variablen zurück, die zu beginn gelesen werden, hier sind ggf. auch indizes für
    //Array Zugriffe einzubeziehen
    abstract fun localsReadAtStart():Set<AsmLocalOperand>
    //Gibt alle lokalen Variablen zurück, die bei der Instruktion neu beschrieben werden
    abstract fun localsWrittenAtEnd():Set<AsmLocalOperand>
    //Für manche Variablen gibt es Bedingungen in welche Register diese müssen (z.B. shl a,b -> b muss in RCX)
    abstract fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>>
}
class AsmMovOp(val destination: AsmOperand, val source: AsmOperand, val use64BitMode: Boolean):AsmStatement() {
    init {
        if (destination.isMemory() && source.isMemory())
            throw Exception("two memory operand are not allowed in a single bin op")
        if (destination is AsmConstantOperand)
            throw Exception("a constant is not allowed as destination")
    }
    override fun toString() = "mov $destination, $source"

    override fun localsReadAtStart():Set<AsmLocalOperand> =
        setOfAsmLocalOperandDirectOrArraySideEffect(source) +
        setOfAsmLocalOperandArraySideEffect(destination)
    override fun localsWrittenAtEnd():Set<AsmLocalOperand> = setOfAsmLocalOperandDirect(destination)
    override fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>> = listOf()
}
class AsmMovZXOp(val destination: AsmOperand, val source: AsmOperand, val accessSize: CpuAccessSize):AsmStatement() {
    init {
        if (destination.isMemory() && source.isMemory())
            throw Exception("two memory operand are not allowed in a single bin op")
        if (destination is AsmConstantOperand)
            throw Exception("a constant is not allowed as destination")
    }
    override fun toString() = "movzx $destination, $source"

    override fun localsReadAtStart():Set<AsmLocalOperand> =
        setOfAsmLocalOperandDirectOrArraySideEffect(source) +
                setOfAsmLocalOperandArraySideEffect(destination)
    override fun localsWrittenAtEnd():Set<AsmLocalOperand> = setOfAsmLocalOperandDirect(destination)
    override fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>> = listOf()
}
class AsmBinOp(val operation: Operation, val destination:AsmOperand, val source:AsmOperand, val use64BitMode:Boolean):AsmStatement() {
    init {
        if (destination.isMemory() && source.isMemory())
            throw Exception("two memory operand are not allowed in a single bin op")
        if (destination is AsmConstantOperand)
            throw Exception("a constant is not allowed as destination")
    }
    enum class Operation {Add,Sub,And,Or,Xor,Shl,Shr;
        fun isCommutative() = when(this) {
            Add,And,Or,Xor->true
            Shl,Shr,Sub->false
        }
        fun isNoOpWithRValue(value:ULong):Boolean = when(this) {
            Add,Sub,Or,Xor,Shl,Shr-> value == 0UL
            And -> false
        }
    }
    override fun toString() = "${operation.toString().lowercase()} $destination, $source"

    override fun localsReadAtStart():Set<AsmLocalOperand> =
            setOfAsmLocalOperandDirectOrArraySideEffect(source) +
            setOfAsmLocalOperandDirectOrArraySideEffect(destination)
    override fun localsWrittenAtEnd():Set<AsmLocalOperand> = setOfAsmLocalOperandDirect(destination)
    override fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>> =
        if ((operation == Operation.Shl || operation == Operation.Shr) && source is AsmLocalOperand)
            listOf(Pair(source,CpuRegister.RCX))
        else
            listOf()
}
class AsmSetCC(val destination:AsmOperand, val condition: AsmCondition):AsmStatement() {
    override fun toString() = "setcc $destination, (${condition})"
    override fun localsReadAtStart():Set<AsmLocalOperand> = condition.localsReadAtStart()
    override fun localsWrittenAtEnd():Set<AsmLocalOperand> = setOfAsmLocalOperandDirect(destination)
    override fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>> = listOf()
}
class AsmCMovCC(val destination: AsmOperand, val source: AsmLocalOperand, val condition: AsmCondition, val use64BitMode: Boolean):AsmStatement() {
    override fun toString() = "cmovcc $destination, $source, (${condition})"
    override fun localsReadAtStart():Set<AsmLocalOperand> =
        setOfAsmLocalOperandDirectOrArraySideEffect(destination)+
        setOfAsmLocalOperandDirectOrArraySideEffect(source) +
        condition.localsReadAtStart()
    override fun localsWrittenAtEnd():Set<AsmLocalOperand> = setOfAsmLocalOperandDirect(destination)
    override fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>> = listOf()
}
class AsmInvertOp(val destination: AsmOperand, val is64Bit:Boolean):AsmStatement() {
    init {
        if (destination !is AsmLocalOperand)
            throw Exception("invert operation only allowed on local operands")
    }
    override fun toString() = "not $destination"
    override fun localsReadAtStart():Set<AsmLocalOperand> = setOfAsmLocalOperandDirectOrArraySideEffect(destination)
    override fun localsWrittenAtEnd():Set<AsmLocalOperand> = setOfAsmLocalOperandDirect(destination)
    override fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>> = listOf()
}
class AsmReserveTmpCpuReg(val tmp:AsmLocalOperand):AsmStatement() {
    override fun toString() = "reservetmpcpureg $tmp"
    override fun localsReadAtStart():Set<AsmLocalOperand> = setOf()
    override fun localsWrittenAtEnd():Set<AsmLocalOperand> = setOfAsmLocalOperandDirect(tmp)
    override fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>> = listOf()
}
class AsmGetParity(val destination: AsmOperand, val source:AsmLocalOperand, val tmp:AsmReserveTmpCpuReg):AsmStatement() {
    init {
        if (destination === source)
            throw Exception("get parity src must be != dst")
    }
    override fun toString() = "parity $destination, $source, ${tmp.tmp}"
    override fun localsReadAtStart():Set<AsmLocalOperand> = setOfAsmLocalOperandDirectOrArraySideEffect(source)+setOfAsmLocalOperandDirect(tmp.tmp)
    override fun localsWrittenAtEnd():Set<AsmLocalOperand> = setOfAsmLocalOperandDirect(destination)+setOfAsmLocalOperandDirect(tmp.tmp)
    override fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>> = listOf()
}
class AsmCall(val destination: AsmOperand, val inputs:List<Pair<AsmLocalOperand,CpuRegister>>, val affectedInternals:List<Pair<AsmLocalOperand,CpuRegister>>):AsmStatement() {
    //derzeit werden inputs als überschrieben markiert
    override fun toString() = "call $destination with $inputs"
    override fun localsReadAtStart():Set<AsmLocalOperand> {
        val result = HashSet(setOfAsmLocalOperandDirectOrArraySideEffect(destination))
        for (inp in inputs)
            result.addAll(setOfAsmLocalOperandDirectOrArraySideEffect(inp.first))
        return result
    }
    override fun localsWrittenAtEnd():Set<AsmLocalOperand> {
        val result = HashSet<AsmLocalOperand>()
        for (inp in inputs+affectedInternals)
            result.addAll(setOfAsmLocalOperandDirectOrArraySideEffect(inp.first))
        return result
    }
    override fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>> = inputs+affectedInternals
}

class AsmResultAndFlagOp(val operation:Operation, val destination: AsmLocalOperand, val right:AsmOperand, val carryInCopy:AsmOperand, val carryResult:AsmOperand?, val carryInverted:Boolean, val zeroResult:AsmOperand?, val zeroInverted:Boolean):AsmStatement() {
    init {
        if (!operation.isArithmetic()) {
            if (carryResult!=null || carryInCopy !is AsmConstantOperand || carryInCopy.value != 0UL)
                throw Exception("illegal arguments")
        }
        else if (carryInCopy !is AsmConstantOperand && carryInCopy !is AsmLocalOperand)
            throw Exception("illegal arguments")
    }
    enum class Operation {
        AddU,SubU,AddS,SubS,Or,And,Xor,Shl,Shr,Neg;
        fun isArithmetic() = when(this) {
            AddU,SubU,AddS,SubS -> true
            Or,And,Xor,Shl,Shr,Neg -> false
        }
    }
    override fun toString() = "${operation.toString().lowercase()} $destination, $right"+
            (if (zeroResult != null) (if (zeroInverted) " !z->" else " z->")+zeroResult else "")+
            (if (carryResult != null) (if (carryInverted) " !c->" else " c->")+carryResult else "")

    override fun localsReadAtStart():Set<AsmLocalOperand> =
        setOfAsmLocalOperandDirectOrArraySideEffect(destination)+
        setOfAsmLocalOperandDirectOrArraySideEffect(right)+
        setOfAsmLocalOperandDirectOrArraySideEffect(carryInCopy)+
        setOfAsmLocalOperandArraySideEffect(carryResult)+
        setOfAsmLocalOperandArraySideEffect(zeroResult)
    override fun localsWrittenAtEnd():Set<AsmLocalOperand> = setOfAsmLocalOperandDirect(carryInCopy)+setOfAsmLocalOperandDirect(destination)+setOfAsmLocalOperandDirect(carryResult)+setOfAsmLocalOperandDirect(zeroResult)
    override fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>> =
        if ((operation == AsmResultAndFlagOp.Operation.Shl || operation == AsmResultAndFlagOp.Operation.Shr) && right is AsmLocalOperand)
            listOf(Pair(right,CpuRegister.RCX))
        else
            listOf()
}

class AsmLookupOp(val table:AsmAbstractLookupTable, val destination: AsmLocalOperand, val index: AsmLocalOperand, val tmp:AsmReserveTmpCpuReg):AsmStatement() {
    override fun toString() = "lookup $destination, ${table.label}[$index]"
    override fun localsReadAtStart():Set<AsmLocalOperand> = setOfAsmLocalOperandArraySideEffect(destination)+setOfAsmLocalOperandDirectOrArraySideEffect(index)+setOfAsmLocalOperandDirect(tmp.tmp)
    override fun localsWrittenAtEnd():Set<AsmLocalOperand> = setOfAsmLocalOperandDirect(destination)+setOfAsmLocalOperandDirect(tmp.tmp)
    override fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>> = listOf()
}

class AsmGetCogNumber(val destination: AsmLocalOperand):AsmStatement() {
    override fun toString() = "getcognumber $destination"
    override fun localsReadAtStart():Set<AsmLocalOperand> = setOf()
    override fun localsWrittenAtEnd():Set<AsmLocalOperand> = setOfAsmLocalOperandDirect(destination)
    override fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>> = listOf()
}

class AsmCondition(val operation:Operation, val left:AsmOperand, val right:AsmOperand, val is64Bit:Boolean) {
    fun localsReadAtStart():Set<AsmLocalOperand> = setOfAsmLocalOperandDirectOrArraySideEffect(left)+setOfAsmLocalOperandDirectOrArraySideEffect(right)
    enum class Operation { CmpLe,CmpLeq,CmpGe,CmpGeq,CmpSgnLe,CmpSgnLeq,CmpSgnGeq,CmpSgnGe,CmpEq,CmpNeq,TestNz,TestZ;
        fun logicInverted() = when(this) {
            CmpEq->CmpNeq
            CmpNeq->CmpEq
            CmpLe->CmpGeq
            CmpGeq->CmpLe
            CmpLeq -> CmpGe
            CmpGe -> CmpLeq
            CmpSgnLe->CmpSgnGeq
            CmpSgnGeq->CmpSgnLe
            CmpSgnLeq->CmpSgnGe
            CmpSgnGe->CmpSgnLeq
            TestNz->TestZ
            TestZ->TestNz
        }
        fun operandsSwapped() = when(this) {
            CmpEq->CmpEq
            CmpNeq->CmpNeq
            CmpLe->CmpGe
            CmpGe->CmpLe
            CmpLeq -> CmpGeq
            CmpGeq -> CmpLeq
            CmpSgnLe->CmpSgnGe
            CmpSgnGe->CmpSgnLe
            CmpSgnLeq->CmpSgnGeq
            CmpSgnGeq->CmpSgnLeq
            TestNz->TestNz
            TestZ->TestZ
        }
        fun isCommutative() = when(this) {
            CmpEq,CmpNeq,TestNz,TestZ->true
            CmpLe,CmpLeq,CmpSgnLe,CmpSgnLeq,CmpGe,CmpGeq,CmpSgnGe,CmpSgnGeq->false
        }
    }
    init {
        if (left.isMemory() && right.isMemory())
            throw Exception("left and right cannot be memory operad at same time")
        if (left is AsmConstantOperand)
            throw Exception("left operand cannot be a constant")
    }
    fun logicInverted() = AsmCondition(operation.logicInverted(),left,right,is64Bit)
    override fun toString() = "${operation.toString().lowercase()} $left, $right"
}
class AsmComment(val comment:String):AsmStatement() {
    override fun toString() = ";$comment"
    override fun localsReadAtStart():Set<AsmLocalOperand> = setOf()
    override fun localsWrittenAtEnd():Set<AsmLocalOperand> = setOf()
    override fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>> = listOf()
}

sealed class EndOfAsmSection : AsmStatement()
class JumpToAsmSection(val nextSection:AsmSection):EndOfAsmSection() {
    override fun toString() = "jmp $nextSection"
    override fun localsReadAtStart():Set<AsmLocalOperand> = setOf()
    override fun localsWrittenAtEnd():Set<AsmLocalOperand> = setOf()
    override fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>> = listOf()
}
class BranchToAsmSection(val condition: AsmCondition, val nextSectionIfTrue:AsmSection, val nextSectionIfFalse:AsmSection):EndOfAsmSection() {
    override fun toString() = "jmp if (${condition}) $nextSectionIfTrue else $nextSectionIfFalse"
    override fun localsReadAtStart():Set<AsmLocalOperand> = condition.localsReadAtStart()
    override fun localsWrittenAtEnd():Set<AsmLocalOperand> = setOf()
    override fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>> = listOf()
}
class AsmReturnToCaller():EndOfAsmSection() {
    override fun toString() = "ret"
    override fun localsReadAtStart():Set<AsmLocalOperand> = setOf()
    override fun localsWrittenAtEnd():Set<AsmLocalOperand> = setOf()
    override fun constraintsForLocals():List<Pair<AsmLocalOperand,CpuRegister>> = listOf()
}

class AsmSection(val id:Int) {
    val body = ArrayList<AsmStatement>()
    var endOfSection:EndOfAsmSection?=null
    fun add(stm:AsmStatement):AsmSection {
        body.add(stm)
        return this
    }
    override fun toString() = ".sect$id"
    fun iterate(cb:(stm:AsmStatement)->Unit) {
        for (s in body)
            cb(s)
        cb(forceEndOfSection())
    }
    fun forceEndOfSection() = endOfSection ?: throw Exception("no end of section found in $id")
}
