package assembler

import ilang.*

//Condition Expressions, sind Ausdrücke, die das Ergebnis in Flags liefern sollen (meisten z.B. cmp)

object AsmConditionTranslator {
    private fun generateGetPllOfCog(state:AsmTranslatorState, cogNo:Expression):AsmCondition {
        val memOp = if (cogNo is ConstantExpression)
            AsmRamHubBaseOperand("HubRelFirstCogRegOffset+cog_reg_pll_saved_both+((0x${cogNo.value.toString(16)})<<PerCogSizeLog2)", null, 0, ILangDataType.Bool)
        else {
            val tmp=state.nextLocal()
            AsmDataExpressionTranslator.generate(state, tmp, cogNo)
            state.addBinOp(AsmBinOp.Operation.Shl,tmp,AsmLabelOperand("PerCogSizeLog2"),false)
            AsmRamHubBaseOperand("HubRelFirstCogRegOffset+cog_reg_pll_saved_both", tmp, 1, ILangDataType.Bool)
        }
        return AsmCondition(AsmCondition.Operation.TestNz,memOp,state.operandOf(1UL),false)
    }
    private fun generateCompare(state:AsmTranslatorState, opCode:AsmCondition.Operation, left: Expression, right:AsmOperand):AsmCondition {
        val is64BIt = left.dataType == ILangDataType.Int64
        if (left is VariableExpression)
            return AsmCondition(opCode,state.createLocalCopy(state.operandReadOf(left), right.isMemory()),right,is64BIt)
        val tmp=state.nextLocal()
        AsmDataExpressionTranslator.generate(state, tmp,left)
        return AsmCondition(opCode,tmp,right,is64BIt)
    }

    fun generateCompare(state:AsmTranslatorState, opCode:AsmCondition.Operation, left: Expression, right: Expression):AsmCondition {
        if (right is ConstantExpression && right.fits32Bit())
            return generateCompare(state,opCode,left,state.operandOf(right))
        if (left is ConstantExpression && left.fits32Bit())
            return generateCompare(state,opCode.operandsSwapped(),right,state.operandOf(left))
        if (right is VariableExpression)
            return generateCompare(state,opCode,left,state.operandReadOf(right))
        if (left is VariableExpression)
            return generateCompare(state,opCode.operandsSwapped(),right,state.operandReadOf(left))
        val tmp=state.nextLocal()
        AsmDataExpressionTranslator.generate(state, tmp,right)
        return generateCompare(state,opCode,left,tmp)
    }

    private fun  generateTestBit(state:AsmTranslatorState, expr:TestBitExpression):AsmCondition {
        if (expr.bitNo !is ConstantExpression)
            throw Exception("testBit only supported for const bitNo")
        val bitNo = state.operandOf(1UL shl expr.bitNo.value.toInt())
        if (expr.bitNo.value < 32UL && expr.expr.dataType != ILangDataType.Int64)
            return generateCompare(state,AsmCondition.Operation.TestNz,expr.expr,bitNo)
        val tmp=state.nextLocal()
        state.addMov(tmp,bitNo,true)
        return generateCompare(state,AsmCondition.Operation.TestNz,expr.expr,tmp)
    }

    fun generate(state:AsmTranslatorState, expr: Expression):AsmCondition = when(expr) {
        is VariableExpression -> generateCompare(state,AsmCondition.Operation.CmpEq,expr,state.operandOf(true))
        is TestBitExpression -> generateTestBit(state,expr)
        is EqualityExpression -> generateCompare(state,if (expr.isEq) AsmCondition.Operation.CmpEq else AsmCondition.Operation.CmpNeq,expr.left,expr.right)
        is CompareExpression -> generateCompare(state,when(expr.operation) {
            CompareExpression.Operation.Le -> AsmCondition.Operation.CmpLe
            CompareExpression.Operation.Leq -> AsmCondition.Operation.CmpLeq
            CompareExpression.Operation.Ge -> AsmCondition.Operation.CmpGe
            CompareExpression.Operation.Geq -> AsmCondition.Operation.CmpGeq
            CompareExpression.Operation.SgnLe -> AsmCondition.Operation.CmpSgnLe
        },expr.left, expr.right)
        is InvertExpression -> generate(state,expr.expr).logicInverted()
        is GetPllOfCogExpression -> generateGetPllOfCog(state,expr.cogNo)
        else -> throw Exception("expression for compare not supported: $expr")
    }
}

