package assembler

import java.lang.Exception

//Hier werden kollidierene lokale Variablen analysiert und dann auf CPU Register verteilt

class AsmLocalLiveStatus {
    val requiredAtStartOfInstruction=HashSet<AsmStatement>()
    val requiredAtEndOfInstruction=HashSet<AsmStatement>()
}

fun getAsmLocalVariableLiveStatus(section: AsmSection, keepVars:List<AsmLocalOperand>, result:HashMap<AsmLocalOperand,AsmLocalLiveStatus>, passedInstructions:HashMap<AsmLocalOperand,HashSet<AsmStatement>>) {
    fun handleStm(stm:AsmStatement) {
        if (stm is AsmComment) //das ist nur ein pseudo stm
            return
        fun markVarAsReadHere(v:AsmLocalOperand) {
            val passed = passedInstructions.getOrPut(v) { HashSet() }
            val resultForVar = result[v] ?: throw Exception("depends on not initialized data ($v) $stm")
            resultForVar.requiredAtEndOfInstruction.addAll(passed)
            resultForVar.requiredAtStartOfInstruction.addAll(passed)
            resultForVar.requiredAtStartOfInstruction.add(stm)
            //die passed stms können jetzt gelöscht werden
            passed.clear()
        }
        for (inputVar in stm.localsReadAtStart()) {
            //diese Variable wird jetzt als Eingabewert gebraucht
            //die Variable muss somit bei allen bisherigen Instruktionen live gehalten werden
            markVarAsReadHere(inputVar)
        }
        val writtenVars = stm.localsWrittenAtEnd()
        for (outputVar in writtenVars) {
            //diese Variable wird jetzt geschrieben
            //alle bisher in der passed Liste gehalten Instruktionen können verfallen, da der Wert zwischenzeitlich nicht gebraucht wurde
            val resultForVar = result.getOrPut(outputVar) { AsmLocalLiveStatus() }
            resultForVar.requiredAtEndOfInstruction.add(stm)
            //in der zwischenzeit durchlaufene instruktionen brauchen den wert nicht, da er ja hier geschrieben wurde
            passedInstructions[outputVar]?.clear()
        }
        //das statement zu allen nicht geschriebenen aber definierten variablen als passed hinzufügen
        for (p in result.keys.subtract(writtenVars)) {
            val passed = passedInstructions.getOrPut(p) { HashSet() }
            passed.add(stm)
        }
        //beim return kommando noch alle keepVars als verwendet markieren
        if (stm is AsmReturnToCaller) {
            for (keepVar in keepVars)
                markVarAsReadHere(keepVar)
        }

    }
    fun createCopyOfPassedInstructions():HashMap<AsmLocalOperand,HashSet<AsmStatement>> {
        //tiefe kopie anlegen
        val copyPassed = HashMap<AsmLocalOperand,HashSet<AsmStatement>>()
        for (v in passedInstructions)
            copyPassed[v.key] = HashSet(v.value)
        return copyPassed
    }
    section.iterate { handleStm(it) }
    when(val endOfSection = section.forceEndOfSection()) {
        is AsmReturnToCaller -> { } //ende
        is JumpToAsmSection -> getAsmLocalVariableLiveStatus(endOfSection.nextSection,keepVars,result,passedInstructions) //einfach weitermachen
        is BranchToAsmSection -> {
            //es gibt zwei mögliche pfade, beide komplett getrennt gehen
            getAsmLocalVariableLiveStatus(endOfSection.nextSectionIfTrue,keepVars,result,createCopyOfPassedInstructions())
            getAsmLocalVariableLiveStatus(endOfSection.nextSectionIfFalse,keepVars,result,createCopyOfPassedInstructions())
        }
    }
}

fun findAsmLocalsLiveAtSameTime(allSections:List<AsmSection>, stmLiveStatus:HashMap<AsmLocalOperand,AsmLocalLiveStatus>):HashMap<AsmLocalOperand,HashSet<AsmLocalOperand>> {
    val liveAtSameTime = HashMap<AsmLocalOperand,HashSet<AsmLocalOperand>>()
    for (op in stmLiveStatus.keys)
        liveAtSameTime[op] = HashSet()

    fun addConflictingVars(lst:Set<AsmLocalOperand>) {
        for (a in lst) {
            for (b in lst) {
                if (a===b)
                    continue
                (liveAtSameTime[a] ?: throw Exception("not found")).add(b)
            }
        }
    }
    for (section in allSections) {
        section.iterate {
            stm ->
            val localsRequiredAtStart = HashSet<AsmLocalOperand>()
            val localsRequiredAtEnd = HashSet<AsmLocalOperand>()
            for (liveSt in stmLiveStatus) {
                if (liveSt.value.requiredAtStartOfInstruction.contains(stm))
                    localsRequiredAtStart.add(liveSt.key)
                if (liveSt.value.requiredAtEndOfInstruction.contains(stm))
                    localsRequiredAtEnd.add(liveSt.key)
            }
            addConflictingVars(localsRequiredAtStart)
            addConflictingVars(localsRequiredAtEnd)
        }
    }

    return liveAtSameTime
}

//analysiert, ob spezielle lokale variablen zwingend in gewisse cpu register müssen
fun getAsmLocalsConstraints(allSections:List<AsmSection>):HashMap<AsmLocalOperand,CpuRegister> {
    val result = HashMap<AsmLocalOperand,CpuRegister>()
    for (section in allSections) {
        section.iterate {
            stm->
            for (c in stm.constraintsForLocals()) {
                val prev = result[c.first]
                if (prev === c.second)
                    continue
                if (prev != null)
                    throw Exception("conflicting contraints for ${c.first}: ${c.second} vs $prev, last stm: $stm")
                result[c.first] = c.second
            }
        }
    }
    return result
}

//zählt, wie häufig eine lokale variable vorkommt, und sortiert nach dem vorkommen absteigend
fun asmRegisterMapOrder(allSections:List<AsmSection>, inputs:List<AsmLocalOperand>):List<AsmLocalOperand> {
    val counter = HashMap<AsmLocalOperand,Int>()
    for (section in allSections) {
        section.iterate { stm ->
            val affected = stm.localsReadAtStart()+stm.localsWrittenAtEnd()
            for (op in affected)
                counter[op] = (counter[op]?:0)+1
        }
    }
    for (inp in inputs)
        if (!counter.containsKey(inp)) //wenn ein input gar nicht verwendet wird hier hinzufügen
            counter[inp]=0
    //deterministische reihenfolge erzeugen, erst nach häufigkeit, dann nach alphabet sortieren
    return counter.map { it }.sortedWith(Comparator {
        a,b -> if (a.value<b.value) 1 else if (a.value>b.value) -1 else if (a.key.name>b.key.name) 1 else if (a.key.name<b.key.name) -1 else 0

    }).map { it.key }
}

fun mapAsmCpuRegisters(liveAtSameTime:HashMap<AsmLocalOperand,HashSet<AsmLocalOperand>>, cpuRegConstraints:HashMap<AsmLocalOperand,CpuRegister>, mapOrder:List<AsmLocalOperand>):HashMap<AsmLocalOperand,CpuRegister> {
    val allowedCpuRegs = listOf(CpuRegister.RAX,CpuRegister.RCX,CpuRegister.RDX,CpuRegister.R8,CpuRegister.R9,CpuRegister.R10,CpuRegister.R11,CpuRegister.R12)
    //zuerst die constraints mappen
    val forbiddenRegisters = HashMap<AsmLocalOperand,HashSet<CpuRegister>>()
    for (v in mapOrder)
        forbiddenRegisters[v] = HashSet()
    val result = HashMap<AsmLocalOperand,CpuRegister>()
    fun getForbiddenRegs(op:AsmLocalOperand) = forbiddenRegisters[op] ?: throw Exception("entry missing in forbidden registers (internal error) $op")
    fun mapRegister(op:AsmLocalOperand, cpuReg:CpuRegister) {
        if (result.containsKey(op))
            throw Exception("remapped $op (internal error)")
        result[op] = cpuReg
        for (opLiveAtSameTime in liveAtSameTime[op] ?: throw Exception("no live at same time entry found for $op (internal error)"))
            getForbiddenRegs(opLiveAtSameTime).add(cpuReg)
    }
    fun mapFirstFreeRegister(op:AsmLocalOperand) {
        val forbidden = getForbiddenRegs(op)
        for (cpuReg in allowedCpuRegs) {
            if (!forbidden.contains(cpuReg))
                return mapRegister(op, cpuReg)
        }
        throw Exception("unable to map cpu register to $op, no more registers available")
    }
    //zuerst die constraints mappen
    for ((op, cpuReg) in cpuRegConstraints) {
        if (getForbiddenRegs(op).contains(cpuReg))
            throw Exception("cannot map $cpuReg to $op -> conflicting")
        mapRegister(op,cpuReg)
    }
    //jetzt die verbleibenden register mappen
    for (op in mapOrder) {
        if (!result.containsKey(op)) //schon gemapped
            mapFirstFreeRegister(op)
    }
    return result
}
