package assembler

import ilang.ILangDataType
import ilang.ILangVariableSet
import java.lang.StringBuilder

enum class CpuAccessSize {None,QWord,DWord,Word,Byte}
enum class CpuRegister { RAX,RCX,RDX,RSI,RDI,R8,R9,R10,R11,R12,   R14;
    fun toAsmCode(size:CpuAccessSize):String = when(size) {
        CpuAccessSize.None -> throw Exception("unsupported")
        CpuAccessSize.QWord -> toString().lowercase()
        CpuAccessSize.DWord -> when (this) {
            RAX -> "eax"
            RCX -> "ecx"
            RDX -> "edx"
            RSI -> "esi"
            RDI -> "edi"
            R8 -> "r8d"
            R9 -> "r9d"
            R10 -> "r10d"
            R11 -> "r11d"
            R12 -> "r12d"
            R14 -> "r14d"

        }
        CpuAccessSize.Word -> when (this) {
            RAX -> "ax"
            RCX -> "cx"
            RDX -> "dx"
            RSI -> "si"
            RDI -> "di"
            R8 -> "r8w"
            R9 -> "r9w"
            R10 -> "r10w"
            R11 -> "r11w"
            R12 -> "r12w"
            R14 -> "r14w"
        }
        CpuAccessSize.Byte -> when (this) {
            RAX -> "al"
            RCX -> "cl"
            RDX -> "dl"
            RSI -> "sil"
            RDI -> "dil"
            R8 -> "r8b"
            R9 -> "r9b"
            R10 -> "r10b"
            R11 -> "r11b"
            R12 -> "r12b"
            R14 -> "r14b"
        }
    }
}



sealed class AMIOperand {
    abstract fun toAsmCode(size:CpuAccessSize):String
}
class AMIConstantOperand(val value:ULong):AMIOperand() {
    override fun toAsmCode(size:CpuAccessSize):String = "0x${value.toString(16)}"
}
class AMILabelOperand(val lbl:String):AMIOperand() {
    override fun toAsmCode(size:CpuAccessSize):String = lbl
}
class AMIRegisterOperand(val cpuRegister: CpuRegister):AMIOperand() {
    override fun toAsmCode(size:CpuAccessSize):String = cpuRegister.toAsmCode(size)
}
class AMIMemoryOperand(val baseRegister: CpuRegister, val indexRegister:CpuRegister?, val indexMulitplicator:Int, val offset:String):AMIOperand() {
    override fun toAsmCode(size:CpuAccessSize):String = when(size) {
        CpuAccessSize.None -> throw Exception("unsupported")
        CpuAccessSize.QWord -> "qword ["
        CpuAccessSize.Word -> "word ["
        CpuAccessSize.DWord -> "dword ["
        CpuAccessSize.Byte -> "byte ["
    }+baseRegister.toAsmCode(CpuAccessSize.QWord)+(if(indexRegister==null) "" else "+"+indexRegister.toAsmCode(CpuAccessSize.QWord)+"*$indexMulitplicator")+(if (offset == "") "" else "+"+offset)+"]"
}
class AMILocalLabelOperand(val id:Int):AMIOperand() {
    override fun toAsmCode(size:CpuAccessSize):String = ".sect$id"
}

//asm mnemonic instruction
sealed class AMIEntity {
    abstract fun toAsmCode():String
}
class AMIComment(val comment:String):AMIEntity() {
    override fun toAsmCode():String = "    ;$comment"
}
class AMILocalLabel(val id:Int):AMIEntity() {
    override fun toAsmCode():String = ".sect$id:"
}
class AMIOpcode(val operation:Operation, val operands:List<AMIOperand>, val operandSize:CpuAccessSize, val conditionFlags:Condition?=null):AMIEntity() {
    enum class Operation {Mov,MovZx,Lea,Add,Sub,Adc,Sbb,Stc,Neg,And,Or,Xor,Shl,Shr,Not,Test,Cmp,JCC,Jmp,SetCC,Ret,CMovCC,Call}
    enum class Condition {G,Ge,L,Le,A,Ae,B,Be,Z,Nz,P,Np,C,Nc,O,No}

    init {
        if (operation == Operation.MovZx && (operandSize != CpuAccessSize.Byte && operandSize != CpuAccessSize.Word))
            throw Exception("invalid operand for movzx")
    }

    fun opCodeName() = when(operation) {
        Operation.SetCC -> "set"+conditionFlags.toString().lowercase()
        Operation.JCC -> "j"+conditionFlags.toString().lowercase()
        Operation.CMovCC -> "cmov"+conditionFlags.toString().lowercase()
        else -> operation.toString().lowercase()
    }

    override fun toAsmCode():String {
        val result = "    "+opCodeName()+if (operands.isNotEmpty()) " " else ""
        if (operation == Operation.Lea) {
            val op1 = (if (operands[1] is AMILabelOperand) "rel " else "")+operands[1].toAsmCode(operandSize)
            return result+operands[0].toAsmCode(operandSize)+", [$op1]"
        }
        if (operation == Operation.MovZx)
            return result+operands[0].toAsmCode(CpuAccessSize.DWord)+", "+operands[1].toAsmCode(operandSize)
        if (operation == Operation.Shl || operation == Operation.Shr) {
            val opRight = operands[1]
            if (opRight is AMIRegisterOperand && opRight.cpuRegister == CpuRegister.RCX)
                return result+operands[0].toAsmCode(operandSize)+", cl"
        }
        return result+operands.joinToString { it.toAsmCode(operandSize)}
    }

}

fun generateAMIStream(sections:List<AsmSection>, registerMapping:HashMap<AsmLocalOperand,CpuRegister>, hubRegs:ILangVariableSet, cogRegs:ILangVariableSet):List<AMIEntity> {
    val result = ArrayList<AMIEntity>()
    val hubBaseCpuReg = CpuRegister.RDI
    val cogBaseCpuReg = CpuRegister.RSI
    fun mapToCpuReg(op:AsmLocalOperand) = registerMapping[op] ?: throw Exception("no register mapping found for $op")
    fun transformOp(op:AsmOperand):AMIOperand = when(op) {
        is AsmConstantOperand -> AMIConstantOperand(op.value)
        is AsmLabelOperand -> AMILabelOperand(op.lbl)
        is AsmLocalOperand -> AMIRegisterOperand(mapToCpuReg(op))
        is AsmMemOperand ->
            if (hubRegs.contains(op.variable))
                AMIMemoryOperand(hubBaseCpuReg, null,0, "hub_"+op.variable.name)
            else if (cogRegs.contains(op.variable))
                AMIMemoryOperand(cogBaseCpuReg, null,0, "cog_"+op.variable.name)
            else throw Exception("array is no hub or cog reg: ${op.variable}")
        is AsmArrayOperand -> {
            val offset = op.arrayVariable.name+if(op.byteOffset != 0) "+${op.byteOffset}" else ""
            if (hubRegs.contains(op.arrayVariable))
                AMIMemoryOperand(
                    hubBaseCpuReg,
                    mapToCpuReg(op.idx),
                    op.idxMultiplicator,
                    "hub_$offset"
                )
            else if (cogRegs.contains(op.arrayVariable))
                AMIMemoryOperand(
                    cogBaseCpuReg,
                    mapToCpuReg(op.idx),
                    op.idxMultiplicator,
                    "cog_$offset"
                )
            else throw Exception("array is no hub or cog reg: ${op.arrayVariable}")
        }
        is AsmRamHubBaseOperand -> AMIMemoryOperand(hubBaseCpuReg, if (op.idx == null) null else mapToCpuReg(op.idx), op.idxMultiplicator, op.offset)
    }
    fun isByteOperand(op:AsmOperand) = when(op) {
        is AsmConstantOperand -> false
        is AsmLocalOperand -> false
        is AsmLabelOperand -> false
        is AsmMemOperand -> op.variable.dataType == ILangDataType.Bool
        is AsmArrayOperand -> op.arrayVariable.dataType == ILangDataType.Bool
        is AsmRamHubBaseOperand -> op.dataType == ILangDataType.Bool
    }
    fun canBeQWordOperand(op:AsmOperand) = when(op) {
        is AsmConstantOperand -> true
        is AsmLabelOperand -> true
        is AsmLocalOperand -> true
        is AsmMemOperand -> op.variable.dataType == ILangDataType.Int64
        is AsmArrayOperand -> op.arrayVariable.dataType == ILangDataType.Int64
        is AsmRamHubBaseOperand -> op.dataType == ILangDataType.Int64
    }
    fun getBinOpSize(left:AsmOperand, right:AsmOperand, is64Bit:Boolean):CpuAccessSize {
        if (is64Bit) {
            if (!canBeQWordOperand(left) || !canBeQWordOperand(right))
                throw Exception("operand size error $left $right $is64Bit")
            return CpuAccessSize.QWord
        }
        if (isByteOperand(left) || isByteOperand(right))
            return CpuAccessSize.Byte
        return CpuAccessSize.DWord
    }
    fun generateCmpOrTest(condition: AsmCondition):Pair<AMIOpcode,AMIOpcode.Condition> {
        //CmpLe,CmpLeq,CmpGe,CmpGeq,CmpSgnLe,CmpSgnGeq,CmpEq,CmpNeq,TestNz,TestZ
        return Pair(AMIOpcode(
            when(condition.operation) {
                AsmCondition.Operation.CmpLe -> AMIOpcode.Operation.Cmp
                AsmCondition.Operation.CmpLeq -> AMIOpcode.Operation.Cmp
                AsmCondition.Operation.CmpGe -> AMIOpcode.Operation.Cmp
                AsmCondition.Operation.CmpGeq -> AMIOpcode.Operation.Cmp
                AsmCondition.Operation.CmpSgnLe -> AMIOpcode.Operation.Cmp
                AsmCondition.Operation.CmpSgnLeq -> AMIOpcode.Operation.Cmp
                AsmCondition.Operation.CmpSgnGe -> AMIOpcode.Operation.Cmp
                AsmCondition.Operation.CmpSgnGeq -> AMIOpcode.Operation.Cmp
                AsmCondition.Operation.CmpEq -> AMIOpcode.Operation.Cmp
                AsmCondition.Operation.CmpNeq -> AMIOpcode.Operation.Cmp
                AsmCondition.Operation.TestNz -> AMIOpcode.Operation.Test
                AsmCondition.Operation.TestZ -> AMIOpcode.Operation.Test
            },
            listOf(transformOp(condition.left), transformOp(condition.right)),
            getBinOpSize(condition.left,condition.right,condition.is64Bit)
        ),when(condition.operation) {
            AsmCondition.Operation.CmpLe -> AMIOpcode.Condition.B
            AsmCondition.Operation.CmpLeq -> AMIOpcode.Condition.Be
            AsmCondition.Operation.CmpGe -> AMIOpcode.Condition.A
            AsmCondition.Operation.CmpGeq -> AMIOpcode.Condition.Ae
            AsmCondition.Operation.CmpSgnLe -> AMIOpcode.Condition.L
            AsmCondition.Operation.CmpSgnLeq -> AMIOpcode.Condition.Le
            AsmCondition.Operation.CmpSgnGe -> AMIOpcode.Condition.G
            AsmCondition.Operation.CmpSgnGeq -> AMIOpcode.Condition.Ge
            AsmCondition.Operation.CmpEq -> AMIOpcode.Condition.Z
            AsmCondition.Operation.CmpNeq -> AMIOpcode.Condition.Nz
            AsmCondition.Operation.TestNz -> AMIOpcode.Condition.Nz
            AsmCondition.Operation.TestZ -> AMIOpcode.Condition.Z
        })
    }
    fun addSetCC(destOp:AMIOperand, conditionFlags: AMIOpcode.Condition) {
        result.add(AMIOpcode(AMIOpcode.Operation.SetCC,listOf(destOp),CpuAccessSize.Byte,conditionFlags))
    }

    fun bitOpWithFlags(opcode:AMIOpcode.Operation, stm:AsmResultAndFlagOp) {
        result.add(AMIOpcode(opcode,
            listOf(transformOp(stm.destination),transformOp(stm.right)),
            CpuAccessSize.DWord
        ))
    }

    fun shiftWithFlags(shiftLeft:Boolean, stm:AsmResultAndFlagOp) {
        result.add(AMIOpcode(if (shiftLeft) AMIOpcode.Operation.Shl else AMIOpcode.Operation.Shr,
            listOf(transformOp(stm.destination),transformOp(stm.right)),
            CpuAccessSize.DWord
        ))
        //damit das zero flag auch bei einem 0 shift richtig gesetzt wird
        if (stm.zeroResult != null)
            result.add(AMIOpcode(AMIOpcode.Operation.Test,listOf(transformOp(stm.destination),transformOp(stm.destination)),CpuAccessSize.DWord))
    }

    fun arithmeticWithFlags(sub:Boolean, isUnsigned:Boolean, stm:AsmResultAndFlagOp) {
        if (stm.carryInCopy is AsmConstantOperand) {
            if (stm.carryInCopy.value == 0UL) //carry in ist statisch false
                result.add(AMIOpcode(if (sub) AMIOpcode.Operation.Sub else AMIOpcode.Operation.Add,listOf(transformOp(stm.destination),transformOp(stm.right)),CpuAccessSize.DWord))
            else { //carry in ist statisch true
                result.add(AMIOpcode(AMIOpcode.Operation.Stc, listOf(),CpuAccessSize.None))
                result.add(AMIOpcode(if (sub) AMIOpcode.Operation.Sbb else AMIOpcode.Operation.Adc,listOf(transformOp(stm.destination),transformOp(stm.right)),CpuAccessSize.DWord))
            }
        }
        else if (stm.carryInCopy is AsmLocalOperand) { //carry in ist dynamisch
            result.add(AMIOpcode(AMIOpcode.Operation.Shr,listOf(transformOp(stm.carryInCopy),transformOp(AsmConstantOperand(1UL))),CpuAccessSize.DWord))
            result.add(AMIOpcode(if (sub) AMIOpcode.Operation.Sbb else AMIOpcode.Operation.Adc,listOf(transformOp(stm.destination),transformOp(stm.right)),CpuAccessSize.DWord))
        }
        else
            throw Exception("illegal carryInCopy: ${stm.carryInCopy}")

        if (stm.carryResult != null) {
            val testFlag =
                if (isUnsigned)
                    if (stm.carryInverted) AMIOpcode.Condition.Nc else AMIOpcode.Condition.C
                else
                    if (stm.carryInverted) AMIOpcode.Condition.No else AMIOpcode.Condition.O
            result.add(
                AMIOpcode(
                    AMIOpcode.Operation.SetCC,
                    listOf(transformOp(stm.carryResult)),
                    CpuAccessSize.Byte,
                    testFlag
                )
            )
            //and res,0x1, kommt erst nach dem setzen des zflags
        }
    }

    for (section in sections) {
        result.add(AMILocalLabel(section.id))
        section.iterate { stm ->
            when(stm) {
                is AsmComment -> result.add(AMIComment(stm.comment))
                is JumpToAsmSection -> result.add(AMIOpcode(AMIOpcode.Operation.Jmp,listOf(AMILocalLabelOperand(stm.nextSection.id)),CpuAccessSize.None))
                is BranchToAsmSection -> {
                    val (cmpOrTestInstr,conditionFlags) = generateCmpOrTest(stm.condition)
                    result.add(cmpOrTestInstr)
                    result.add(AMIOpcode(AMIOpcode.Operation.JCC,listOf(AMILocalLabelOperand(stm.nextSectionIfTrue.id)),CpuAccessSize.None,conditionFlags))
                    result.add(AMIOpcode(AMIOpcode.Operation.Jmp,listOf(AMILocalLabelOperand(stm.nextSectionIfFalse.id)),CpuAccessSize.None))
                }
                is AsmReturnToCaller -> result.add(AMIOpcode(AMIOpcode.Operation.Ret,listOf(),CpuAccessSize.None))
                is AsmMovOp -> {
                    if (isByteOperand(stm.source) && !isByteOperand(stm.destination))
                        result.add(AMIOpcode(AMIOpcode.Operation.MovZx,listOf(
                            transformOp(stm.destination),
                            transformOp(stm.source)
                        ),CpuAccessSize.Byte))
                    else
                        result.add(AMIOpcode(AMIOpcode.Operation.Mov,listOf(
                            transformOp(stm.destination),
                            transformOp(stm.source)
                        ),getBinOpSize(stm.destination,stm.source,stm.use64BitMode)))
                }
                is AsmMovZXOp -> {
                    val dest = transformOp(stm.destination)
                    val src = transformOp(stm.source)
                    result.add(AMIOpcode(
                        (if (dest is AMIMemoryOperand || stm.accessSize == CpuAccessSize.DWord) AMIOpcode.Operation.Mov
                        else if (stm.accessSize == CpuAccessSize.Word || stm.accessSize == CpuAccessSize.Byte) AMIOpcode.Operation.MovZx
                        else throw Exception("illegal movzx operands")),
                        listOf(dest,src),stm.accessSize))
                }
                is AsmBinOp -> {
                    result.add(AMIOpcode(when(stm.operation) {
                            AsmBinOp.Operation.And -> AMIOpcode.Operation.And
                            AsmBinOp.Operation.Or  -> AMIOpcode.Operation.Or
                            AsmBinOp.Operation.Xor -> AMIOpcode.Operation.Xor
                            AsmBinOp.Operation.Shl -> AMIOpcode.Operation.Shl
                            AsmBinOp.Operation.Shr -> AMIOpcode.Operation.Shr
                            AsmBinOp.Operation.Add -> AMIOpcode.Operation.Add
                            AsmBinOp.Operation.Sub -> AMIOpcode.Operation.Sub
                        },
                        listOf(
                            transformOp(stm.destination),
                            transformOp(stm.source)
                        ),getBinOpSize(stm.destination,stm.source,stm.use64BitMode)
                    ))
                }
                is AsmSetCC -> {
                    val (cmpOrTestInstr,conditionFlags) = generateCmpOrTest(stm.condition)
                    result.add(cmpOrTestInstr)
                    val destOp = transformOp(stm.destination)
                    addSetCC(destOp,conditionFlags)
                    if (destOp is AMIRegisterOperand)
                        result.add(AMIOpcode(AMIOpcode.Operation.And, listOf(destOp,AMIConstantOperand(1UL)),CpuAccessSize.DWord))
                }
                is AsmCMovCC -> {
                    val (cmpOrTestInstr,conditionFlags) = generateCmpOrTest(stm.condition)
                    result.add(cmpOrTestInstr)
                    result.add(AMIOpcode(AMIOpcode.Operation.CMovCC,listOf(transformOp(stm.destination),transformOp(stm.source)),if(stm.use64BitMode) CpuAccessSize.QWord else CpuAccessSize.DWord,conditionFlags))
                }
                is AsmInvertOp -> {
                    result.add(AMIOpcode(
                        AMIOpcode.Operation.Not,
                        listOf(transformOp(stm.destination)),
                        if (stm.is64Bit) CpuAccessSize.QWord else CpuAccessSize.DWord))
                }
                is AsmGetParity -> {
                    val srcOp = AMIRegisterOperand(mapToCpuReg(stm.source)) //ist immer ein register
                    val tmp = AMIRegisterOperand(mapToCpuReg(stm.tmp.tmp)) //ist immer ein register

                    result.add(AMIOpcode(AMIOpcode.Operation.Mov,listOf(tmp,srcOp),CpuAccessSize.DWord)) //mov tmp,srcOp
                    result.add(AMIOpcode(AMIOpcode.Operation.Shr,listOf(tmp,AMIConstantOperand(16UL)),CpuAccessSize.DWord)) //shr tmp,16
                    result.add(AMIOpcode(AMIOpcode.Operation.Xor,listOf(srcOp,tmp),CpuAccessSize.DWord)) //xor srcOp,tmp

                    result.add(AMIOpcode(AMIOpcode.Operation.Mov,listOf(tmp,srcOp),CpuAccessSize.DWord)) //mov tmp,srcOp
                    result.add(AMIOpcode(AMIOpcode.Operation.Shr,listOf(tmp,AMIConstantOperand(8UL)),CpuAccessSize.DWord)) //shr tmp,8
                    result.add(AMIOpcode(AMIOpcode.Operation.Xor,listOf(srcOp,tmp),CpuAccessSize.DWord)) //xor srcOp,tmp
                    val destOp = transformOp(stm.destination)
                    addSetCC(destOp,AMIOpcode.Condition.Np) ///setnp dest (ggf and dest,1)
                    if (destOp is AMIRegisterOperand)
                        result.add(AMIOpcode(AMIOpcode.Operation.And, listOf(destOp,AMIConstantOperand(1UL)),CpuAccessSize.DWord))
                }
                is AsmResultAndFlagOp -> {
                    result.add(AMIComment("ResultWithFlag Operation: ${stm.operation}"))

                    when(stm.operation) {
                        AsmResultAndFlagOp.Operation.And -> bitOpWithFlags(AMIOpcode.Operation.And,stm)
                        AsmResultAndFlagOp.Operation.Or  -> bitOpWithFlags(AMIOpcode.Operation.Or,stm)
                        AsmResultAndFlagOp.Operation.Xor -> bitOpWithFlags(AMIOpcode.Operation.Xor,stm)
                        AsmResultAndFlagOp.Operation.Shl -> shiftWithFlags(true,stm)
                        AsmResultAndFlagOp.Operation.Shr -> shiftWithFlags(false,stm)
                        AsmResultAndFlagOp.Operation.AddU -> arithmeticWithFlags(false,true,stm)
                        AsmResultAndFlagOp.Operation.SubU -> arithmeticWithFlags(true,true,stm)
                        AsmResultAndFlagOp.Operation.AddS -> arithmeticWithFlags(false,false,stm)
                        AsmResultAndFlagOp.Operation.SubS -> arithmeticWithFlags(true,false,stm)
                        AsmResultAndFlagOp.Operation.Neg -> result.add(AMIOpcode(AMIOpcode.Operation.Neg, listOf(transformOp(stm.destination)), CpuAccessSize.DWord))
                    }
                    if (stm.zeroResult != null) {
                        val testFlag = if (stm.zeroInverted) AMIOpcode.Condition.Nz else AMIOpcode.Condition.Z
                        result.add(
                            AMIOpcode(
                                AMIOpcode.Operation.SetCC,
                                listOf(transformOp(stm.zeroResult)),
                                CpuAccessSize.Byte,
                                testFlag
                            )
                        )
                        if (stm.zeroResult is AsmLocalOperand)
                            result.add(AMIOpcode(AMIOpcode.Operation.And,listOf(transformOp(stm.zeroResult),transformOp(AsmConstantOperand(1UL))),CpuAccessSize.DWord))
                    }
                    //erst jetzt ggf carry mit 0x1 anden, damit das zero flag nicht beeinträchtigt wird
                    if (stm.carryResult != null && stm.carryResult is AsmLocalOperand)
                        result.add(AMIOpcode(AMIOpcode.Operation.And,listOf(transformOp(stm.carryResult),transformOp(AsmConstantOperand(1UL))),CpuAccessSize.DWord))
                }
                is AsmCall -> result.add(AMIOpcode(AMIOpcode.Operation.Call,listOf(transformOp(stm.destination)),CpuAccessSize.QWord))
                is AsmLookupOp -> {
                    val tmpReg = mapToCpuReg(stm.tmp.tmp)
                    result.add(AMIOpcode(AMIOpcode.Operation.Lea,listOf(AMIRegisterOperand(tmpReg),transformOp(stm.table.label)),CpuAccessSize.QWord))
                    result.add(when(stm.table.tableAccessSize) {
                        CpuAccessSize.None,CpuAccessSize.QWord -> throw Exception("cpu access size error")
                        CpuAccessSize.Byte -> throw Exception("not yet supported")
                        CpuAccessSize.Word -> throw Exception("not yet supported")
                        CpuAccessSize.DWord -> AMIOpcode(AMIOpcode.Operation.Mov,listOf(transformOp(stm.destination),AMIMemoryOperand(
                            tmpReg,
                            AMIRegisterOperand(mapToCpuReg(stm.index)).cpuRegister,
                            4,"0"
                        )),CpuAccessSize.DWord)
                    })
                }
                is AsmGetCogNumber -> {
                    //(cogptr - (hubptr+HubRelFirstCogRegOffset))>>PerCogSizeLog2
                    result.add(AMIOpcode(AMIOpcode.Operation.Mov,listOf(transformOp(stm.destination),AMIRegisterOperand(cogBaseCpuReg)),CpuAccessSize.QWord))
                    result.add(AMIOpcode(AMIOpcode.Operation.Sub,listOf(transformOp(stm.destination),AMIRegisterOperand(hubBaseCpuReg)),CpuAccessSize.QWord))
                    result.add(AMIOpcode(AMIOpcode.Operation.Sub,listOf(transformOp(stm.destination),AMILabelOperand("HubRelFirstCogRegOffset")),CpuAccessSize.DWord))
                    result.add(AMIOpcode(AMIOpcode.Operation.Shr,listOf(transformOp(stm.destination),AMILabelOperand("PerCogSizeLog2")),CpuAccessSize.DWord))
                }
                is AsmReserveTmpCpuReg -> {} //NOP
                else -> throw Exception("$stm not yet supported")
            }
        }
    }
    removeTrivialJumps(result)
    removeTrivialMovs(result)
    return result
}

fun removeTrivialJumps(lst:ArrayList<AMIEntity>) {
    var i=0

    fun isLabelFollowing(checkLbl:AMILocalLabelOperand?):Boolean {
        if (checkLbl == null)
            return false
        for (j in i+1 until lst.size) {
            if (lst[j] is AMIComment)
                continue
            val lbl = lst[j] as? AMILocalLabel ?: return false
            if (lbl.id == checkLbl.id)
                return true
        }
        return false
    }

    while (i<lst.size) {
        val stm = lst[i]
        if (stm is AMIOpcode && stm.operation == AMIOpcode.Operation.Jmp && isLabelFollowing(stm.operands[0] as? AMILocalLabelOperand)) {
            lst.removeAt(i)
            continue
        }
        ++i
    }
}

fun removeTrivialMovs(lst:ArrayList<AMIEntity>) {
    var i=0
    while (i<lst.size) {
        val stm = lst[i++]
        if (stm !is AMIOpcode || stm.operation != AMIOpcode.Operation.Mov)
            continue
        val p1 = stm.operands[0]
        val p2 = stm.operands[1]
        if (p1 is AMIRegisterOperand && p2 is AMIRegisterOperand && p1.cpuRegister == p2.cpuRegister) {
            --i
            lst.removeAt(i)
        }
    }
}

fun generateAsmCodeFromStream(stream:List<AMIEntity>,indent:String):String {
    val result = StringBuilder()
    for (e in stream) {
        result.append(indent+e.toAsmCode()+"\n")
    }
    return result.toString()
}