package assembler

import ilang.*
import ilang.Function

//analyisert, welche Register exakt einmal gelesen werden
//analysiert, ob ein register nach einem schreibzugriff nochmal gelesen wird

class AsmMemoryVariableState {
    val readOnce:HashSet<ILangVariable> //bezieht sich auf einen möglichen pfad, also a ? b+c : b+d ist b nur einmal gelesen!
    val readMulti:HashSet<ILangVariable>
    val written:HashSet<ILangVariable>
    val temporaryResultAndFlagsVariables:HashSet<ILangVariable>
    constructor(parent:AsmMemoryVariableState) {
        readOnce = HashSet(parent.readOnce)
        readMulti = HashSet(parent.readMulti)
        written = HashSet(parent.written)
        temporaryResultAndFlagsVariables = parent.temporaryResultAndFlagsVariables //das muss später nicht gemerged werden
    }
    constructor() {
        readOnce = HashSet()
        readMulti = HashSet()
        written = HashSet()
        temporaryResultAndFlagsVariables = HashSet()
    }

    fun mergeChildStates(children:List<AsmMemoryVariableState>) {
        for (c in children) {
            readOnce.addAll(c.readOnce)
            readMulti.addAll(c.readMulti)
            written.addAll(c.written)
        }
        readOnce.removeAll(readMulti)
        //resultAndFlagsVariables wird implizit gemerged
    }
}

fun asmPreAnalyseExpression(expr:Expression, varState: AsmMemoryVariableState) {
    when (expr) {
        is ConstantExpression,is GetCogNumberExpression,is ConstantJumpDestinationExpression -> {}
        is VariableExpression -> {
            if (expr.variable.direction == ILangVariableDirection.Register) {
                if (varState.written.contains(expr.variable))
                    throw Exception("register ${expr.variable.name} was read after it was written")
                if (varState.readOnce.contains(expr.variable)) {
                    varState.readMulti.add(expr.variable)
                    varState.readOnce.remove(expr.variable)
                }
                else if (!varState.readMulti.contains(expr.variable))
                    varState.readOnce.add(expr.variable)
            }
        }
        is AdditiveOrBitwiseExpression -> {
            asmPreAnalyseExpression(expr.left,varState)
            asmPreAnalyseExpression(expr.right,varState)
        }
        is ShiftExpression -> {
            asmPreAnalyseExpression(expr.left,varState)
            asmPreAnalyseExpression(expr.right,varState)
        }
        is TestBitExpression -> {
            asmPreAnalyseExpression(expr.expr,varState)
            asmPreAnalyseExpression(expr.bitNo,varState)
        }
        is CompareExpression -> {
            asmPreAnalyseExpression(expr.left,varState)
            asmPreAnalyseExpression(expr.right,varState)
        }
        is EqualityExpression -> {
            asmPreAnalyseExpression(expr.left,varState)
            asmPreAnalyseExpression(expr.right,varState)
        }
        is ArrayAccessExpression -> asmPreAnalyseExpression(expr.indexExpr,varState)
        is PartialArrayAccessExpression -> asmPreAnalyseExpression(expr.indexExpr,varState)
        is DecryptExpression -> asmPreAnalyseExpression(expr.expr,varState)
        is ParityExpression -> asmPreAnalyseExpression(expr.expr,varState)
        is ExtendToU64ShiftSignToUpperExpression -> asmPreAnalyseExpression(expr.expr,varState)
        is Cast64Expression -> asmPreAnalyseExpression(expr.expr,varState)
        is Lower32Expression -> asmPreAnalyseExpression(expr.expr,varState)
        is InvertExpression -> asmPreAnalyseExpression(expr.expr,varState)
        is LookupExpression -> asmPreAnalyseExpression(expr.index,varState)
        is JumpDestinationLookupExpression -> asmPreAnalyseExpression(expr.index,varState)
        is InRange32Expression -> asmPreAnalyseExpression(expr.expr,varState)
        is AddOrSubExpression -> {
            asmPreAnalyseExpression(expr.left,varState)
            asmPreAnalyseExpression(expr.right,varState)
            asmPreAnalyseExpression(expr.isSub,varState)
            asmPreAnalyseExpression(expr.extraCarry,varState)
        }
        is NegateWithFlagsExpression -> asmPreAnalyseExpression(expr.expr, varState)
        is MuxExpression -> {
            asmPreAnalyseExpression(expr.condition,varState)
            val trueState = AsmMemoryVariableState(varState)
            asmPreAnalyseExpression(expr.trueExpr,trueState)
            val falseState = AsmMemoryVariableState(varState)
            asmPreAnalyseExpression(expr.falseExpr,falseState)
            varState.mergeChildStates(listOf(trueState,falseState))
        }
        is ExtractFromCombinedOperation -> varState.temporaryResultAndFlagsVariables.add(expr.variable)
        is GetPllOfCogExpression -> asmPreAnalyseExpression(expr.cogNo, varState)
    }
}

fun asmPreAnalyseStmList(stmList:List<Statement>, varState:AsmMemoryVariableState) {
    for (stm in stmList) {
        when(stm) {
            is LetStatement -> {
                asmPreAnalyseExpression(stm.expr,varState)
                if (stm.destVar.direction == ILangVariableDirection.Register) {
                    if (varState.written.contains(stm.destVar))
                        throw Exception("register ${stm.destVar.name} written twice")
                    varState.written.add(stm.destVar)
                }
                if (stm.destVar.direction == ILangVariableDirection.Input)
                    throw Exception("input variable ${stm.destVar} written")
            }
            is LetArrayStatement -> {
                asmPreAnalyseExpression(stm.expr,varState)
                asmPreAnalyseExpression(stm.idx,varState)
            }
            is SwitchStatement -> {
                asmPreAnalyseExpression(stm.expr,varState)
                varState.mergeChildStates(stm.cases.map {
                    val childState = AsmMemoryVariableState(varState)
                    asmPreAnalyseStmList(it.body, childState)
                    childState
                })
            }
            is MacroStatement -> { }
        }
    }
}

fun asmPreanalyse(f:Function):AsmMemoryVariableState {
    val result = AsmMemoryVariableState()
    asmPreAnalyseStmList(f.body, result)
    return result
}