package assembler

import ilang.*

//Hier werden die Statements übersetzt

object AsmZFlagMacroRegisters {
    val destroyedRegisters = listOf(CpuRegister.RAX,CpuRegister.RCX,CpuRegister.RDX,CpuRegister.R8,CpuRegister.R9,CpuRegister.R10,CpuRegister.R11)
    fun getInputMapping(inputs:List<ILangVariable>):List<AsmInputVariable> {
        if (inputs.isEmpty())
            return listOf()
        val inputNames = inputs.map { it.name.removeSuffix("_input") }
        if (inputNames == listOf("instr_dst") || inputNames == listOf("instr_src"))
            return listOf(AsmInputVariable(inputs[0], CpuRegister.RDX, false))
        if (inputNames == listOf("instr_dst","outp_cog_alu_r"))
            return listOf(
                AsmInputVariable(inputs[0], CpuRegister.RDX, false),
                AsmInputVariable(inputs[1], CpuRegister.R8, false)
            )
        if (inputNames == listOf("instr_dst","instr_src","outp_cog_alu_r"))
            return listOf(
                AsmInputVariable(inputs[0], CpuRegister.RDX, false),
                AsmInputVariable(inputs[1], CpuRegister.R8, false),
                AsmInputVariable(inputs[2], CpuRegister.R9, false)
            )
        throw Exception("unknown inputs for zflag macro $inputNames")
    }
}

class AsmStatementTranslator(val state:AsmTranslatorState, val body:ArrayList<Statement>, val isRoot:Boolean) { //die ilang statements werden kopiert, da sie hier lokal modifiziert werden
    var idx=0
    private fun callMacro(macroName:String,inputs:List<Pair<ILangVariable,CpuRegister>>,macroInnerCpuRegs:List<CpuRegister>) {
        state.add(AsmComment("Inputs: ${inputs.map { "${it.second.toString().lowercase()}: ${it.first.name}" }}"))
        state.add(AsmComment("Modified: ${(inputs.map { it.second }+macroInnerCpuRegs).map { it.toString().lowercase() }}"))
        state.add(AsmCall(AsmLabelOperand(macroName),inputs.map {
            val tmpLocal = state.nextLocal()
            state.addMov(tmpLocal,state.operandReadOf(it.first),it.first.dataType == ILangDataType.Int64)
            Pair(tmpLocal,it.second)
        },macroInnerCpuRegs.map { Pair(state.nextLocal(),it) }))
    }
    init {
        while (idx<body.size) {
            val stm = body[idx++]
            if (isRoot)
                preLoadReadMultiVars(stm)
            state.add(AsmComment(stm.toString()))
            when(stm) {
                is LetStatement -> {
                    if (state.isDummyValueAndFlagVariable(stm.destVar))
                        AsmValueAndFlagExpressionTranslator(this, stm)
                    else
                        AsmDataExpressionTranslator.generate(state, state.operandWriteOf(stm.destVar), stm.expr)
                }
                is LetArrayStatement -> buildLetArrayStm(stm)
                is SwitchStatement -> {
                    if (stm.expr.dataType != ILangDataType.Bool)
                        throw Exception("switch not yet supported")
                    val trueStms = ArrayList(stm.matchingCase(1UL).body)
                    val falseStms = ArrayList(stm.matchingCase(0UL).body)
                    if (trueStms.isEmpty())
                        state.finishSectionWithSkipIf(stm.expr,true) { AsmStatementTranslator(it, falseStms, false) }
                    else if (falseStms.isEmpty())
                        state.finishSectionWithSkipIf(stm.expr,false) { AsmStatementTranslator(it, trueStms, false) }
                    else
                        state.finishSectionWithBranch(stm.expr,
                            { AsmStatementTranslator(it, trueStms, false) },
                            { AsmStatementTranslator(it, falseStms, false) }
                        )
                }
                is MacroStatement -> {
                    if (!stm.doFunctionCall)
                        throw Exception("only functions allowed, no macros")

                    if (stm.macroName == "macroCogPeripherals" && stm.inputs.isEmpty())
                        callMacro(stm.macroName,listOf(),listOf(CpuRegister.RAX,CpuRegister.RDX,CpuRegister.RCX))
                    else if (stm.macroName.startsWith("macroCogState4Z")) {
                        val inputs = AsmZFlagMacroRegisters.getInputMapping(stm.inputs)
                        val modRegs = ArrayList(inputs.map { it.cpuRegister }.toList())
                        for (inp in inputs)
                            modRegs.remove(inp.cpuRegister)
                        callMacro(stm.macroName,inputs.map { Pair(it.variable,it.cpuRegister) },modRegs)
                    }
                    else if (stm.macroName == "macroCog4SetControlRegisters" && stm.inputs.size == 2 && stm.inputs[0].name == "instr_dst" && stm.inputs[1].name == "outp_cog_alu_r") {
                        callMacro(stm.macroName, listOf(
                            Pair(stm.inputs[0],CpuRegister.RDX),
                            Pair(stm.inputs[1],CpuRegister.RCX)
                        ), listOf(CpuRegister.RAX))
                    }
                    else if ((stm.macroName == "hubClkBusMacroNack" || stm.macroName == "hubClkBusMacroAck") && stm.inputs.isEmpty())
                        callMacro(stm.macroName, listOf(), listOf(CpuRegister.RAX,CpuRegister.RDX,CpuRegister.RCX,CpuRegister.R8))
                    else if (stm.macroName == "hubClkBusMacroAckCustom" && stm.inputs.size == 2 && stm.inputs[0].name == "cog_e_loc" && stm.inputs[1].name == "bus_sel_idx_loc")
                        callMacro(stm.macroName, listOf(
                            Pair(stm.inputs[0],CpuRegister.RDX),
                            Pair(stm.inputs[1],CpuRegister.RCX)
                        ), listOf(CpuRegister.RAX,CpuRegister.R8))
                    else if (stm.macroName == "hubClkBusMacroAckStartCog" && stm.inputs.size == 3 && stm.inputs[0].name == "cog_e_loc" && stm.inputs[1].name == "bus_sel_idx_loc" && stm.inputs[2].name == "num_loc")
                        callMacro(stm.macroName, listOf(
                            Pair(stm.inputs[0],CpuRegister.RDX),
                            Pair(stm.inputs[1],CpuRegister.RCX),
                            Pair(stm.inputs[2],CpuRegister.R8),
                        ), listOf(CpuRegister.RAX))
                    else if ((stm.macroName == "macroCogReset" || stm.macroName == "cogUpdateAllDynamicFunctions") && stm.inputs.isEmpty())
                        callMacro(stm.macroName, listOf(), listOf(CpuRegister.RAX,CpuRegister.RCX,CpuRegister.RDX,CpuRegister.R8,CpuRegister.R9))
                    else
                        throw Exception("unknown macro ${stm.macroName} ${stm.inputs}")
                }
            }
        }
    }

    private fun preLoadReadMultiVars(stm:Statement) {
        //Das laden findet nur im Root Zweig statt
        for (v in state.getReadMultiVarsThatNeedToBeLoaded(stm)) {
            state.add(AsmComment("preload ${v.name}"))
            val preLoadedOp = state.variablePreLoaded(v)
            val loadFrom = state.operandWriteOf(v) //hier operandWriteOf nehmen, damit die memory location benutzt wird
            state.addMov(preLoadedOp,loadFrom,v.dataType == ILangDataType.Int64)
        }
    }

    private fun buildLetArrayStm(stm:LetArrayStatement) {
        if (stm.destArrayVar.dataType != ILangDataType.Int32)
            throw Exception("only Int32 supported for arrays")
        val idxVar = if (stm.idx is VariableExpression) {
            state.forceLocal(state.operandReadOf(stm.idx))
        }
        else {
            val tmpI = state.nextLocal()
            AsmDataExpressionTranslator.generate(state, tmpI,stm.idx)
            tmpI
        }
        when (stm.accessSize) {
            CpuAccessSize.DWord -> AsmDataExpressionTranslator.generate(state, AsmArrayOperand(stm.destArrayVar, idxVar, 4, 0), stm.expr)
            CpuAccessSize.Word,CpuAccessSize.Byte -> {
                val tmpValue = state.nextLocal()
                AsmDataExpressionTranslator.generate(state, tmpValue, stm.expr)
                state.add(AsmMovZXOp(AsmArrayOperand(stm.destArrayVar, idxVar, 4, stm.byteOffset),tmpValue,stm.accessSize))
            }
            CpuAccessSize.QWord,CpuAccessSize.None -> throw Exception("not supported")
        }
    }
}