package assembler

import ilang.*
import special.dependendVars
import java.lang.Exception
import java.lang.StringBuilder

class AsmInputVariable(val variable:ILangVariable, val cpuRegister: CpuRegister, val keepValue:Boolean)
class AsmCodeForFunction(val name:String, val code:List<AMIEntity>, val inputs:List<AsmInputVariable>, val modifiedRegisters:Set<CpuRegister>, val hasReturnValue:Boolean=false)

class AsmTransformer(name:String, val memVarsReadMulti:Set<ILangVariable>, otherVars:Set<ILangVariable>, inputs:List<AsmInputVariable>, val tempFlagVars:Set<ILangVariable>, body:List<Statement>, hubRegs:ILangVariableSet, cogRegs:ILangVariableSet, val lookupTables:ArrayList<AsmAbstractLookupTable>) {
    var nextAnonymousLocal=1
    val varToOperandMapRead = HashMap<ILangVariable,AsmOperand>()
    val varToOperandMapWrite = HashMap<ILangVariable,AsmOperand>()
    var sectionCounter=0
    private val allSections = ArrayList<AsmSection>()
    val notYetLoadedReadMultiVars = HashSet(memVarsReadMulti)
    val resultFun:AsmCodeForFunction
    init {
        //otherVars sind lokale variablen oder register die nur einmal gelesen werden
        for (v in otherVars) {
            val op = if (v.direction == ILangVariableDirection.Register) AsmMemOperand(v) else AsmLocalOperand("\$${v.name}",v)
            varToOperandMapRead[v] = op
            varToOperandMapWrite[v] = op
        }
        for (v in memVarsReadMulti) {
            if (v.direction != ILangVariableDirection.Register)
                throw Exception("this var is not a register")
            varToOperandMapWrite[v] = AsmMemOperand(v)
        }
    }
    fun reserveSection():AsmSection {
        return AsmSection(sectionCounter++)
    }
    fun placeSection(section:AsmSection) {
        allSections.add(section)
    }
    fun reserveAndPlaceSection():AsmSection {
        val s = reserveSection()
        placeSection(s)
        return s
    }

    fun serialize():String {
        val res = StringBuilder()
        for (s in allSections) {
            res.append("$s:\n")
            for (c in s.body)
                res.append("    $c\n")
            res.append("    ${s.endOfSection}\n")
        }
        return res.toString()
    }

    private fun getRdLocal(localVar:ILangVariable):AsmLocalOperand = varToOperandMapRead[localVar] as? AsmLocalOperand ?: throw Exception("input $localVar not found in varMap")

    init {
        val rootSection = reserveAndPlaceSection()
        val state = AsmTranslatorState(this,rootSection,lookupTables)
        AsmStatementTranslator(state,ArrayList(body), true)
        state.finishSectionWithRet()

        val localLiveStatus = HashMap<AsmLocalOperand,AsmLocalLiveStatus>()
        //für die inputs müssen die register von anfang an getracked werden, da diese außerhalb gesetzt wurden
        for (inp in inputs)
            localLiveStatus[getRdLocal(inp.variable)] = AsmLocalLiveStatus()
        val keepVars = inputs.filter { it.keepValue }.map { getRdLocal(it.variable) }
        getAsmLocalVariableLiveStatus(rootSection,keepVars,localLiveStatus,HashMap())

        val liveAtSameTime = findAsmLocalsLiveAtSameTime(allSections,localLiveStatus)
        val cpuRegConstraints = getAsmLocalsConstraints(allSections)
        for (inp in inputs) {
            val inpOp = getRdLocal(inp.variable)
            if (cpuRegConstraints.containsKey(inpOp))
                throw Exception("duplicate constraints for $inpOp")
            cpuRegConstraints[inpOp] = inp.cpuRegister
        }
        val varMapOrder = asmRegisterMapOrder(allSections, inputs.map { getRdLocal(it.variable) })
        val mappedRegisters = mapAsmCpuRegisters(liveAtSameTime,cpuRegConstraints,varMapOrder)
        resultFun=AsmCodeForFunction(name, generateAMIStream(allSections,mappedRegisters,hubRegs,cogRegs), inputs, mappedRegisters.values.toSet().subtract(keepVars.map { mappedRegisters[it]?:throw Exception("no mapped register found for $it") }))
    }
}



class AsmTranslatorState(private val master: AsmTransformer, private var currentSection: AsmSection, private val lookupTables:ArrayList<AsmAbstractLookupTable>) {
    fun nextLocal():AsmLocalOperand = AsmLocalOperand("\$${master.nextAnonymousLocal++}",null)
    fun nextLocalWithDummyVar(dataType: ILangDataType):Pair<AsmLocalOperand,ILangVariable> {
        val op = nextLocal()
        val v=ILangVariable(op.name,dataType,ILangVariableDirection.Local)
        master.varToOperandMapRead[v] = op
        master.varToOperandMapWrite[v] = op
        return Pair(op,v)
    }
    fun variablePreLoaded(v:ILangVariable):AsmLocalOperand {
        if (!master.notYetLoadedReadMultiVars.remove(v))
            throw Exception("this variable is not on the list of not yet preloaded: $v")
        if (master.varToOperandMapRead.containsKey(v))
            throw Exception("this variable is already in the operand map: $v")
        val localOp = AsmLocalOperand("\$${v.name}",v)
        master.varToOperandMapRead[v] = localOp
        return localOp
    }
    fun operandReadOf(variable: ILangVariable):AsmOperand = master.varToOperandMapRead[variable] ?: throw Exception("unknown var ${variable.name}")
    fun operandReadOf(expr: VariableExpression):AsmOperand = operandReadOf(expr.variable)

    fun operandWriteOf(variable: ILangVariable):AsmOperand = master.varToOperandMapWrite[variable] ?: throw Exception("unknown var ${variable.name}")
    fun operandWriteOf(expr: VariableExpression):AsmOperand = operandReadOf(expr.variable)

    fun operandInplaceWritable(variable: ILangVariable):Boolean = !master.memVarsReadMulti.contains(variable)
    fun operandInplaceWritable(expr: VariableExpression):Boolean = operandInplaceWritable(expr.variable)

    fun operandOf(constValue:ULong)=AsmConstantOperand(constValue)
    fun operandOf(constValue:Boolean)=AsmConstantOperand(if (constValue) 1UL else 0UL)
    fun operandOf(expr: ConstantExpression)=AsmConstantOperand(expr.value)
    fun isDummyValueAndFlagVariable(v:ILangVariable):Boolean = master.tempFlagVars.contains(v)
    fun createLocalCopy(asmOperand: AsmOperand, condition:Boolean):AsmOperand {
        if (!condition || !asmOperand.isMemory())
            return asmOperand
        val tmp = nextLocal()
        addMov(tmp,asmOperand,asmOperand.memoryIs64Bit())
        return tmp
    }
    fun forceLocal(asmOperand: AsmOperand):AsmLocalOperand {
        if (asmOperand is AsmLocalOperand)
            return asmOperand
        val tmp = nextLocal()
        addMov(tmp,asmOperand,asmOperand.memoryIs64Bit())
        return tmp
    }
    fun duplicateLocal(asmOperand: AsmOperand):AsmLocalOperand {
        val tmp = nextLocal()
        addMov(tmp,asmOperand,asmOperand.memoryIs64Bit())
        return tmp
    }
    fun add(stm:AsmStatement) {
        if (currentSection.endOfSection != null)
            throw Exception("cannot add asm stms to finished section")
        currentSection.add(stm)
    }
    fun addBinOp(operation: AsmBinOp.Operation, destination:AsmOperand, source:AsmOperand, use64BitMode:Boolean) = add(AsmBinOp(operation, destination, source, use64BitMode))
    fun addMov(destination:AsmOperand, source:AsmOperand, use64BitMode:Boolean) = add(AsmMovOp(destination, source, use64BitMode))

    fun addOrGetNumericLookupTable(values:List<ULong>):AsmAbstractLookupTable {
        for (t in lookupTables) {
            if (t is AsmNumericLookupTable && t.values == values)
                return t
        }
        //val maxValue = values.maxOrNull() ?: throw Exception("lookup table was empty")
        val newTbl = AsmNumericLookupTable(values, AsmLabelOperand("lookup_table_${lookupTables.size}"), CpuAccessSize.DWord)
        lookupTables.add(newTbl)
        return newTbl
    }

    fun addOrGetJumpLookupTable(values:List<JumpDestinationLookupExpression.Entry>):AsmAbstractLookupTable {
        for (t in lookupTables) {
            if (t is AsmJumpLookupTable && t.values == values)
                return t
        }
        //val maxValue = values.maxOrNull() ?: throw Exception("lookup table was empty")
        val newTbl = AsmJumpLookupTable(values, AsmLabelOperand("lookup_table_${lookupTables.size}"))
        lookupTables.add(newTbl)
        return newTbl
    }

    private fun finishSection(stop:EndOfAsmSection) {
        if (currentSection.endOfSection != null)
            throw Exception("cannot finish a section twice")
        currentSection.endOfSection = stop
    }
    fun finishSectionWithJump(destination: AsmSection) = finishSection(JumpToAsmSection(destination))
    fun finishSectionWithSkipIf(conditionExpr: Expression, skipIfTrue:Boolean, cbNotSkipped:(notSkippedState:AsmTranslatorState)->Unit) {
        val notSkippedState = AsmTranslatorState(master, master.reserveSection(), lookupTables)

        val initialNotSkippedSection = notSkippedState.currentSection
        val sectionAfterBranches = master.reserveSection()
        generateFastEvalBranch(conditionExpr, !skipIfTrue, sectionAfterBranches, initialNotSkippedSection)
        master.placeSection(initialNotSkippedSection)
        cbNotSkipped(notSkippedState)
        master.placeSection(sectionAfterBranches)
        notSkippedState.finishSectionWithJump(sectionAfterBranches)
        currentSection = sectionAfterBranches
    }
    fun finishSectionWithBranch(conditionExpr:Expression, cbTrue:(trueState:AsmTranslatorState)->Unit, cbFalse:(falseState:AsmTranslatorState)->Unit) {
        val trueChildState = AsmTranslatorState(master, master.reserveSection(), lookupTables)
        val falseChildState = AsmTranslatorState(master, master.reserveSection(), lookupTables)

        val initialTrueSection = trueChildState.currentSection
        val initialFalseSection = falseChildState.currentSection

        generateFastEvalBranch(conditionExpr, false, initialTrueSection, initialFalseSection)
        master.placeSection(initialFalseSection)
        cbFalse(falseChildState)
        master.placeSection(initialTrueSection)
        cbTrue(trueChildState)

        val sectionAfterBranches = master.reserveAndPlaceSection()
        trueChildState.finishSectionWithJump(sectionAfterBranches)
        falseChildState.finishSectionWithJump(sectionAfterBranches)
        currentSection = sectionAfterBranches
    }
    private fun generateFastEvalBranch(conditionExpr: Expression, invert:Boolean, trueSection: AsmSection, falseSection: AsmSection) {
        if (conditionExpr is AdditiveOrBitwiseExpression && !conditionExpr.withFlags && conditionExpr.operation == AdditiveOrBitwiseExpression.Operation.And) {
            val testFurtherState = AsmTranslatorState(master, master.reserveSection(), lookupTables)
            val testFurtherInitialSection = testFurtherState.currentSection
            if (invert) {
                //nand, wenn links schon falsch, dann direkt abbrechen
                generateFastEvalBranch(conditionExpr.left,true,trueSection,testFurtherInitialSection)
                master.placeSection(testFurtherInitialSection)
                currentSection = testFurtherState.currentSection
                generateFastEvalBranch(conditionExpr.right,true,trueSection,falseSection)
            }
            else {
                //and, wenn links schon falsch, dann direkt abbrechen
                generateFastEvalBranch(conditionExpr.left,false,testFurtherInitialSection, falseSection)
                master.placeSection(testFurtherInitialSection)
                currentSection = testFurtherState.currentSection
                generateFastEvalBranch(conditionExpr.right,false,trueSection,falseSection)
            }
            return
        }
        if (conditionExpr is AdditiveOrBitwiseExpression && !conditionExpr.withFlags && conditionExpr.operation == AdditiveOrBitwiseExpression.Operation.Or) {
            val testFurtherState = AsmTranslatorState(master, master.reserveSection(), lookupTables)
            val testFurtherInitialSection = testFurtherState.currentSection
            if (invert) {
                //nor, wenn links schon richtig, dann direkt abbrechen
                generateFastEvalBranch(conditionExpr.left,false,falseSection,testFurtherInitialSection)
                master.placeSection(testFurtherInitialSection)
                currentSection = testFurtherState.currentSection
                generateFastEvalBranch(conditionExpr.right,true,trueSection,falseSection)
            }
            else {
                //or, wenn links schon richtig, dann direkt abbrechen
                generateFastEvalBranch(conditionExpr.left,true,testFurtherInitialSection,trueSection)
                master.placeSection(testFurtherInitialSection)
                currentSection = testFurtherState.currentSection
                generateFastEvalBranch(conditionExpr.right,false,trueSection,falseSection)
            }
            return
        }
        val c = AsmConditionTranslator.generate(this, conditionExpr)
        finishSection(BranchToAsmSection(if (invert) c.logicInverted() else c, trueSection, falseSection))
    }
    fun finishSectionWithRet() = finishSection(AsmReturnToCaller())
    fun getReadMultiVarsThatNeedToBeLoaded(stm:Statement):List<ILangVariable> {
        return master.notYetLoadedReadMultiVars.intersect(dependendVars(stm).vars).sortedBy { it.name }
    }
    fun calculateInsideLocalIfDestIsMemLocation(dest:AsmOperand, cb:(tmpDest:AsmLocalOperand)->Unit) = if(dest.isMemory()) {
            val tmpDest = nextLocal()
            cb(tmpDest)
            addMov(dest, tmpDest, dest.memoryIs64Bit())
        }
        else if (dest is AsmLocalOperand)
            cb(dest)
        else
            throw Exception("illegal operand")
}
