package assembler

import ilang.*
import java.lang.Exception

//Hier werden spezielle Ausdrücke umgewandelt, bei denen sowohl das Ergebnis als auch die Flags weitergereicht werden

class AsmValueAndFlagExpressionTranslator(private val stmCls:AsmStatementTranslator, baseStm: LetStatement) {
    private val dummyFlagAndResultVar = baseStm.destVar
    private var resultOperand:AsmOperand?=null
    private var zFlagOperand:AsmOperand?=null
    private var cFlagOperand:AsmOperand?=null
    private var zFlagInverted=false
    private var cFlagInverted=false
    private fun resultRequested(resultType: ExtractFromCombinedOperation.ExtractType, dest: AsmOperand, inverted:Boolean) {
        when(resultType) {
            ExtractFromCombinedOperation.ExtractType.Result -> {
                if (resultOperand != null)
                    throw Exception("result operand requested twice")
                resultOperand = dest
                if (inverted)
                    throw Exception("inverted result not supported")
            }
            ExtractFromCombinedOperation.ExtractType.ZFlag -> {
                if (zFlagOperand != null)
                    throw Exception("zflag operand requested twice")
                zFlagOperand = dest
                zFlagInverted = inverted
            }
            ExtractFromCombinedOperation.ExtractType.CFlag -> {
                if (cFlagOperand != null)
                    throw Exception("cflag operand requested twice")
                cFlagOperand = dest
                cFlagInverted = inverted
            }
        }
    }

    private fun analyseSimpleFollowingStatements(state:AsmTranslatorState) {
        //Direkt folgende Let statements, die das Ergebnis oder die Flags direkt in einer Variable zwischenspeichern
        //können hier direkt behandelt und damit übersprungen werden
        while (stmCls.idx<stmCls.body.size) {
            val tmpStm = stmCls.body[stmCls.idx] as? LetStatement ?: break
            //direktes flag oder result
            if (tmpStm.expr is ExtractFromCombinedOperation && tmpStm.expr.variable === dummyFlagAndResultVar) {
                state.add(AsmComment("> $tmpStm"))
                ++stmCls.idx
                resultRequested(tmpStm.expr.extractType,state.operandWriteOf(tmpStm.destVar), false)
                continue
            }
            //invertiertes flag
            if (tmpStm.expr is InvertExpression && tmpStm.expr.dataType == ILangDataType.Bool && tmpStm.expr.expr is ExtractFromCombinedOperation && tmpStm.expr.expr.variable === dummyFlagAndResultVar) {
                state.add(AsmComment("> $tmpStm"))
                ++stmCls.idx
                resultRequested(tmpStm.expr.expr.extractType,state.operandWriteOf(tmpStm.destVar), true)
                continue
            }
            break
        }
    }

    private fun modifiyExpressionIfDependsOnResult(state:AsmTranslatorState, rootExpression: Expression, checkForInverted:Boolean) : Expression = rootExpression.map { expr ->
        if (!checkForInverted && expr is ExtractFromCombinedOperation && expr.variable === dummyFlagAndResultVar) { //direkter wert oder flag
            val dt = if (expr.extractType == ExtractFromCombinedOperation.ExtractType.Result) ILangDataType.Int32 else ILangDataType.Bool
            val (tmpOperand, tmpVar) = state.nextLocalWithDummyVar(dt)
            resultRequested(expr.extractType, tmpOperand, false)
            VariableExpression(tmpVar)
        }
        else if (checkForInverted && expr is InvertExpression && expr.dataType == ILangDataType.Bool && expr.expr is ExtractFromCombinedOperation && expr.expr.variable === dummyFlagAndResultVar) {
            val (tmpOperand, tmpVar) = state.nextLocalWithDummyVar(ILangDataType.Bool)
            resultRequested(expr.expr.extractType, tmpOperand, true)
            VariableExpression(tmpVar)
        }
        else
            expr
    }

    private fun modifiyExpressionIfDependsOnResult(state:AsmTranslatorState, rootExpression: Expression):Expression =
        modifiyExpressionIfDependsOnResult(state,modifiyExpressionIfDependsOnResult(state,rootExpression,true),false)

    private fun analyseComplexStatements(state:AsmTranslatorState) {
        //folgen komplexre Ausdrücke, so wird das ergebnis erstmal in einer lokalen variable gespeichert
        //der Ausdruck wird hier dann modifiziert und dass stm erst später abgearbeitet (als herkörmliches stm)
        for (lookaheadIdx in stmCls.idx until stmCls.body.size) {
            val stm = stmCls.body[lookaheadIdx]
            stmCls.body[lookaheadIdx] = when(stm) {
                is LetStatement -> stm.modified(modifiyExpressionIfDependsOnResult(state,stm.expr))
                is LetArrayStatement -> stm.modified(modifiyExpressionIfDependsOnResult(state,stm.idx),modifiyExpressionIfDependsOnResult(state,stm.expr))
                is SwitchStatement -> stm.modified(modifiyExpressionIfDependsOnResult(state,stm.expr),stm.cases) //innere elemente werden nicht durchsucht, es ist verboten, dort zu lesen
                is MacroStatement -> stm
            }
        }
    }

    private fun addOperation(destState:AsmTranslatorState, opCode:AsmResultAndFlagOp.Operation, dest:AsmLocalOperand, right:AsmOperand, carryIn:AsmOperand) {
        destState.add(AsmResultAndFlagOp(opCode,dest,right,carryIn,cFlagOperand,cFlagInverted,zFlagOperand,zFlagInverted))
    }

    private fun addOperation(destState:AsmTranslatorState, opCode:AsmResultAndFlagOp.Operation, left:Expression, right:Expression, extraCarry:Expression) {
        destState.calculateInsideLocalIfDestIsMemLocation(resultOperand ?: destState.nextLocal()) {
            tmpDest ->
            //linke seite auswerten (nach tmpDest)
            AsmDataExpressionTranslator.generate(destState, tmpDest, left)

            val rightOp =
                if (right is VariableExpression) destState.operandReadOf(right)
                else if (right is ConstantExpression) destState.operandOf(right)
                else {
                    val tmpR = destState.nextLocal()
                    AsmDataExpressionTranslator.generate(destState, tmpR, right)
                    tmpR
                }
            val extraCarryOp = if (extraCarry is ConstantExpression) destState.operandOf(extraCarry)
                else {
                    val tmpC = destState.nextLocal()
                    AsmDataExpressionTranslator.generate(destState,tmpC,extraCarry)
                    //es wird zwingend eine kopie angeführt, da das register sich ändert
                    destState.duplicateLocal(tmpC)
                }
            addOperation(destState, opCode, tmpDest, rightOp, extraCarryOp)
        }
    }

    private fun getOperationForAddSub(signed:Boolean, isSub:Boolean):AsmResultAndFlagOp.Operation {
        if (signed)
            return if (isSub) AsmResultAndFlagOp.Operation.SubS else AsmResultAndFlagOp.Operation.AddS
        return if (isSub) AsmResultAndFlagOp.Operation.SubU else AsmResultAndFlagOp.Operation.AddU
    }

    private fun buildOperation(state:AsmTranslatorState, expr:Expression) {
        //Jetzt wurde ermittelt, welche Ergebnisse benötigt werden, es kann die Operation gebaut werden
        when(expr) {
            is AddOrSubExpression -> {
                if (expr.isSub is ConstantExpression)
                    addOperation(state, getOperationForAddSub(expr.signed, expr.isSub.value != 0UL), expr.left, expr.right, expr.extraCarry)
                else {
                    state.finishSectionWithBranch(expr.isSub,
                        { addOperation(it, getOperationForAddSub(expr.signed, true), expr.left, expr.right, expr.extraCarry) },
                        { addOperation(it, getOperationForAddSub(expr.signed, false), expr.left, expr.right, expr.extraCarry) }
                    )
                }
            }
            is AdditiveOrBitwiseExpression -> {
                if (!expr.withFlags)
                    throw Exception("bitwise op was without flags")
                addOperation(state, when(expr.operation) {
                    AdditiveOrBitwiseExpression.Operation.And -> AsmResultAndFlagOp.Operation.And
                    AdditiveOrBitwiseExpression.Operation.Or -> AsmResultAndFlagOp.Operation.Or
                    AdditiveOrBitwiseExpression.Operation.Xor -> AsmResultAndFlagOp.Operation.Xor
                    AdditiveOrBitwiseExpression.Operation.Add,AdditiveOrBitwiseExpression.Operation.Sub -> throw Exception("unsupported")
                }, expr.left, expr.right, ConstantExpression(false))
            }
            is ShiftExpression -> {
                if (!expr.withFlags)
                    throw Exception("shift op was without flags")
                addOperation(state, if (expr.shiftLeft) AsmResultAndFlagOp.Operation.Shl else AsmResultAndFlagOp.Operation.Shr, expr.left, expr.right, ConstantExpression(false))
            }
            is NegateWithFlagsExpression ->
                addOperation(state, AsmResultAndFlagOp.Operation.Neg, expr.expr, ConstantExpression(0UL,ILangDataType.Int32), ConstantExpression(false))
            else -> throw Exception("illegal result and flag operation: ${expr}")
        }
    }

    init {
        analyseSimpleFollowingStatements(stmCls.state)
        analyseComplexStatements(stmCls.state)
        buildOperation(stmCls.state, baseStm.expr)
    }
}
