package cpp

import assembler.CpuAccessSize
import ilang.*
import ilang.Function
import java.lang.StringBuilder

fun generateCppCode(regSets:List<ILangVariableSet>, function: List<Function>):String {
    //val builder=StringBuilder()
    fun dataTypeStr(dataType: ILangDataType) = when(dataType) {
        ILangDataType.Bool -> "bool"
        ILangDataType.Int32 -> "quint32"
        ILangDataType.Int64 -> "quint64"
    }
    fun numericValue(dataType: ILangDataType, value:ULong) = when(dataType) {
        ILangDataType.Bool -> if (value != 0UL) "true" else "false"
        ILangDataType.Int32 -> "0x"+value.toString(16)
        ILangDataType.Int64 -> "0x"+value.toString(16)
    }
    val lookupTables = ArrayList<List<ULong>>()
    fun lookupTableName(idx:Int):String = "lookuptable$idx"
    fun regSetName(rs:ILangVariableSet, type:Boolean) = if (type) "rs_${rs.name}" else "reg${rs.name}"
    fun generateRegisterSet(indent:String, set:ILangVariableSet, builder:StringBuilder) {
        fun orderOfDataType(dt:ILangDataType) = when(dt) {
            ILangDataType.Int64 -> 0
            ILangDataType.Int32 -> 1
            ILangDataType.Bool -> 2
        }
        val varList = set.getVariableList().sortedWith{ a,b ->
            val ad = orderOfDataType(a.dataType)
            val bd = orderOfDataType(b.dataType)
            if (ad<bd) -1 else if (ad>bd) 1 else {
                if (a.name<b.name) -1 else if (a.name>b.name) 1 else 0
            }
        }
        builder.append("${indent}struct ${regSetName(set, true)} {\n")
        //Konstruktor
        builder.append("${indent}    ${regSetName(set, true)}():"+varList.map { "${it.name}(${numericValue(it.dataType,0UL)})" }.joinToString(",")+" {\n")
        for (v in set.getArrayVariableList()) {
            builder.append("${indent}        for (int locIterator=0; locIterator<${v.size}; locIterator++)\n")
            builder.append("${indent}            ${v.name}[locIterator] = ${numericValue(v.dataType, 0UL)};\n")
        }
        builder.append("${indent}    }\n")

        builder.append("${indent}    void serialize(QDataStream& stream) const {\n")
        for (v in varList)
            builder.append("${indent}        stream << ${v.name};\n")
        for (v in set.getArrayVariableList()) {
            builder.append("${indent}        for (int locIterator=0; locIterator<${v.size}; locIterator++)\n")
            builder.append("${indent}            stream << ${v.name}[locIterator];\n")
        }
        builder.append("${indent}    }\n")

        builder.append("${indent}    void deserialize(QDataStream& stream) {\n")
        for (v in varList)
            builder.append("${indent}        stream >> ${v.name};\n")
        for (v in set.getArrayVariableList()) {
            builder.append("${indent}        for (int locIterator=0; locIterator<${v.size}; locIterator++)\n")
            builder.append("${indent}            stream >> ${v.name}[locIterator];\n")
        }
        builder.append("${indent}    }\n")

        builder.append("${indent}    QList<QPair<QString,QString> > listInnerVars() const {\n")
        builder.append("${indent}        QList<QPair<QString,QString> > resultList;\n")
        for (v in varList)
            if (v.dataType == ILangDataType.Bool)
                builder.append("${indent}        resultList << QPair<QString,QString>(\"${v.name}\",${v.name} ? \"TRUE\" : \"false\");\n")
            else
                builder.append("${indent}        resultList << QPair<QString,QString>(\"${v.name}\",\"0x\"+QString::number(${v.name},16));\n")
        builder.append("${indent}        return resultList;\n")
        builder.append("${indent}    }\n")

        for (v in varList)
            builder.append("${indent}    ${dataTypeStr(v.dataType)} ${v.name};\n")
        for (v in set.getArrayVariableList())
            builder.append("${indent}    ${dataTypeStr(v.dataType)} ${v.name}[${v.size}];\n")

        //hack
        if (set.name == "hub")
            builder.append("$indent    rs_cog *cogs[8];\n")
        else if (set.name == "cog")
            builder.append("$indent    quint32 cogId;\n")

        builder.append("${indent}};\n\n")
    }
    fun generateFunction(indent:String, f:Function, funcBuilder:StringBuilder) {
        val regs = f.registerVarSet()
        val locals = f.localVarSet()
        val usedRegsSets = HashSet<ILangVariableSet>()

        fun varName(v:ILangVariable):String {
            for (set in regs.sets) {
                if (set.contains(v)) {
                    usedRegsSets.add(set)
                    return regSetName(set, false) + ".${v.name}"
                }
            }
            if (locals.contains(v))
                return v.name
            throw Exception("var $v not found")
        }
        fun arrayVarName(v:ILangArrayVariable):String {
            for (set in regs.sets) {
                if (set.contains(v))
                    return regSetName(set,false)+".${v.name}"
            }
            if (locals.contains(v))
                return v.name
            throw Exception("var $v not found")
        }
        fun generateExpression(expr: Expression):String = when(expr) {
            is ConstantExpression -> numericValue(expr.dataType,expr.value)
            is ConstantJumpDestinationExpression -> numericValue(expr.dataType,expr.value)
            is VariableExpression -> varName(expr.variable)
            is AdditiveOrBitwiseExpression -> {
                val l = generateExpression(expr.left)
                val r = generateExpression(expr.right)
                when(expr.operation) {
                    AdditiveOrBitwiseExpression.Operation.Add -> "($l + $r)"
                    AdditiveOrBitwiseExpression.Operation.Sub -> "($l - $r)"
                    AdditiveOrBitwiseExpression.Operation.And -> if (expr.dataType == ILangDataType.Bool) "($l && $r)" else "($l & $r)"
                    AdditiveOrBitwiseExpression.Operation.Or -> if (expr.dataType == ILangDataType.Bool) "($l || $r)" else "($l | $r)"
                    AdditiveOrBitwiseExpression.Operation.Xor -> if (expr.dataType == ILangDataType.Bool) "($l != $r)" else "($l ^ $r)"
                }
            }
            is EqualityExpression -> {
                val l = generateExpression(expr.left)
                val r = generateExpression(expr.right)
                if (expr.isEq) "($l == $r)" else "($l != $r)"
            }
            is ShiftExpression -> {
                val l = generateExpression(expr.left)
                val r = generateExpression(expr.right)
                if (expr.shiftLeft) "($l << $r)" else "($l >> $r)"
            }
            is TestBitExpression -> {
                val e = generateExpression(expr.expr)
                val bitNo = generateExpression(expr.bitNo)
                val one = numericValue(expr.expr.dataType,1UL)
                "((($e >> $bitNo) & $one) == $one)"
            }
            is InRange32Expression -> {
                val e = generateExpression(expr.expr)
                "($e >= ${numericValue(expr.expr.dataType,expr.lower)} && $e <= ${numericValue(expr.expr.dataType,expr.upper)})"
            }
            is ArrayAccessExpression -> {
                val idx = generateExpression(expr.indexExpr)
                "${arrayVarName(expr.arrayVariable)}[$idx]"
            }
            is PartialArrayAccessExpression -> {
                val idx = generateExpression(expr.indexExpr)
                val accessShifted = "(${arrayVarName(expr.arrayVariable)}[$idx]>>${expr.byteOffset*8})"
                if (expr.isByte) "($accessShifted & 0xFF)" else "($accessShifted & 0xFFFF)"
            }
            is MuxExpression -> {
                val cond = generateExpression(expr.condition)
                val trueExpr = generateExpression(expr.trueExpr)
                val falseExpr = generateExpression(expr.falseExpr)
                "($cond ? $trueExpr : $falseExpr)"
            }
            is CompareExpression -> {
                val l = generateExpression(expr.left)
                val r = generateExpression(expr.right)
                when (expr.operation) {
                    CompareExpression.Operation.Le -> "($l < $r)"
                    CompareExpression.Operation.Leq -> "($l <= $r)"
                    CompareExpression.Operation.Ge -> "($l > $r)"
                    CompareExpression.Operation.Geq -> "($l >= $r)"
                    CompareExpression.Operation.SgnLe -> if (expr.left.dataType == ILangDataType.Int32) "(qint32($l) < qint32($r))" else throw Exception("sgnle only supported for 32 bit")
                }
            }
            is InvertExpression -> {
                val e = generateExpression(expr.expr)
                if (expr.dataType == ILangDataType.Bool) "(!$e)" else "(~$e)"
            }
            is DecryptExpression -> "decrypt(${generateExpression(expr.expr)})"
            is ParityExpression -> "getParity(${generateExpression(expr.expr)})"
            is ExtendToU64ShiftSignToUpperExpression -> "extendToU64ShiftSignToUpper(${generateExpression(expr.expr)})"
            is Cast64Expression -> "quint64(${generateExpression(expr.expr)})"
            is Lower32Expression -> "quint32(${generateExpression(expr.expr)})"
            is LookupExpression -> {
                var lookupTblIdx = lookupTables.indexOf(expr.values)
                if (lookupTblIdx<0) {
                    lookupTblIdx = lookupTables.size
                    lookupTables.add(expr.values)
                }
                "${lookupTableName(lookupTblIdx)}[${generateExpression(expr.index)}]"
            }
            is JumpDestinationLookupExpression -> {
                val onlyIndices = expr.values.map { it.value }
                var lookupTblIdx = lookupTables.indexOf(onlyIndices)
                if (lookupTblIdx<0) {
                    lookupTblIdx = lookupTables.size
                    lookupTables.add(onlyIndices)
                }
                "${lookupTableName(lookupTblIdx)}[${generateExpression(expr.index)}]"
            }
            is ExtractFromCombinedOperation -> {
                when(expr.operationType) {
                    ExtractFromCombinedOperation.OperationType.BitOp -> {
                        when(expr.extractType) {
                            ExtractFromCombinedOperation.ExtractType.Result -> varName(expr.variable)
                            ExtractFromCombinedOperation.ExtractType.ZFlag -> "(${varName(expr.variable)} == 0UL)"
                            ExtractFromCombinedOperation.ExtractType.CFlag -> throw Exception("unsupported")
                        }
                    }
                    ExtractFromCombinedOperation.OperationType.UnsignedAddOrSub, ExtractFromCombinedOperation.OperationType.SignedAddOrSub -> expr.functionName()+"("+varName(expr.variable)+")"

                }
            }
            is AddOrSubExpression -> {
                "${expr.functionName()}(${generateExpression(expr.left)}, ${generateExpression(expr.right)}, ${generateExpression(expr.isSub)}, ${generateExpression(expr.extraCarry)})"
            }
            is NegateWithFlagsExpression -> "negWithZFlag(${generateExpression(expr.expr)})"
            is GetCogNumberExpression -> "${regSetName(f.registerVarSet().sets[1], false)}.cogId"
            is GetPllOfCogExpression -> {
                val hubreg = f.registerVarSet().sets[0]
                usedRegsSets.add(hubreg)
                "((${regSetName(hubreg, false)}.cogs[${generateExpression(expr.cogNo)}]->reg_pll_saved_both & 1UL) == 1UL)"
            }
        }
        fun removeBrackets(inStr:String):String {
            var str = inStr
            while (str.isNotEmpty() && str.startsWith("(") && str.endsWith(")"))
                str = str.substring(1,str.length-1)
            var depth=0
            var minDepth=0
            for (c in str) {
                if (c == '(')
                    depth++
                else if (c == ')') {
                    depth--
                    if (depth<minDepth)
                        minDepth=depth
                }
            }
            //zu viele klammern entfernt
            while (minDepth<0) {
                minDepth++
                str = "($str)"
            }
            return str
        }
        fun generateStatements(indent:String,stmList:List<Statement>, builder:StringBuilder) {
            for (stm in stmList) {
                when(stm) {
                    is LetStatement -> builder.append("${indent}${varName(stm.destVar)} = ${generateExpression(stm.expr)};\n")
                    is LetArrayStatement -> {
                        val idxExpr = generateExpression(stm.idx)
                        val valueExpr = generateExpression(stm.expr)
                        val arrayName = arrayVarName(stm.destArrayVar)
                        val shiftAmount = stm.byteOffset*8
                        builder.append(when(stm.accessSize) {
                            CpuAccessSize.DWord -> "${indent}$arrayName[$idxExpr] = $valueExpr;\n"
                            CpuAccessSize.Word -> "${indent}writePartialArray($arrayName[$idxExpr],(($valueExpr)&0xFFFFU)<<$shiftAmount,0x${(0xFFFFU shl shiftAmount).inv().toString(16)}U);\n"
                            CpuAccessSize.Byte -> "${indent}writePartialArray($arrayName[$idxExpr],(($valueExpr)&0xFFU)<<$shiftAmount,0x${(0xFFU shl shiftAmount).inv().toString(16)}U);\n"
                            CpuAccessSize.None,CpuAccessSize.QWord -> throw Exception("not supported")
                        })
                    }
                    is SwitchStatement -> {
                        val subIndent = "$indent    "
                        if (stm.expr.dataType == ILangDataType.Bool) { //if
                            val trueBody = stm.matchingCase(1UL).body
                            val falseBody = stm.matchingCase(0UL).body
                            if (trueBody.isEmpty()) {
                                if (falseBody.isEmpty())
                                    throw Exception("empty if found")
                                builder.append("${indent}if (!(${removeBrackets(generateExpression(stm.expr))})) {\n")
                                generateStatements(subIndent,falseBody, builder)
                                builder.append("${indent}}\n")
                            }
                            else {
                                builder.append("${indent}if (${removeBrackets(generateExpression(stm.expr))}) {\n")
                                generateStatements(subIndent, trueBody, builder)
                                builder.append("${indent}}\n")
                                if (falseBody.isNotEmpty()) {
                                    builder.append("${indent}else {\n")
                                    generateStatements(subIndent, falseBody, builder)
                                    builder.append("${indent}}\n")
                                }
                            }
                        }
                        else { //switch
                            builder.append("${indent}switch (${removeBrackets(generateExpression(stm.expr))}) {\n")
                            for (c in stm.cases) {
                                builder.append(subIndent)
                                for (v in c.values)
                                    builder.append("case ${numericValue(ILangDataType.Int32, v)}: ")
                                builder.append("{\n")
                                generateStatements("$subIndent    ",c.body, builder)
                                builder.append("$subIndent    break;\n")
                                builder.append("$subIndent}\n")
                            }
                            builder.append("${indent}}\n")
                        }
                    }
                    is MacroStatement -> {
                        if (stm.doFunctionCall) {
                            val args = ArrayList(stm.regSet.sets.map {
                                usedRegsSets.add(it)
                                regSetName(it, false)
                            })
                            args.addAll(stm.inputs.map { varName(it) })
                            builder.append("${indent}${stm.macroName}(" + args.joinToString(",") + ");\n")
                        }
                        else {
                            builder.append("${indent}//Start of macro ${stm.macroName}\n")
                            generateStatements(indent,stm.body,builder)
                            builder.append("${indent}//End of macro ${stm.macroName}\n")
                        }
                    }
                }
            }
        }
        val pureOutputs = locals.getVariableList().filter { it.direction == ILangVariableDirection.Output }


        val bodyBuilder = StringBuilder()
        generateStatements("$indent    ",f.body,bodyBuilder)
        if (pureOutputs.size == 1)
            bodyBuilder.append("$indent    return ${varName(pureOutputs.first())};\n")



        if (pureOutputs.size == 1)
            funcBuilder.append("${indent}static ${dataTypeStr(pureOutputs.first().dataType)} ${f.name} (")
        else
            funcBuilder.append("${indent}static void ${f.name} (")
        val arguments = ArrayList(regs.sets.map {
            if (usedRegsSets.contains(it))
                regSetName(it, true) + "& " + regSetName(it, false)
            else
                regSetName(it, true) + "&"
        })
        for (v in locals.getVariableList()) {
            if (v.direction == ILangVariableDirection.Input)
                arguments.add("const ${dataTypeStr(v.dataType)} ${v.name}")
            else if (v.direction == ILangVariableDirection.InOut)
                arguments.add("${dataTypeStr(v.dataType)}& ${v.name}")
            else if (v.direction == ILangVariableDirection.Output && pureOutputs.size != 1)
                arguments.add("${dataTypeStr(v.dataType)}& ${v.name}")
        }
        funcBuilder.append(arguments.joinToString(","))
        funcBuilder.append(") {\n")
        for (v in locals.getVariableList()) {
            if (v.direction == ILangVariableDirection.Local || (v.direction == ILangVariableDirection.Output && pureOutputs.size==1))
                funcBuilder.append("$indent    ${dataTypeStr(v.dataType)} ${v.name} = ${numericValue(v.dataType,0UL)};\n")
            else if (v.direction != ILangVariableDirection.Input && v.direction != ILangVariableDirection.InOut && v.direction != ILangVariableDirection.Output)
                throw Exception("no ${v.direction} supported in cpp generator: ${v.name} of ${f.name}")
        }
        funcBuilder.append(bodyBuilder)
        funcBuilder.append("${indent}}\n\n")
    }

    val mainBuilder=StringBuilder()
    mainBuilder.append("#ifndef GENERATED_PROPELLER_FUNCTIONS\n")
    mainBuilder.append("#define GENERATED_PROPELLER_FUNCTIONS\n")
    mainBuilder.append("// Warning: Generated File! (by kotlin program: propeller-jvm-backend)\n")
    mainBuilder.append("#include <QDataStream>\n\n")
    mainBuilder.append("#include <QtDebug>\n\n")
    mainBuilder.append("struct PropellerNewCppImplGen {\n")
    mainBuilder.append("    struct rs_cog;\n") //hack
    for (set in regSets)
        generateRegisterSet("    ", set,mainBuilder)
    mainBuilder.append("    static void macroCogStartUp (rs_hub& reghub, const quint32 cogNo) {\n")
    mainBuilder.append("        reghub.cogs[cogNo]->progctr = reghub.dc >> 0x4;\n")
    mainBuilder.append("    }\n\n")
    mainBuilder.append("    static void writePartialArray (quint32& destArray, const quint32 value, const quint32 maskArray) {\n")
    mainBuilder.append("        destArray = (destArray&maskArray) | value;\n")
    mainBuilder.append("    }\n\n")
    mainBuilder.append("    static void cogUpdateAllDynamicFunctions(rs_hub&, rs_cog&) {} //In der CPP Version wird kein dynamischer Code generiert \n\n")
    for (f in function)
        generateFunction("    ",f,mainBuilder)
    for ((i,l) in lookupTables.withIndex())
        mainBuilder.append("    static constexpr quint32 ${lookupTableName(i)}[${l.size}] = {${l.map { numericValue(ILangDataType.Int32,it) }.joinToString(",")}};\n")
    mainBuilder.append("};\n")
    mainBuilder.append("#endif //GENERATED_PROPELLER_FUNCTIONS\n")
    return mainBuilder.toString()
}