package ilang

import java.lang.Exception

class PartialValue(dataType: ILangDataType, rawValue:ULong, rawUnknownBits:ULong) {
    val unknownBits:ULong
    val value:ULong
    init {
        unknownBits = rawUnknownBits and dataType.mask()
        value = rawValue and dataType.mask() and unknownBits.inv()
    }
    fun valueAsInt(maxValue:Int):Int =  if (value > maxValue.toULong()) maxValue else value.toInt()
    fun fullySpecified():Boolean = unknownBits == 0UL
    fun knownBits():ULong = unknownBits.inv()
    fun knownOnes():ULong = knownBits() and value
    fun knownZeros():ULong = knownBits() and value.inv()
    fun bitIsKnown(bitIdx:Int):Boolean = ((unknownBits shr bitIdx) and 1UL) == 0UL
    fun bitValue(bitIdx:Int):Boolean = ((value shr bitIdx) and 1UL) == 1UL
    fun boolean():Boolean = (value and 1UL) == 1UL
    fun merged(dataType: ILangDataType, other:PartialValue):PartialValue {
        val mismatchBits = value xor other.value //unterschiedliche bits
        return PartialValue(dataType, value, unknownBits or other.unknownBits or mismatchBits)
    }

    override fun toString(): String = "0x${value.toString(16)}:0x${knownBits().toString(16)}"
}
class PartialResult(val value:PartialValue, val expr:Expression) {
    override fun toString(): String = "$value:$expr"
}
fun constantPartialResult(dataType: ILangDataType, value: ULong) = PartialResult(PartialValue(dataType,value,0UL), ConstantExpression(value, dataType))
fun constantPartialResult(value: Boolean) = constantPartialResult(ILangDataType.Bool, if (value) 1UL else 0UL)
fun partialResultOf(value:PartialValue, expr:Expression) = PartialResult(value, if (!value.fullySpecified() || expr is ConstantExpression) expr else ConstantExpression(value.value,expr.dataType))
fun fromKnownBits(dataType: ILangDataType, value:ULong, knownBits:ULong) = PartialValue(dataType, value, knownBits.inv())
fun fromExactValue(dataType: ILangDataType, value:ULong) = PartialValue(dataType, value, 0UL)
fun fromExactValue(value:Boolean) = PartialValue(ILangDataType.Bool, if (value) 1UL else 0UL, 0UL)
fun fromUnknownValue(dataType: ILangDataType) = fromKnownBits(dataType, 0UL, 0UL)

class PartialValueVarState {
    class State(var value:PartialValue, var forwaredVariable:ILangVariable?) { //forwardedVariable ist gesetzt, wenn es direkt einer variablen entspricht
        fun merge(dataType: ILangDataType, other:State) {
            value = value.merged(dataType, other.value)
            forwaredVariable = if (forwaredVariable === other.forwaredVariable) forwaredVariable else null
        }
    }
    private val map:HashMap<ILangVariable, State>
    private val neverFwdVar:HashSet<ILangVariable>
    constructor(parent:PartialValueVarState) {
        map = HashMap(parent.map)
        neverFwdVar = parent.neverFwdVar //wird nicht kopiert
    }
    constructor() {
        map = HashMap()
        neverFwdVar = HashSet()
    }
    constructor(function:Function, fixedBooleanValues:List<Pair<String,Boolean>>) {
        map = HashMap()
        neverFwdVar = HashSet()
        for (v in fixedBooleanValues)
            setPartialValue(function.getVariableByName(v.first), v.second, false)
    }
    fun getNeverForwardVars():Set<ILangVariable> = neverFwdVar
    fun getPartialValue(v:ILangVariable):State {
        val r =  map[v] ?: return State(fromUnknownValue(v.dataType),null)
        if (r.forwaredVariable == null)
            return r
        return if (!neverFwdVar.contains(v)) r else State(r.value, null)
    }
    fun setNeverForwardVariable(v:ILangVariable):PartialValueVarState {
        neverFwdVar.add(v)
        return this
    }
    fun setPartialValueInitialConstraint(v:ILangVariable, p:PartialValue):PartialValueVarState {
        map[v] = State(p, null)
        return this
    }
    fun setPartialValue(v:ILangVariable, p:PartialResult, implicit:Boolean):PartialValueVarState {
        //implicit gibt an, ob die Variable wirklich gesetzt wurde mit einem let (false), oder sich aus dem kontext (durch if condition) ergibt
        if (!implicit) {
            for (s in map.values) {
                if (s.forwaredVariable === v) //variable wurde ungültg, da hier überschrieben
                    s.forwaredVariable = null
            }
        }
        map[v] = State(p.value, if (p.expr is VariableExpression) p.expr.variable else null)
        return this
    }
    fun setPartialValue(v:ILangVariable, p:Boolean, implicit:Boolean):PartialValueVarState {
        return setPartialValue(v, constantPartialResult(p), implicit)
    }
    fun setPartialValue(v:ILangVariable, dataType: ILangDataType, value:ULong, implicit:Boolean):PartialValueVarState {
        return setPartialValue(v, constantPartialResult(dataType, value), implicit)
    }
    //von einer switch anweisung gibt es verschiedene Pfade, die das Programm durchlaufen hat
    fun joinFromChildren(children:List<PartialValueVarState>) {
        if (children.isEmpty())
            return
        val affectedVarsInChildren = HashSet<ILangVariable>() //welche variablen wurden in den zweigen modifiziert?
        for (c in children)
            affectedVarsInChildren.addAll(c.map.keys)

        val remainingChildren = children.drop(1)
        //alle variablen durchgehen
        for (v in affectedVarsInChildren) {
            var varState = children.first().getPartialValue(v)
            for (c in remainingChildren)
                varState.merge(v.dataType, c.getPartialValue(v))
            map[v] = varState
        }
    }
}

fun partialEvaluateExpression(rootExpr:Expression, varState:PartialValueVarState):PartialResult {
    fun evaluate(expr:Expression):PartialResult = when(expr) {
        is VariableExpression -> {
            val p = varState.getPartialValue(expr.variable)
            //wenn der Wert exakt bekannt ist, kann die Variable durch eine konstante ersetzt werden
            val fwdVar = p.forwaredVariable
            partialResultOf(p.value, if (fwdVar == null) expr else VariableExpression(fwdVar))
        }
        is ConstantExpression -> PartialResult(fromExactValue(expr.dataType,expr.value),expr)
        is ConstantJumpDestinationExpression -> partialResultOf(fromUnknownValue(expr.dataType),expr) //nicht wegoptimieren
        is EqualityExpression -> {
            val l = evaluate(expr.left)
            val r = evaluate(expr.right)
            val commonKnownBits = l.value.knownBits() and r.value.knownBits() //Bits die links und rechts bekannt sind
            val fullySpecified = l.value.fullySpecified() and r.value.fullySpecified()
            //prüfen, ob es bei den bereits bekannten bits gleichheit gibt
            val isEqCommonBits = (l.value.value and commonKnownBits) == (r.value.value and commonKnownBits)
            val p = if (expr.isEq) {
                //== operator
                if (fullySpecified)
                    fromExactValue(isEqCommonBits)
                else if (!isEqCommonBits) //zwar nicht vollständig bekannt, kann aber nicht mehr gleich werden
                    fromExactValue(false)
                else
                    fromUnknownValue(ILangDataType.Bool)
            }
            else {
                //!= operator
                if (fullySpecified)
                    fromExactValue(!isEqCommonBits)
                else if (!isEqCommonBits) //zwar nicht vollständig bekannt, kann aber nicht mehr (nicht ungleich) werden
                    fromExactValue(true)
                else
                    fromUnknownValue(ILangDataType.Bool)
            }
            partialResultOf(p, expr.modified(l.expr, r.expr))
        }
        is CompareExpression -> {
            val l = evaluate(expr.left)
            val r = evaluate(expr.right)
            //Einfacher Fall, wenn einer der Seiten teilweise unbekannt ist, ist das Ergebnis unbekannt
            //Das ginge auch besser
            if (!l.value.fullySpecified() or !r.value.fullySpecified())
                PartialResult(fromUnknownValue(ILangDataType.Bool),expr.modified(l.expr, r.expr))
            else {
                //beide Seiten sind vollständig bekannt
                constantPartialResult(when(expr.operation) {
                    CompareExpression.Operation.Le -> l.value.value < r.value.value
                    CompareExpression.Operation.Leq -> l.value.value <= r.value.value
                    CompareExpression.Operation.Ge -> l.value.value > r.value.value
                    CompareExpression.Operation.Geq -> l.value.value >= r.value.value
                    CompareExpression.Operation.SgnLe -> {
                        //TODO ist das richtig
                        println("TODO SGN LE ${l.value.value} ${r.value.value}")
                        l.value.value.toInt() < r.value.value.toInt()
                    }
                })
            }
        }
        is AdditiveOrBitwiseExpression -> {
            val l = evaluate(expr.left)
            val r = evaluate(expr.right)
            val modExpr = expr.modified(l.expr, r.expr)
            when(expr.operation) {
                AdditiveOrBitwiseExpression.Operation.And -> {
                    /*if (expr.toString()=="(and32 p_reg 0I1ff)")
                        println("optimize and32: ${l.value} ${r.value} ${((l.value.knownBits() and r.value.knownBits()) or l.value.knownZeros() or r.value.knownZeros()).toString(16)}" +
                                " ${(r.value.value or l.value.knownZeros()).toString(16)}")*/
                    //Wenn Links voll bekannt ist, dann ist das And überflüssig, wenn ein Bit links 1 ist oder rechts bekannt und 0
                    if (l.value.fullySpecified() && (l.value.value or r.value.knownZeros()) == 0xFFFFFFFFFFFFFFFFUL) r
                    else if (r.value.fullySpecified() && (r.value.value or l.value.knownZeros()) == 0xFFFFFFFFFFFFFFFFUL) l
                    else
                        partialResultOf(
                            fromKnownBits(
                                expr.dataType,
                                l.value.value and r.value.value,
                                //Ein Bit ist bekannt, wenn es rechts und links bekannt ist, oder es bei der bekannten Seite 0 ist
                                (l.value.knownBits() and r.value.knownBits()) or l.value.knownZeros() or r.value.knownZeros()
                            ), modExpr
                        )
                }
                AdditiveOrBitwiseExpression.Operation.Or ->
                    if (l.value.fullySpecified() && l.value.value == 0UL) r
                    else if (r.value.fullySpecified() && r.value.value == 0UL) l
                    else
                        partialResultOf(fromKnownBits(
                            expr.dataType,
                            l.value.value or r.value.value,
                            //Ein Bit ist bekannt, wenn es rechts und links bekannt ist, oder es bei der bekannten Seite 1 ist
                            (l.value.knownBits() and r.value.knownBits()) or l.value.knownOnes() or r.value.knownOnes()
                        ),modExpr)
                AdditiveOrBitwiseExpression.Operation.Xor -> partialResultOf(fromKnownBits(
                    expr.dataType,
                    l.value.value xor r.value.value,
                    //Ein Bit ist bekannt, wenn es rechts und links bekannt ist
                    l.value.knownBits() and r.value.knownBits()
                ),modExpr)
                AdditiveOrBitwiseExpression.Operation.Add ->
                    //Einfache Lösung: Addition ist nur bekannt, wenn beide Operanden vollständig bekannt sind
                    //das geht auch besser, da unbekannte Bits nur nach vorne propagieren können
                    partialResultOf((
                        if (l.value.fullySpecified() && r.value.fullySpecified())
                            fromExactValue(expr.dataType, l.value.value+r.value.value)
                        else
                            fromUnknownValue(expr.dataType)
                        ),modExpr)
                AdditiveOrBitwiseExpression.Operation.Sub ->
                    //Einfache Lösung: Subtraktion ist nur bekannt, wenn beide Operanden vollständig bekannt sind
                    //das geht auch besser, wie bei Addition
                    partialResultOf((
                        if (l.value.fullySpecified() && r.value.fullySpecified())
                            fromExactValue(expr.dataType, l.value.value-r.value.value)
                        else
                            fromUnknownValue(expr.dataType)
                        ),modExpr)
            }
        }
        is InvertExpression -> {
            val e = evaluate(expr.expr)
            partialResultOf(PartialValue(expr.dataType, e.value.value.inv(), e.value.unknownBits),expr.modified(e.expr))
        }
        is Lower32Expression -> {
            val e = evaluate(expr.expr)
            partialResultOf(PartialValue(expr.dataType, e.value.value and 0xFFFFFFFFUL, e.value.unknownBits and 0xFFFFFFFFUL),expr.modified(e.expr))
        }
        is Cast64Expression -> {
            val e = evaluate(expr.expr)
            partialResultOf(PartialValue(expr.dataType, e.value.value and 0xFFFFFFFFUL, e.value.unknownBits and 0xFFFFFFFFUL),expr.modified(e.expr))
        }
        is ShiftExpression -> {
            val l = evaluate(expr.left)
            val r = evaluate(expr.right)
            val tmpRes = expr.modified(l.expr,r.expr)
            if (r.value.fullySpecified()) {
                val shiftBy = r.value.valueAsInt(64)
                if (shiftBy == 0) l
                else partialResultOf((if (expr.shiftLeft)
                                    PartialValue(expr.dataType, l.value.value shl shiftBy,l.value.unknownBits shl shiftBy)
                                else
                                    PartialValue(expr.dataType, l.value.value shr shiftBy,l.value.unknownBits shr shiftBy)),
                tmpRes)
            }
            else {
                partialResultOf((if (l.value.fullySpecified() && l.value.value == 0UL) //0 bleibt 0 egal wie geshiftet
                                    fromExactValue(expr.dataType, 0UL)
                                else
                                    fromUnknownValue(expr.dataType)),
                tmpRes)
            }
        }
        is TestBitExpression -> {
            val baseExpr = evaluate(expr.expr)
            val bitNo = evaluate(expr.bitNo)
            val p = if (bitNo.value.fullySpecified() && baseExpr.value.bitIsKnown(bitNo.value.valueAsInt(64)))
                fromExactValue(baseExpr.value.bitValue(bitNo.value.valueAsInt(64)))
            else if (baseExpr.value.fullySpecified() && baseExpr.value.value == 0UL) //jedes bit ist 0
                fromExactValue(false)
            else
                fromUnknownValue(ILangDataType.Bool)
            partialResultOf(p, expr.modified(baseExpr.expr, bitNo.expr))
        }
        is InRange32Expression -> {
            val baseExpr = evaluate(expr.expr)
            val p = if (baseExpr.value.fullySpecified())
                fromExactValue(baseExpr.value.value >= expr.lower && baseExpr.value.value <= expr.upper)
            else
                fromUnknownValue(ILangDataType.Bool)
            partialResultOf(p, expr.modified(baseExpr.expr))
        }
        is MuxExpression -> {
            val c = evaluate(expr.condition)
            val t = evaluate(expr.trueExpr)
            val f = evaluate(expr.falseExpr)
            if (c.value.fullySpecified())
                if (c.value.boolean()) t else f
            else
                partialResultOf(fromUnknownValue(expr.dataType),expr.modified(c.expr,t.expr,f.expr))
        }
        is ArrayAccessExpression -> {
            val idx = evaluate(expr.indexExpr)
            partialResultOf(fromUnknownValue(expr.dataType),expr.modified(idx.expr))
        }
        is PartialArrayAccessExpression -> {
            val idx = evaluate(expr.indexExpr)
            partialResultOf(fromUnknownValue(expr.dataType),expr.modified(idx.expr))
        }
        is DecryptExpression -> partialResultOf(fromUnknownValue(expr.dataType),expr.modified(evaluate(expr.expr).expr))
        is ParityExpression -> partialResultOf(fromUnknownValue(expr.dataType),expr.modified(evaluate(expr.expr).expr))
        is ExtendToU64ShiftSignToUpperExpression -> partialResultOf(fromUnknownValue(expr.dataType),expr.modified(evaluate(expr.expr).expr))
        is LookupExpression -> {
            val idxExpr = evaluate(expr.index)
            if (!idxExpr.value.fullySpecified())
                partialResultOf(fromUnknownValue(expr.dataType),expr.modified(idxExpr.expr))
            else {
                if (idxExpr.value.value>=expr.values.size.toULong())
                    throw Exception("index out of range in lookup expression: $idxExpr; ${expr.values.size}")
                constantPartialResult(expr.dataType,expr.values[idxExpr.value.value.toInt()])
            }
        }
        is JumpDestinationLookupExpression -> {
            val idxExpr = evaluate(expr.index)
            if (!idxExpr.value.fullySpecified())
                partialResultOf(fromUnknownValue(expr.dataType),expr.modified(idxExpr.expr))
            else {
                if (idxExpr.value.value>=expr.values.size.toULong())
                    throw Exception("index out of range in lookup expression: $idxExpr; ${expr.values.size}")
                val tmp = expr.values[idxExpr.value.value.toInt()]
                partialResultOf(fromUnknownValue(expr.dataType),ConstantJumpDestinationExpression(tmp.value,tmp.label))
            }
        }
        is ExtractFromCombinedOperation -> {
            partialResultOf(fromUnknownValue(expr.dataType),expr)
        }
        is AddOrSubExpression -> {
            partialResultOf(fromUnknownValue(expr.dataType),expr.modified(
                evaluate(expr.left).expr,
                evaluate(expr.right).expr,
                evaluate(expr.isSub).expr,
                evaluate(expr.extraCarry).expr)
            )
        }
        is NegateWithFlagsExpression -> partialResultOf(fromUnknownValue(expr.dataType),expr.modified(evaluate(expr.expr).expr))
        is GetCogNumberExpression -> partialResultOf(fromUnknownValue(expr.dataType),expr)
        is GetPllOfCogExpression -> partialResultOf(fromUnknownValue(expr.dataType),expr.modified(evaluate(expr.cogNo).expr))
    }
    return evaluate(rootExpr)
}
