package ilang

import assembler.CpuAccessSize

enum class ILangDataType {
    Bool,Int32,Int64;
    fun mask():ULong = when(this) {
        ILangDataType.Bool -> 1UL
        ILangDataType.Int32 -> 0xFFFFFFFFUL
        ILangDataType.Int64 -> 0xFFFFFFFFFFFFFFFFUL
    }
}
enum class ILangVariableDirection {Local,Input,Output,InOut,Register}
class ILangVariable(val name:String, val dataType:ILangDataType, val direction:ILangVariableDirection) {
    override fun toString(): String = name
}
class ILangArrayVariable(val name:String, val dataType:ILangDataType, val size:Int) {
    override fun toString(): String = name
}
class ILangVariableSet {
    val name:String
    private val variablesMap:HashMap<String,ILangVariable>
    private val variablesList:ArrayList<ILangVariable>
    private val arrayVariablesMap:HashMap<String,ILangArrayVariable>
    private val arrayVariablesList:ArrayList<ILangArrayVariable>

    constructor(setName:String) {
        name = setName
        variablesMap = HashMap()
        variablesList = ArrayList()
        arrayVariablesMap = HashMap()
        arrayVariablesList = ArrayList()
    }

    constructor(copyFrom:ILangVariableSet) {
        name = copyFrom.name
        variablesMap = HashMap(copyFrom.variablesMap)
        variablesList = ArrayList(copyFrom.variablesList)
        arrayVariablesMap = HashMap(copyFrom.arrayVariablesMap)
        arrayVariablesList = ArrayList(copyFrom.arrayVariablesList)
    }

    fun getVariable(name:String):ILangVariable? = variablesMap[name]
    fun getArrayVariable(name:String):ILangArrayVariable? = arrayVariablesMap[name]

    fun contains(v:ILangVariable) = variablesList.contains(v)
    fun contains(v:ILangArrayVariable) = arrayVariablesList.contains(v)

    fun addVariable(v:ILangVariable) {
        if (variablesMap.containsKey(v.name))
            throw Exception("duplicate variable ${v.name}")
        variablesMap[v.name] = v
        variablesList.add(v)
    }

    fun addArrayVariable(v:ILangArrayVariable) {
        if (arrayVariablesMap.containsKey(v.name))
            throw Exception("duplicate array variable ${v.name}")
        arrayVariablesMap[v.name] = v
        arrayVariablesList.add(v)
    }

    fun removeVariable(v:ILangVariable) {
        variablesList.remove(v)
        variablesMap.remove(v.name)
    }

    fun removeArrayVariable(v:ILangArrayVariable) {
        arrayVariablesList.remove(v)
        arrayVariablesMap.remove(v.name)
    }

    fun getVariableList():List<ILangVariable> = variablesList
    fun getArrayVariableList():List<ILangArrayVariable> = arrayVariablesList
}

class ILangMultiVariableSet(val sets:List<ILangVariableSet>) {
    fun getVariableByName(name:String):ILangVariable {
        var result:ILangVariable? = null
        for (regSet in sets) {
            val v = regSet.getVariable(name) ?: continue
            if (result == null)
                result = v
            else
                throw Exception("duplicate variable $name")
        }
        return result ?: throw Exception("unknown variable $name")
    }
    fun getArrayVariableByName(name:String):ILangArrayVariable {
        var result:ILangArrayVariable? = null
        for (regSet in sets) {
            val v = regSet.getArrayVariable(name) ?: continue
            if (result == null)
                result = v
            else
                throw Exception("duplicate variable $name")
        }
        return result ?: throw Exception("unknown array variable $name")
    }
}

class Function(val name:String, val allVarSets:ILangMultiVariableSet, val body:List<Statement>) {
    fun getVariableByName(name:String):ILangVariable = allVarSets.getVariableByName(name)
    fun localVarSet() = allVarSets.sets.first()
    fun registerVarSet() = ILangMultiVariableSet(allVarSets.sets.subList(1,allVarSets.sets.size))
    fun modified(newBody:List<Statement>):Function {
        if (newBody === body)
            return this
        return Function(name, allVarSets, newBody)
    }
    fun modified(newLocals:ILangVariableSet):Function {
        val newMultiSet = arrayListOf(newLocals)
        newMultiSet.addAll(registerVarSet().sets)
        return Function(name,ILangMultiVariableSet(newMultiSet),body)
    }
    fun localsAdded(newLocalVars:List<ILangVariable>):Function {
        val newLocalSet = ILangVariableSet(localVarSet())
        for (l in newLocalVars)
            newLocalSet.addVariable(l)
        return modified(newLocalSet)
    }
    fun rename(newName:String) = Function(newName, allVarSets, body)
    fun iterate(cb:(stm:Statement)->Unit) {
        for (s in body)
            s.iterate(cb)
    }
}

class ILangVarAndArraySet(val vars:Set<ILangVariable>, val arrs:Set<ILangArrayVariable>) {
    fun union(other:ILangVarAndArraySet) = ILangVarAndArraySet(vars+other.vars, arrs+other.arrs)
    fun intersects(other:ILangVarAndArraySet):Boolean = vars.intersect(other.vars).isNotEmpty() || arrs.intersect(other.arrs).isNotEmpty()
}

sealed class Expression(val dataType: ILangDataType) {
    abstract fun toILang():String
    abstract fun iterate(cb:(e:Expression)->Unit)
    abstract fun map(cb:(e:Expression)->Expression):Expression
    abstract fun isSameExpression(other:Expression):Boolean
    override fun toString(): String = toILang()
    fun getUsedVariables():List<ILangVariable> { //wenn eine variable im ausdruck mehrfach verwendet wird, kommt sie auch mehrfach vor
        val result = ArrayList<ILangVariable>()
        iterate {
            if (it is VariableExpression)
                result.add(it.variable)
            if (it is ExtractFromCombinedOperation) {
                //2x hinzufügen, damit es nie geforwarded wird
                result.add(it.variable)
                result.add(it.variable)
            }
        }
        return result
    }
    fun getUsedVariablesAndArrays():ILangVarAndArraySet {
        val varSet = HashSet<ILangVariable>()
        val arrSet = HashSet<ILangArrayVariable>()
        iterate {
            if (it is VariableExpression)
                varSet.add(it.variable)
            if (it is ArrayAccessExpression)
                arrSet.add(it.arrayVariable)
            if (it is PartialArrayAccessExpression)
                arrSet.add(it.arrayVariable)
            if (it is ExtractFromCombinedOperation)
                varSet.add(it.variable)
        }
        return ILangVarAndArraySet(varSet,arrSet)
    }
    abstract fun stackComplexity():Int
}
class VariableExpression(val variable:ILangVariable):Expression(variable.dataType) {
    override fun toILang():String = variable.name
    override fun iterate(cb:(e:Expression)->Unit) = cb(this)
    override fun map(cb:(e:Expression)->Expression):Expression = cb(this)
    fun modified(newVariable:ILangVariable):VariableExpression = if (newVariable === variable) this else VariableExpression(newVariable)
    override fun isSameExpression(other:Expression):Boolean = other is VariableExpression && other.dataType == dataType && other.variable === variable
    override fun stackComplexity():Int = 2
}
class ConstantExpression(val value:ULong, dataType: ILangDataType):Expression(dataType) {
    constructor(v:Boolean):this(if (v) 1UL else 0UL, ILangDataType.Bool)
    override fun toILang():String = when(dataType) {
        ILangDataType.Bool -> if (value == 1UL) "true" else "false"
        ILangDataType.Int32 -> "0I"+value.toString(16)
        ILangDataType.Int64 -> "0L"+value.toString(16)
    }
    override fun stackComplexity():Int = 1
    override fun iterate(cb:(e:Expression)->Unit) = cb(this)
    override fun map(cb:(e:Expression)->Expression):Expression = cb(this)
    override fun isSameExpression(other:Expression):Boolean = other is ConstantExpression && other.dataType == dataType && other.value == value
    fun fits32Bit():Boolean = value<0x100000000UL
    fun is1ShlN():Int {
        for (i in 0 until 64)
            if (value == 1UL shl i)
                return i
        return -1
    }
}
class ConstantJumpDestinationExpression(val value:ULong, val label:String):Expression(ILangDataType.Int32) {
    constructor(e:JumpDestinationLookupExpression.Entry):this(e.value,e.label)
    override fun toILang(): String = "(jump 0I${value.toString(16)})"
    override fun stackComplexity():Int = 1
    override fun iterate(cb:(e:Expression)->Unit) = cb(this)
    override fun map(cb:(e:Expression)->Expression):Expression = cb(this)
    override fun isSameExpression(other:Expression):Boolean = other is ConstantJumpDestinationExpression && other.value == value && other.label == label
}
fun opConst32(value:ULong) = ConstantExpression(value,ILangDataType.Int32)
class EqualityExpression(dataType: ILangDataType, val isEq:Boolean, val left:Expression, val right:Expression):Expression(ILangDataType.Bool) {
    init {
        if (left.dataType != dataType || right.dataType != dataType)
            throw Exception("datatype mismatch ${left.dataType} vs $dataType /  ${right.dataType} vs $dataType")
    }
    fun modified(newLeft:Expression, newRight:Expression):EqualityExpression = if (newLeft === left && newRight === right) this else EqualityExpression(newLeft.dataType, isEq, newLeft, newRight)
    override fun toILang():String = when(left.dataType) {
        ILangDataType.Bool -> if (isEq) "(eqB" else "(neqB"
        ILangDataType.Int32 -> if (isEq) "(eq32" else "(neq32"
        ILangDataType.Int64 -> if (isEq) "(eq64" else "(neq64"
    }+" ${left.toILang()} ${right.toILang()})"
    override fun iterate(cb:(e:Expression)->Unit) {
        left.iterate(cb)
        right.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(left.map(cb),right.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is EqualityExpression && other.isEq == isEq && other.left.isSameExpression(left) && other.right.isSameExpression(right)
    override fun stackComplexity():Int = left.stackComplexity()+right.stackComplexity()+2
}
fun opEq32(left:Expression, right:Expression) = EqualityExpression(ILangDataType.Int32,true, left, right)
fun opNeq32(left:Expression, right:Expression) = EqualityExpression(ILangDataType.Int32,false, left, right)

class CompareExpression(val operation:Operation, val left:Expression, val right:Expression):Expression(ILangDataType.Bool) {
    enum class Operation {Le,Leq,Ge,Geq,SgnLe}
    init {
        if (left.dataType != ILangDataType.Int32 || right.dataType != ILangDataType.Int32)
            throw Exception("datatype mismatch")
    }
    fun modified(newLeft:Expression, newRight:Expression):CompareExpression = if (newLeft === left && newRight === right) this else CompareExpression(operation, newLeft, newRight)
    override fun toILang():String = when(operation) {
        Operation.Le -> "(le32"
        Operation.Leq -> "(leq32"
        Operation.Ge -> "(ge32"
        Operation.Geq -> "(geq32"
        Operation.SgnLe -> "(sgnle32"
    }+" ${left.toILang()} ${right.toILang()})"
    override fun iterate(cb:(e:Expression)->Unit) {
        left.iterate(cb)
        right.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(left.map(cb),right.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is CompareExpression && other.operation == operation && other.left.isSameExpression(left) && other.right.isSameExpression(right)
    override fun stackComplexity():Int = left.stackComplexity()+right.stackComplexity()+2
}
class AdditiveOrBitwiseExpression(dataType: ILangDataType, val operation:Operation, val withFlags:Boolean, val left: Expression, val right:Expression):Expression(dataType) {
    enum class Operation {Add,Sub,And,Or,Xor}
    init {
        if (left.dataType != dataType || right.dataType != dataType)
            throw Exception("datatype mismatch")
        if (withFlags && dataType != ILangDataType.Int32 || (withFlags && operation != Operation.And && operation != Operation.Or && operation != Operation.Xor))
            throw Exception("bitwise with flags, invalid combination mismatch")
    }
    fun isCommutative() = when(operation) {
        Operation.Add,Operation.And,Operation.Xor,Operation.Or -> true
        Operation.Sub -> false
    }
    fun modified(newLeft:Expression, newRight:Expression):AdditiveOrBitwiseExpression = if (newLeft === left && newRight === right) this else AdditiveOrBitwiseExpression(dataType, operation, withFlags, newLeft, newRight)
    override fun toILang():String = (if (withFlags) when(operation) {
        Operation.Add,Operation.Sub -> throw Exception("not supported")
        Operation.And -> "(andWithZFlag"
        Operation.Or -> "(orWithZFlag"
        Operation.Xor -> "(xorWithZFlag"
    }
    else when(operation) {
        Operation.Add -> "(add"
        Operation.Sub -> "(sub"
        Operation.And -> "(and"
        Operation.Or -> "(or"
        Operation.Xor -> "(xor"
    }+when(left.dataType) {
        ILangDataType.Bool -> "B"
        ILangDataType.Int32 -> "32"
        ILangDataType.Int64 -> "64"
    })+" ${left.toILang()} ${right.toILang()})"
    override fun iterate(cb:(e:Expression)->Unit) {
        left.iterate(cb)
        right.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(left.map(cb),right.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is AdditiveOrBitwiseExpression && other.operation == operation && other.left.isSameExpression(left) && other.right.isSameExpression(right)
    override fun stackComplexity():Int = left.stackComplexity()+right.stackComplexity()+2
}
fun opAndB(left:Expression, right:Expression) = AdditiveOrBitwiseExpression(ILangDataType.Bool, AdditiveOrBitwiseExpression.Operation.And, false, left, right)
fun opOrB(left:Expression, right:Expression) = AdditiveOrBitwiseExpression(ILangDataType.Bool, AdditiveOrBitwiseExpression.Operation.Or, false, left, right)
fun opAnd32(left:Expression, right:Expression) = AdditiveOrBitwiseExpression(ILangDataType.Int32, AdditiveOrBitwiseExpression.Operation.And, false, left, right)
fun opOr32(left:Expression, right:Expression) = AdditiveOrBitwiseExpression(ILangDataType.Int32, AdditiveOrBitwiseExpression.Operation.Or, false, left, right)
fun opXor32(left:Expression, right:Expression) = AdditiveOrBitwiseExpression(ILangDataType.Int32, AdditiveOrBitwiseExpression.Operation.Xor, false, left, right)
fun opAdd32(left:Expression, right:Expression) = AdditiveOrBitwiseExpression(ILangDataType.Int32, AdditiveOrBitwiseExpression.Operation.Add, false, left, right)
fun opSub32(left:Expression, right:Expression) = AdditiveOrBitwiseExpression(ILangDataType.Int32, AdditiveOrBitwiseExpression.Operation.Sub, false, left, right)

class ShiftExpression(dataType: ILangDataType, val shiftLeft:Boolean, val withFlags: Boolean, val left:Expression, val right:Expression):Expression(dataType) {
    init {
        if (left.dataType != dataType || right.dataType != ILangDataType.Int32)
            throw Exception("datatype mismatch")
        if (withFlags && dataType != ILangDataType.Int32)
            throw Exception("illegal operation")
    }
    fun modified(newLeft:Expression, newRight:Expression):ShiftExpression = if (newLeft === left && newRight === right) this else ShiftExpression(dataType, shiftLeft, withFlags, newLeft, newRight)
    override fun toILang():String = when(dataType) {
        ILangDataType.Bool -> throw Exception("not supported")
        ILangDataType.Int32 -> if (withFlags) if (shiftLeft) "(shlWithZFlag" else "(shrWithZFlag" else if (shiftLeft) "(shl32" else "(shr32"
        ILangDataType.Int64 -> if (shiftLeft) "(shl64" else "(shr64"
    }+" ${left.toILang()} ${right.toILang()})"
    override fun iterate(cb:(e:Expression)->Unit) {
        left.iterate(cb)
        right.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(left.map(cb),right.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is ShiftExpression && other.shiftLeft == shiftLeft && other.left.isSameExpression(left) && other.right.isSameExpression(right)
    override fun stackComplexity():Int = left.stackComplexity()+right.stackComplexity()+2
}

fun opShl32(left:Expression, right:Expression) = ShiftExpression(ILangDataType.Int32, true, false, left, right)
fun opShr32(left:Expression, right:Expression) = ShiftExpression(ILangDataType.Int32, false, false, left, right)
fun opShl64(left:Expression, right:Expression) = ShiftExpression(ILangDataType.Int64, true, false, left, right)
fun opShr64(left:Expression, right:Expression) = ShiftExpression(ILangDataType.Int64, false, false, left, right)

class InvertExpression(dataType: ILangDataType, val expr: Expression):Expression(dataType) {
    enum class Operation {Add,Sub,And,Or,Xor}
    init {
        if (expr.dataType != dataType)
            throw Exception("datatype mismatch")
    }
    fun modified(newExpr:Expression):InvertExpression = if (newExpr === expr) this else InvertExpression(dataType, newExpr)
    override fun toILang():String = when(dataType) {
        ILangDataType.Bool -> "(notB"
        ILangDataType.Int32 -> "(invert32"
        ILangDataType.Int64 -> "(invert64"
    }+" ${expr.toILang()})"
    override fun iterate(cb:(e:Expression)->Unit) {
        expr.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(expr.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is InvertExpression && other.expr.isSameExpression(expr)
    override fun stackComplexity():Int = expr.stackComplexity()+1
}
fun opNotB(expr:Expression) = InvertExpression(ILangDataType.Bool, expr)

class Lower32Expression(val expr:Expression):Expression(ILangDataType.Int32) {
    init {
        if (expr.dataType != ILangDataType.Int64)
            throw Exception("datatype mismatch")
    }
    fun modified(newExpr:Expression):Lower32Expression = if (newExpr === expr) this else Lower32Expression(newExpr)
    override fun toILang():String = "(lower32 ${expr.toILang()})"
    override fun iterate(cb:(e:Expression)->Unit) {
        expr.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(expr.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is Lower32Expression && other.expr.isSameExpression(expr)
    override fun stackComplexity():Int = expr.stackComplexity()+1
}

fun opLower32(expr:Expression) = Lower32Expression(expr)

class Cast64Expression(val expr:Expression):Expression(ILangDataType.Int64) {
    init {
        if (expr.dataType != ILangDataType.Int32)
            throw Exception("datatype mismatch")
    }
    fun modified(newExpr:Expression):Cast64Expression = if (newExpr === expr) this else Cast64Expression(newExpr)
    override fun toILang():String = "(cast64 ${expr.toILang()})"
    override fun iterate(cb:(e:Expression)->Unit) {
        expr.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(expr.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is Cast64Expression && other.expr.isSameExpression(expr)
    override fun stackComplexity():Int = expr.stackComplexity()+1
}
class MuxExpression(dataType: ILangDataType, val condition:Expression, val trueExpr:Expression, val falseExpr: Expression):Expression(dataType) {
    init {
        if (trueExpr.dataType != dataType || falseExpr.dataType != dataType || condition.dataType != ILangDataType.Bool)
            throw Exception("datatype mismatch")
    }
    fun modified(newCondition: Expression, newTrueExpr:Expression, newFalseExpr:Expression):MuxExpression = if (newCondition === condition && newTrueExpr === trueExpr && newFalseExpr === falseExpr) this else MuxExpression(dataType, newCondition, newTrueExpr, newFalseExpr)
    override fun toILang():String = when(dataType) {
        ILangDataType.Bool -> "(muxB"
        ILangDataType.Int32 -> "(mux32"
        ILangDataType.Int64 -> "(mux64"
    }+" ${condition.toILang()} ${trueExpr.toILang()} ${falseExpr.toILang()})"
    override fun iterate(cb:(e:Expression)->Unit) {
        condition.iterate(cb)
        trueExpr.iterate(cb)
        falseExpr.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(condition.map(cb),trueExpr.map(cb),falseExpr.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is MuxExpression && other.condition.isSameExpression(condition) && other.trueExpr.isSameExpression(trueExpr) && other.falseExpr.isSameExpression(falseExpr)
    override fun stackComplexity():Int = condition.stackComplexity()+trueExpr.stackComplexity()+falseExpr.stackComplexity()+4
}
fun opMux32(condition: Expression, trueExpr: Expression, falseExpr: Expression) = MuxExpression(ILangDataType.Int32,condition,trueExpr,falseExpr)
class TestBitExpression(dataType: ILangDataType, val expr:Expression, val bitNo:Expression):Expression(ILangDataType.Bool) {
    init {
        if (expr.dataType != dataType || bitNo.dataType != ILangDataType.Int32)
            throw Exception("datatype mismatch")
    }
    fun modified(newExpr:Expression, newBitNo:Expression):TestBitExpression = if (newExpr === expr && newBitNo === bitNo) this else TestBitExpression(newExpr.dataType, newExpr, newBitNo)
    override fun toILang():String = when(expr.dataType) {
        ILangDataType.Bool -> throw Exception("not supported")
        ILangDataType.Int32 -> "(testBit32"
        ILangDataType.Int64 -> "(testBit64"
    }+" ${expr.toILang()} ${bitNo.toILang()})"
    override fun iterate(cb:(e:Expression)->Unit) {
        expr.iterate(cb)
        bitNo.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(expr.map(cb),bitNo.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is TestBitExpression && other.expr.isSameExpression(expr) && other.bitNo.isSameExpression(bitNo)
    override fun stackComplexity():Int = expr.stackComplexity()+bitNo.stackComplexity()+2
}
fun opTestBit32(expr:Expression, bitNo: Expression) = TestBitExpression(ILangDataType.Int32, expr, bitNo)
class InRange32Expression(val expr:Expression, val lower:ULong, val upper:ULong):Expression(ILangDataType.Bool) {
    init {
        if (expr.dataType != ILangDataType.Int32)
            throw Exception("datatype mismatch")
    }
    fun modified(newExpr:Expression):InRange32Expression = if (newExpr === expr) this else InRange32Expression(newExpr, lower, upper)
    override fun toILang():String = "(inRange32 ${expr.toILang()} 0I${lower.toString(16)} 0I${upper.toString(16)})"
    override fun iterate(cb:(e:Expression)->Unit) {
        expr.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(expr.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is InRange32Expression && other.lower == lower && other.upper == upper && other.expr.isSameExpression(expr)
    override fun stackComplexity():Int = expr.stackComplexity()+2
}
class ArrayAccessExpression(val arrayVariable: ILangArrayVariable, val indexExpr:Expression):Expression(arrayVariable.dataType) {
    init {
        if (indexExpr.dataType != ILangDataType.Int32)
            throw Exception("datatype mismatch")
    }
    fun modified(newIndexExpr:Expression):ArrayAccessExpression = if (newIndexExpr === indexExpr) this else ArrayAccessExpression(arrayVariable, newIndexExpr)
    fun modified(newArrayVariable: ILangArrayVariable):ArrayAccessExpression = if (newArrayVariable === arrayVariable) this else ArrayAccessExpression(newArrayVariable, indexExpr)
    override fun toILang():String = "(arrayAccess ${arrayVariable.name} ${indexExpr.toILang()})"
    override fun iterate(cb:(e:Expression)->Unit) {
        indexExpr.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(indexExpr.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is ArrayAccessExpression && other.arrayVariable === arrayVariable && other.indexExpr.isSameExpression(indexExpr)
    override fun stackComplexity():Int = indexExpr.stackComplexity()+2
}
class PartialArrayAccessExpression(val arrayVariable: ILangArrayVariable, val indexExpr:Expression, val isByte:Boolean, val byteOffset:Int):Expression(arrayVariable.dataType) {
    init {
        if (indexExpr.dataType != ILangDataType.Int32)
            throw Exception("datatype mismatch")
    }
    fun modified(newIndexExpr:Expression):PartialArrayAccessExpression = if (newIndexExpr === indexExpr) this else PartialArrayAccessExpression(arrayVariable, newIndexExpr, isByte, byteOffset)
    override fun toILang():String = "(${if (isByte) "byteArrayAccess" else "wordArrayAccess"} ${arrayVariable.name} ${indexExpr.toILang()} 0I${byteOffset.toString(16)})"
    override fun iterate(cb:(e:Expression)->Unit) {
        indexExpr.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(indexExpr.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is PartialArrayAccessExpression && other.arrayVariable === arrayVariable && other.isByte == isByte && other.byteOffset == byteOffset && other.indexExpr.isSameExpression(indexExpr)
    override fun stackComplexity():Int = indexExpr.stackComplexity()+2
}
class DecryptExpression(val expr:Expression):Expression(ILangDataType.Int32) {
    init {
        if (expr.dataType != ILangDataType.Int32)
            throw Exception("datatype mismatch")
    }
    fun modified(newExpr:Expression):DecryptExpression = if (newExpr === expr) this else DecryptExpression(newExpr)
    override fun toILang():String = "(decrypt ${expr.toILang()})"
    override fun iterate(cb:(e:Expression)->Unit) {
        expr.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(expr.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is DecryptExpression && other.expr.isSameExpression(expr)
    override fun stackComplexity():Int = expr.stackComplexity()+10
}
class ParityExpression(val expr:Expression):Expression(ILangDataType.Bool) {
    init {
        if (expr.dataType != ILangDataType.Int32)
            throw Exception("datatype mismatch")
    }
    fun modified(newExpr:Expression):ParityExpression = if (newExpr === expr) this else ParityExpression(newExpr)
    override fun toILang():String = "(getParity ${expr.toILang()})"
    override fun iterate(cb:(e:Expression)->Unit) {
        expr.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(expr.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is ParityExpression && other.expr.isSameExpression(expr)
    override fun stackComplexity():Int = expr.stackComplexity()+10
}
class ExtendToU64ShiftSignToUpperExpression(val expr:Expression):Expression(ILangDataType.Int64) {
    init {
        if (expr.dataType != ILangDataType.Int32)
            throw Exception("datatype mismatch")
    }
    fun modified(newExpr:Expression):ExtendToU64ShiftSignToUpperExpression = if (newExpr === expr) this else ExtendToU64ShiftSignToUpperExpression(newExpr)
    override fun toILang():String = "(extendToU64ShiftSignToUpper ${expr.toILang()})"
    override fun iterate(cb:(e:Expression)->Unit) {
        expr.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(expr.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is ExtendToU64ShiftSignToUpperExpression && other.expr.isSameExpression(expr)
    override fun stackComplexity():Int = expr.stackComplexity()+5
}
class NegateWithFlagsExpression(val expr:Expression):Expression(ILangDataType.Int32) {
    init {
        if (expr.dataType != ILangDataType.Int32)
            throw Exception("datatype mismatch")
    }
    fun modified(newExpr:Expression):NegateWithFlagsExpression = if (newExpr === expr) this else NegateWithFlagsExpression(newExpr)
    override fun toILang():String = "(negWithZFlag ${expr.toILang()})"
    override fun iterate(cb:(e:Expression)->Unit) {
        expr.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(expr.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is ExtendToU64ShiftSignToUpperExpression && other.expr.isSameExpression(expr)
    override fun stackComplexity():Int = expr.stackComplexity()+2
}

class LookupExpression(val index:Expression, val values:List<ULong>):Expression(ILangDataType.Int32) {
    init {
        if (index.dataType != ILangDataType.Int32)
            throw Exception("datatype mismatch")
        if (values.size<2)
            throw Exception("lookup expects at least 2 values")
    }
    fun modified(newIndex:Expression):LookupExpression = if (newIndex === index) this else LookupExpression(newIndex, values)
    override fun toILang():String = "(lookup ${index.toILang()} ${values.map { "0I"+it.toString(16) }.joinToString(" ")})"
    override fun iterate(cb:(e:Expression)->Unit) {
        index.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(index.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is LookupExpression && other.index.isSameExpression(index) && other.values == values
    override fun stackComplexity():Int = index.stackComplexity()+4
}
class JumpDestinationLookupExpression(val index:Expression, val values:List<Entry>):Expression(ILangDataType.Int32) {
    class Entry(val value:ULong, val label:String)
    init {
        if (values.size<2)
            throw Exception("lookup expects at least 2 values")
    }
    fun modified(newIndex:Expression):JumpDestinationLookupExpression = if (newIndex === index) this else JumpDestinationLookupExpression(newIndex, values)
    override fun toILang():String = "(lookupJump ${index.toILang()} ${values.map { "0I${it.value.toString(16)}" }.joinToString(" ")})"
    override fun iterate(cb:(e:Expression)->Unit) {
        index.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(index.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is JumpDestinationLookupExpression && other.index.isSameExpression(index) && other.values == values
    override fun stackComplexity():Int = index.stackComplexity()+4
}
class ExtractFromCombinedOperation(val operationType:OperationType, val extractType:ExtractType, val variable: ILangVariable):Expression(when (extractType) {
    ExtractType.Result -> ILangDataType.Int32
    ExtractType.CFlag,ExtractType.ZFlag -> ILangDataType.Bool
}) {
    enum class OperationType {UnsignedAddOrSub,SignedAddOrSub,BitOp}
    enum class ExtractType {Result,ZFlag,CFlag}

    init {
        if (variable.dataType != when(operationType) {
            OperationType.UnsignedAddOrSub, OperationType.SignedAddOrSub -> ILangDataType.Int64
            OperationType.BitOp -> ILangDataType.Int32
        })
            throw Exception("mismatch datatype")
    }

    fun functionName() = "get"+when(operationType) {
        OperationType.UnsignedAddOrSub -> "Unsigned"
        OperationType.SignedAddOrSub -> "Signed"
        OperationType.BitOp -> "BitOp"
    }+when(extractType) {
        ExtractType.Result -> "Result"
        ExtractType.CFlag -> "Carry"
        ExtractType.ZFlag -> "Zero"
    }
    override fun toILang():String = "(${functionName()} ${variable.name})"
    override fun iterate(cb:(e:Expression)->Unit) {
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(this)
    override fun isSameExpression(other:Expression):Boolean = other is ExtractFromCombinedOperation && operationType == other.operationType && extractType == other.extractType && other.variable === variable
    override fun stackComplexity():Int = 2
}

class AddOrSubExpression(val signed:Boolean, val left:Expression, val right:Expression, val isSub:Expression, val extraCarry:Expression):Expression(ILangDataType.Int64) {
    init {
        if (left.dataType != ILangDataType.Int32 || right.dataType != ILangDataType.Int32 || isSub.dataType != ILangDataType.Bool || extraCarry.dataType != ILangDataType.Bool)
            throw Exception("datatype mismatch")
    }

    fun modified(newLeft:Expression, newRight: Expression, newIsSub:Expression, newExtraCarry:Expression):AddOrSubExpression = if (newLeft === left && newRight === right && newIsSub === isSub && newExtraCarry === extraCarry) this else AddOrSubExpression(signed,newLeft, newRight, newIsSub, newExtraCarry)
    fun functionName() = if (signed) "signedAddOrSub" else "unsignedAddOrSub"
    override fun toILang():String = "(${functionName()} ${left.toILang()} ${right.toILang()} ${isSub.toILang()} ${extraCarry.toILang()})"
    override fun iterate(cb:(e:Expression)->Unit) {
        left.iterate(cb)
        right.iterate(cb)
        isSub.iterate(cb)
        extraCarry.iterate(cb)
        cb(this)
    }
    override fun map(cb:(e:Expression)->Expression):Expression = cb(modified(left.map(cb),right.map(cb),isSub.map(cb),extraCarry.map(cb)))
    override fun isSameExpression(other:Expression):Boolean = other is AddOrSubExpression && signed == other.signed && left.isSameExpression(other.left) && right.isSameExpression(other.right) && isSub.isSameExpression(other.isSub) && extraCarry.isSameExpression(other.extraCarry)
    override fun stackComplexity():Int = left.stackComplexity()+right.stackComplexity()+isSub.stackComplexity()+extraCarry.stackComplexity()+4
}

class GetCogNumberExpression():Expression(ILangDataType.Int32) {
    override fun toILang():String = "getCogNumber"
    override fun stackComplexity():Int = 4
    override fun iterate(cb:(e:Expression)->Unit) = cb(this)
    override fun map(cb:(e:Expression)->Expression):Expression = cb(this)
    override fun isSameExpression(other:Expression):Boolean = other is GetCogNumberExpression
}

class GetPllOfCogExpression(val cogNo:Expression):Expression(ILangDataType.Bool) {
    override fun toILang():String = "(getPllOfCog ${cogNo.toILang()})"
    override fun stackComplexity():Int = 4
    override fun iterate(cb:(e:Expression)->Unit) = cb(this)
    override fun map(cb:(e:Expression)->Expression):Expression = cb(this)
    override fun isSameExpression(other:Expression):Boolean = other is GetPllOfCogExpression && cogNo.isSameExpression(other.cogNo)
    fun modified(newCogNo:Expression):GetPllOfCogExpression = if (newCogNo === cogNo) this else GetPllOfCogExpression(newCogNo)
}

sealed class Statement {
    abstract fun iterate(cb:(stm:Statement)->Unit)
    abstract fun map(cb:(stm:Statement)->Statement):Statement
    abstract fun mapMulti(cb:(stm:Statement)->List<Statement>):List<Statement>
    abstract fun isSameStatement(other:Statement):Boolean
}
class LetStatement(val destVar:ILangVariable, val expr:Expression):Statement() {
    init {
        if (destVar.dataType != expr.dataType)
            throw Exception("cannot assign ${expr.dataType} to ${destVar.dataType} (variable ${destVar.name})")
        if (destVar.direction == ILangVariableDirection.Input)
            throw Exception("cannot assign to input variable (variable ${destVar.name})")
    }
    fun modified(newDestVar: ILangVariable):LetStatement = if (newDestVar === destVar) this else LetStatement(newDestVar, expr)
    fun modified(newExpr:Expression):LetStatement = if (newExpr === expr) this else LetStatement(destVar, newExpr)
    override fun toString(): String = "let ${destVar.name} $expr"
    override fun iterate(cb:(stm:Statement)->Unit) = cb(this)
    override fun map(cb:(stm:Statement)->Statement):Statement = cb(this)
    override fun mapMulti(cb:(stm:Statement)->List<Statement>):List<Statement> = cb(this)
    override fun isSameStatement(other:Statement):Boolean = other is LetStatement && other.destVar === destVar && other.expr.isSameExpression(expr)
}
class LetArrayStatement(val destArrayVar:ILangArrayVariable, val idx:Expression, val expr:Expression, val accessSize:CpuAccessSize, val byteOffset: Int):Statement() {
    init {
        if (destArrayVar.dataType != expr.dataType)
            throw Exception("cannot assign ${expr.dataType} to ${destArrayVar.dataType} (variable ${destArrayVar.name})")
        if (idx.dataType != ILangDataType.Int32)
            throw Exception("invalid index datatype")
        if (!when(accessSize) {
                CpuAccessSize.QWord,CpuAccessSize.None -> throw Exception("illegal accesssize")
                CpuAccessSize.DWord -> listOf(0)
                CpuAccessSize.Word -> listOf(0,2)
                CpuAccessSize.Byte -> listOf(0,1,2,3)
        }.contains(byteOffset))
            throw Exception("illegal let array stm byte offset $byteOffset")
    }
    override fun toString(): String = "letarr ${destArrayVar.name} $idx $expr"
    fun modified(newIdx:Expression, newExpr:Expression):LetArrayStatement = if (newIdx === idx && newExpr === expr) this else LetArrayStatement(destArrayVar, newIdx, newExpr, accessSize, byteOffset)
    fun modified(newDestArrayVar:ILangArrayVariable):LetArrayStatement = if (newDestArrayVar === destArrayVar) this else LetArrayStatement(newDestArrayVar, idx, expr, accessSize, byteOffset)
    override fun iterate(cb:(stm:Statement)->Unit) = cb(this)
    override fun map(cb:(stm:Statement)->Statement):Statement = cb(this)
    override fun mapMulti(cb:(stm:Statement)->List<Statement>):List<Statement> = cb(this)
    override fun isSameStatement(other:Statement):Boolean = other is LetArrayStatement && other.destArrayVar === destArrayVar && other.idx.isSameExpression(idx) && other.expr.isSameExpression(expr) && other.accessSize == accessSize && other.byteOffset == byteOffset
}
class SwitchStatement(val expr:Expression, val cases:List<Case>):Statement() {
    class Case(val values:List<ULong>, val body:List<Statement>) {
        fun modified(newBody:List<Statement>):Case = if (newBody === body) this else Case(values, newBody)
    }
    override fun toString(): String = if (expr.dataType == ILangDataType.Bool) "if $expr" else "switch $expr"
    fun modified(newExpr:Expression, newCases:List<Case>):SwitchStatement {
        if (newExpr !== expr || newCases.size != cases.size)
            return SwitchStatement(newExpr, newCases)
        for (i in newCases.indices)
            if (newCases[i] !== cases[i])
                return SwitchStatement(newExpr, newCases)
        return this
    }
    fun tryMatchingCase(v:ULong):Case? = cases.find { it.values.contains(v) }
    fun matchingCase(v:ULong):Case = cases.find { it.values.contains(v) } ?: throw Exception("no switch case found for value $v")
    override fun iterate(cb:(stm:Statement)->Unit) {
        for (c in cases) {
            iterateStmList(c.body,cb)
        }
        cb(this)
    }
    override fun map(cb:(stm:Statement)->Statement):Statement {
        val newCases = cases.map {
            it.modified(mapStmList(it.body,cb))
        }
        return cb(modified(expr,newCases))
    }
    override fun mapMulti(cb:(stm:Statement)->List<Statement>):List<Statement> {
        val newCases = cases.map {
            it.modified(mapStmMultiList(it.body,cb))
        }
        return cb(modified(expr,newCases))
    }
    override fun isSameStatement(other:Statement):Boolean {
        if (other !is SwitchStatement || !other.expr.isSameExpression(expr))
            return false
        for (c in cases) {
            for (v in c.values) {
                val otherCase = other.tryMatchingCase(v) ?: return false
                if (!isSameStatementList(c.body, otherCase.body))
                    return false
            }
        }
        return true
    }
}

fun iterateStmList(stms:List<Statement>, cb:(stm:Statement)->Unit) {
    for (s in stms)
        s.iterate(cb)
}

fun mapStmList(stms:List<Statement>, cb:(stm:Statement)->Statement):List<Statement> {
    val result = ArrayList<Statement>()
    for (stm in stms)
        result.add(stm.map(cb))
    return useNewStatements(stms,result)
}

fun mapStmMultiList(stms:List<Statement>, cb:(stm:Statement)->List<Statement>):List<Statement> {
    val result = ArrayList<Statement>()
    for (stm in stms)
        result.addAll(stm.mapMulti(cb))
    return useNewStatements(stms,result)
}

fun mapExpressionsRecursive(stms:List<Statement>, cb:(expr:Expression)->Expression):List<Statement> = mapStmList(stms) {
    stm->
    when(stm) {
        is LetStatement -> stm.modified(cb(stm.expr))
        is LetArrayStatement -> stm.modified(cb(stm.idx),cb(stm.expr))
        is SwitchStatement -> stm.modified(cb(stm.expr),stm.cases)
        is MacroStatement -> stm
    }
}

class MacroStatement(val macroName:String, val doFunctionCall:Boolean, val regSet:ILangMultiVariableSet, val body:List<Statement>, val pureLocals:Set<ILangVariable>, val inputs:List<ILangVariable>) :Statement() {
    override fun toString(): String = "Call Macro $macroName"
    override fun isSameStatement(other:Statement):Boolean = other === this
    override fun iterate(cb:(stm:Statement)->Unit) = cb(this)
    override fun map(cb:(stm:Statement)->Statement):Statement = cb(this)
    override fun mapMulti(cb:(stm:Statement)->List<Statement>):List<Statement> = cb(this)
    fun toFunction():Function {
        val myLocals = ILangVariableSet("*")
        for (v in pureLocals)
            myLocals.addVariable(v)
        val newBody = ArrayList<Statement>()
        for (v in inputs) {
            myLocals.addVariable(v)
            val i = ILangVariable(v.name+"_input",v.dataType,ILangVariableDirection.Input)
            myLocals.addVariable(i)
            newBody.add(LetStatement(v,VariableExpression(i)))
        }
        return Function(macroName, ILangMultiVariableSet(listOf(myLocals)+regSet.sets),newBody+body)
    }
    fun modified(newBody:List<Statement>) = MacroStatement(macroName,doFunctionCall,regSet,newBody,pureLocals,inputs)
}

fun isSameStatementList(stm1:List<Statement>, stm2:List<Statement>):Boolean {
    if (stm1 === stm2)
        return true
    if (stm1.size != stm2.size)
        return false
    for (i in stm1.indices)
        if (!stm1[i].isSameStatement(stm2[i]))
            return false
    return true
}

fun useNewStatements(oldStatemens:List<Statement>, newStatements:List<Statement>):List<Statement> {
    if (newStatements.size != oldStatemens.size)
        return newStatements
    for (i in oldStatemens.indices)
        if (newStatements[i] !== oldStatemens[i])
            return newStatements
    //kein statement hat sich geändert
    return oldStatemens
}
