package ilang

import java.lang.Exception

sealed class ILangTokenParameter
class ILangModule(val varSets:HashMap<String,ILangVariableSet>,val functions:HashMap<String, Function>)
class ILangTokenParameterAtom(val atom:String):ILangTokenParameter()
class ILangTokenParameterGroup(val children:List<ILangTokenParameter>):ILangTokenParameter() {
    fun expectAt(idx:Int):ILangTokenParameter {
        if (idx>=children.size)
            throw Exception("insufficient parameters")
        return children[idx]
    }
    fun expectAtomAt(idx:Int):String = (expectAt(idx) as? ILangTokenParameterAtom ?: throw Exception("expected atom")).atom
    fun expectTypeAt(idx:Int):ILangDataType = when(expectAtomAt(idx)) {
        "B" -> ILangDataType.Bool
        "I" -> ILangDataType.Int32
        "L" -> ILangDataType.Int64
        else -> throw Exception("invalid ilang data type")
    }
}
class ILangToken(val command:String, val parameter:ILangTokenParameterGroup, val isBlock:Boolean) {
    val children = ArrayList<ILangToken>()
    fun validate(expectedParameterCount:Int, expectedBlock:Boolean) {
        if (expectedParameterCount != parameter.children.size)
            throw Exception("ilang invalid parameter count for $command, ${parameter.children.size} but expected $expectedParameterCount")
        if (isBlock != expectedBlock)
            throw Exception("ilang block error, isBlock: $isBlock, but expected $expectedBlock")
    }
}

fun tokenizeILang(sourceCode:String):ILangToken {
    val tokenStack = arrayListOf(ILangToken("",ILangTokenParameterGroup(listOf()),true))
    fun expectedIndentation() = (tokenStack.size-1)*4

    fun parseParameters(str:String):ILangTokenParameterGroup {
        var idx = 0
        fun implParse():ILangTokenParameterGroup {
            var result = ArrayList<ILangTokenParameter>()
            while (idx<str.length && str[idx] != ')') {
                if (str[idx] == '(') {
                    idx++
                    result.add(implParse())
                    if (idx>=str.length || str[idx] != ')')
                        throw Exception("expected ')'")
                    idx++
                    continue
                }
                if (str[idx] == ' ') {
                    ++idx
                    continue
                }
                //bezeichner lesen
                var paramName = ""
                while (idx<str.length && str[idx] != ' ' && str[idx] != '(' && str[idx] != ')')
                    paramName += str[idx++]
                result.add(ILangTokenParameterAtom(paramName))
            }
            return ILangTokenParameterGroup(result)
        }
        return implParse()
    }

    for (lne in sourceCode.replace("\r","").split("\n")) {
        if (lne.isEmpty() || lne.startsWith("//"))
            continue
        val lneNoIndentation = lne.dropWhile { it == ' ' }
        val spaces = lne.length-lneNoIndentation.length
        val isStartOfBlock = lneNoIndentation.endsWith(":")
        while (spaces<expectedIndentation() && tokenStack.isNotEmpty())
            tokenStack.removeLast()
        if (spaces != expectedIndentation())
            throw Exception("ilang indentation error, got $spaces spaces")

        val splitted = parseParameters((if (isStartOfBlock) lneNoIndentation.dropLast(1) else lneNoIndentation))
        if (splitted.children.isEmpty())
            throw Exception("ilang no command given")
        val cmd = splitted.children.first() as? ILangTokenParameterAtom ?: throw Exception("expected atom")
        val token = ILangToken(cmd.atom,ILangTokenParameterGroup(splitted.children.subList(1,splitted.children.size)),isStartOfBlock)
        tokenStack.last().children.add(token)
        if (isStartOfBlock)
            tokenStack.add(token)
    }
    return tokenStack.first()
}

fun deserializeILang(sourceCode:String):ILangModule {
    val allRegSets = HashMap<String,ILangVariableSet>()
    val functions = HashMap<String, Function>()
    fun parseRegSet(regsetRoot:ILangToken) {
        regsetRoot.validate(1, true)
        val regSet = ILangVariableSet(regsetRoot.parameter.expectAtomAt(0))
        for (r in regsetRoot.children) {
            if (r.command == "reg") {
                r.validate(2, false)
                val regType = r.parameter.expectTypeAt(0)
                val regName = r.parameter.expectAtomAt(1)
                regSet.addVariable(ILangVariable(regName, regType, ILangVariableDirection.Register))
            }
            else if (r.command == "regarray") {
                r.validate(3, false)
                val regType = r.parameter.expectTypeAt(0)
                val regName = r.parameter.expectAtomAt(1)
                val arraySize = r.parameter.expectAtomAt(2)
                regSet.addArrayVariable(ILangArrayVariable(regName, regType, arraySize.drop(2).toInt(16)))
            }
            else
                throw Exception("invalid ilang item ${r.command}")
        }
        allRegSets[regSet.name] = regSet
    }
    fun parseFunction(functionRoot:ILangToken) {
        functionRoot.validate(1, true)
        val functionName = functionRoot.parameter.expectAtomAt(0)
        var functionBody:ILangToken? = null
        val usedRegSets = ArrayList<ILangVariableSet>()
        val locals = ILangVariableSet("*")
        usedRegSets.add(locals)
        val implicitInitLocals = ArrayList<ILangVariable>()
        for (r in functionRoot.children) {
            if (r.command == "local" || r.command == "input" || r.command == "output") {
                r.validate(2, false)
                val regType = r.parameter.expectTypeAt(0)
                val regName = r.parameter.expectAtomAt(1)
                val loc = ILangVariable(regName, regType, when(r.command) {
                    "local" -> ILangVariableDirection.Local
                    "inout" -> ILangVariableDirection.InOut
                    "input" -> ILangVariableDirection.Input
                    "output" -> ILangVariableDirection.Output
                    else -> throw Exception("invalid variable direction")
                })
                locals.addVariable(loc)
                if (loc.direction == ILangVariableDirection.Local || loc.direction == ILangVariableDirection.Output)
                    implicitInitLocals.add(loc)
            }
            else if (r.command == "useregset") {
                r.validate(1, false)
                val regSetName = r.parameter.expectAtomAt(0)
                usedRegSets.add(allRegSets[regSetName] ?: throw Exception("unknown regset $regSetName"))
            }
            else if (r.command == "body") {
                r.validate(0, true)
                if (functionBody!=null)
                    throw Exception("duplicate function body")
                functionBody = r
            }
            else
                throw Exception("invalid ilang item ${r.command}")
        }
        val multiRegSet = ILangMultiVariableSet(usedRegSets)
        //Lokale Variablen und Outputs werden hier initialisiert
        val initLocalSequence = implicitInitLocals.map { LetStatement(it,ConstantExpression(0UL,it.dataType)) }
        functions[functionName] = Function(functionName,multiRegSet,initLocalSequence + parseFunctionBody(functionBody?:throw Exception("expected function body"),multiRegSet))
    }
    val root = tokenizeILang(sourceCode)
    for (c in root.children) {
        if (c.command == "regset")
            parseRegSet(c)
        else if (c.command == "function")
            parseFunction(c)
        else
            throw Exception("invalid ilang item ${c.command}")
    }
    return ILangModule(allRegSets, functions)
}
