package ilang

import assembler.CpuAccessSize
import java.lang.Exception

fun parseILangExpression(expr: ILangTokenParameter, usedRegSets:ILangMultiVariableSet): Expression = when (expr) {
    is ILangTokenParameterAtom -> {
        if (expr.atom == "true")
            ConstantExpression(1UL, ILangDataType.Bool)
        else if (expr.atom == "false")
            ConstantExpression(0UL, ILangDataType.Bool)
        else if (expr.atom.startsWith("0I"))
            ConstantExpression(expr.atom.drop(2).toULong(16), ILangDataType.Int32)
        else if (expr.atom.startsWith("0L"))
            ConstantExpression(expr.atom.drop(2).toULong(16), ILangDataType.Int64)
        else
            VariableExpression(usedRegSets.getVariableByName(expr.atom))
    }
    is ILangTokenParameterGroup -> {
        val funcName = expr.expectAtomAt(0)
        if (funcName == "arrayAccess")
            ArrayAccessExpression(usedRegSets.getArrayVariableByName(expr.expectAtomAt(1)),parseILangExpression(expr.expectAt(2),usedRegSets))
        else {
            val params = expr.children.subList(1, expr.children.size).map { parseILangExpression(it,usedRegSets) }
            fun forceVariable(p:Expression):ILangVariable = (p as? VariableExpression)?.variable ?: throw Exception("expected variable")
            when (funcName) {
                "eqB" -> EqualityExpression(ILangDataType.Bool, true, params[0], params[1])
                "eq32" -> EqualityExpression(ILangDataType.Int32, true, params[0], params[1])
                "eq64" -> EqualityExpression(ILangDataType.Int64, true, params[0], params[1])
                "neqB" -> EqualityExpression(ILangDataType.Bool, false, params[0], params[1])
                "neq32" -> EqualityExpression(ILangDataType.Int32, false, params[0], params[1])
                "neq64" -> EqualityExpression(ILangDataType.Int64, false, params[0], params[1])

                "le32" -> CompareExpression(CompareExpression.Operation.Le, params[0], params[1])
                "leq32" -> CompareExpression(CompareExpression.Operation.Leq, params[0], params[1])
                "ge32" -> CompareExpression(CompareExpression.Operation.Ge, params[0], params[1])
                "geq32" -> CompareExpression(CompareExpression.Operation.Geq, params[0], params[1])
                "sgnle32" -> CompareExpression(CompareExpression.Operation.SgnLe, params[0], params[1])

                "andB" -> AdditiveOrBitwiseExpression(
                    ILangDataType.Bool,
                    AdditiveOrBitwiseExpression.Operation.And,
                    false,
                    params[0],
                    params[1]
                )
                "orB" -> AdditiveOrBitwiseExpression(
                    ILangDataType.Bool,
                    AdditiveOrBitwiseExpression.Operation.Or,
                    false,
                    params[0],
                    params[1]
                )
                "add32" -> AdditiveOrBitwiseExpression(
                    ILangDataType.Int32,
                    AdditiveOrBitwiseExpression.Operation.Add,
                    false,
                    params[0],
                    params[1]
                )
                "sub32" -> AdditiveOrBitwiseExpression(
                    ILangDataType.Int32,
                    AdditiveOrBitwiseExpression.Operation.Sub,
                    false,
                    params[0],
                    params[1]
                )
                "and32" -> AdditiveOrBitwiseExpression(
                    ILangDataType.Int32,
                    AdditiveOrBitwiseExpression.Operation.And,
                    false,
                    params[0],
                    params[1]
                )
                "andWithZFlag" -> AdditiveOrBitwiseExpression(
                    ILangDataType.Int32,
                    AdditiveOrBitwiseExpression.Operation.And,
                    true,
                    params[0],
                    params[1]
                )
                "or32" -> AdditiveOrBitwiseExpression(
                    ILangDataType.Int32,
                    AdditiveOrBitwiseExpression.Operation.Or,
                    false,
                    params[0],
                    params[1]
                )
                "orWithZFlag" -> AdditiveOrBitwiseExpression(
                    ILangDataType.Int32,
                    AdditiveOrBitwiseExpression.Operation.Or,
                    true,
                    params[0],
                    params[1]
                )
                "xor32" -> AdditiveOrBitwiseExpression(
                    ILangDataType.Int32,
                    AdditiveOrBitwiseExpression.Operation.Xor,
                    false,
                    params[0],
                    params[1]
                )
                "xorWithZFlag" -> AdditiveOrBitwiseExpression(
                    ILangDataType.Int32,
                    AdditiveOrBitwiseExpression.Operation.Xor,
                    true,
                    params[0],
                    params[1]
                )

                "add64" -> AdditiveOrBitwiseExpression(
                    ILangDataType.Int64,
                    AdditiveOrBitwiseExpression.Operation.Add,
                    false,
                    params[0],
                    params[1]
                )
                "sub64" -> AdditiveOrBitwiseExpression(
                    ILangDataType.Int64,
                    AdditiveOrBitwiseExpression.Operation.Sub,
                    false,
                    params[0],
                    params[1]
                )
                "and64" -> AdditiveOrBitwiseExpression(
                    ILangDataType.Int64,
                    AdditiveOrBitwiseExpression.Operation.And,
                    false,
                    params[0],
                    params[1]
                )
                "or64" -> AdditiveOrBitwiseExpression(
                    ILangDataType.Int64,
                    AdditiveOrBitwiseExpression.Operation.Or,
                    false,
                    params[0],
                    params[1]
                )
                "xor64" -> AdditiveOrBitwiseExpression(
                    ILangDataType.Int64,
                    AdditiveOrBitwiseExpression.Operation.Xor,
                    false,
                    params[0],
                    params[1]
                )

                "notB" -> InvertExpression(ILangDataType.Bool, params[0])
                "invert32" -> InvertExpression(ILangDataType.Int32, params[0])

                "shl32" -> ShiftExpression(ILangDataType.Int32, true, false, params[0], params[1])
                "shr32" -> ShiftExpression(ILangDataType.Int32, false, false, params[0], params[1])
                "shlWithZFlag" -> ShiftExpression(ILangDataType.Int32, true, true, params[0], params[1])
                "shrWithZFlag" -> ShiftExpression(ILangDataType.Int32, false, true, params[0], params[1])
                "shl64" -> ShiftExpression(ILangDataType.Int64, true, false, params[0], params[1])
                "shr64" -> ShiftExpression(ILangDataType.Int64, false, false, params[0], params[1])

                "lower32" -> Lower32Expression(params[0])
                "cast64" -> Cast64Expression(params[0])
                "muxB" -> MuxExpression(ILangDataType.Bool, params[0], params[1], params[2])
                "mux32" -> MuxExpression(ILangDataType.Int32, params[0], params[1], params[2])
                "mux64" -> MuxExpression(ILangDataType.Int64, params[0], params[1], params[2])

                "testBit32" -> TestBitExpression(ILangDataType.Int32, params[0], params[1])
                "testBit64" -> TestBitExpression(ILangDataType.Int64, params[0], params[1])
                "inRange32" -> {
                    val lower = params[1] as? ConstantExpression ?: throw Exception("invalid inRange32 range")
                    val upper = params[2] as? ConstantExpression ?: throw Exception("invalid inRange32 range")
                    if (lower.dataType != ILangDataType.Int32 || upper.dataType != ILangDataType.Int32)
                        throw Exception("invalid datatype")
                    InRange32Expression(params[0], lower.value, upper.value)
                }
                "decrypt" -> DecryptExpression(params[0])
                "getParity" -> ParityExpression(params[0])
                "negWithZFlag" -> NegateWithFlagsExpression(params[0])
                "extendToU64ShiftSignToUpper" -> ExtendToU64ShiftSignToUpperExpression(params[0])
                "getBitOpResult" -> ExtractFromCombinedOperation(ExtractFromCombinedOperation.OperationType.BitOp,ExtractFromCombinedOperation.ExtractType.Result,forceVariable(params[0]))
                "getBitOpZero" -> ExtractFromCombinedOperation(ExtractFromCombinedOperation.OperationType.BitOp,ExtractFromCombinedOperation.ExtractType.ZFlag,forceVariable(params[0]))
                "getUnsignedResult" -> ExtractFromCombinedOperation(ExtractFromCombinedOperation.OperationType.UnsignedAddOrSub,ExtractFromCombinedOperation.ExtractType.Result,forceVariable(params[0]))
                "getUnsignedZero" -> ExtractFromCombinedOperation(ExtractFromCombinedOperation.OperationType.UnsignedAddOrSub,ExtractFromCombinedOperation.ExtractType.ZFlag,forceVariable(params[0]))
                "getUnsignedCarry" -> ExtractFromCombinedOperation(ExtractFromCombinedOperation.OperationType.UnsignedAddOrSub,ExtractFromCombinedOperation.ExtractType.CFlag,forceVariable(params[0]))
                "getSignedResult" -> ExtractFromCombinedOperation(ExtractFromCombinedOperation.OperationType.SignedAddOrSub,ExtractFromCombinedOperation.ExtractType.Result,forceVariable(params[0]))
                "getSignedZero" -> ExtractFromCombinedOperation(ExtractFromCombinedOperation.OperationType.SignedAddOrSub,ExtractFromCombinedOperation.ExtractType.ZFlag,forceVariable(params[0]))
                "getSignedCarry" -> ExtractFromCombinedOperation(ExtractFromCombinedOperation.OperationType.SignedAddOrSub,ExtractFromCombinedOperation.ExtractType.CFlag,forceVariable(params[0]))
                "unsignedAddOrSub" -> AddOrSubExpression(false,params[0],params[1],params[2],params[3])
                "signedAddOrSub" -> AddOrSubExpression(true,params[0],params[1],params[2],params[3])
                else -> throw Exception("unknown function $funcName")
            }
        }
    }
}

fun parseILangStatements(stmBlk: ILangToken, usedRegSets:ILangMultiVariableSet):List<Statement> {
    val result = ArrayList<Statement>()
    for (stm in stmBlk.children) {
        result.add(when(stm.command) {
            "let" -> {
                stm.validate(2, false)
                val destVar = usedRegSets.getVariableByName(stm.parameter.expectAtomAt(0))
                val srcExpr = parseILangExpression(stm.parameter.expectAt(1), usedRegSets)
                LetStatement(destVar, srcExpr)
            }
            "letarr" -> {
                stm.validate(3, false)
                val destVar = usedRegSets.getArrayVariableByName(stm.parameter.expectAtomAt(0))
                val idxExpr = parseILangExpression(stm.parameter.expectAt(1), usedRegSets)
                val srcExpr = parseILangExpression(stm.parameter.expectAt(2), usedRegSets)
                LetArrayStatement(destVar, idxExpr, srcExpr, CpuAccessSize.DWord, 0)
            }
            "if" -> {
                stm.validate(1, true)
                val conditionExpr = parseILangExpression(stm.parameter.expectAt(0),usedRegSets)
                var trueStm:List<Statement>? = null
                var falseStm:List<Statement>? = null
                for (c in stm.children) {
                    if (c.isBlock && c.command == "true" && trueStm == null)
                        trueStm = parseILangStatements(c, usedRegSets)
                    else if (c.isBlock && c.command == "false" && falseStm == null)
                        falseStm = parseILangStatements(c, usedRegSets)
                    else
                        throw Exception("invalid if case")
                }
                SwitchStatement(conditionExpr, listOf(
                    SwitchStatement.Case(listOf(1U),trueStm?:listOf()),
                    SwitchStatement.Case(listOf(0U),falseStm?:listOf())
                ))
            }
            "switch" -> {
                stm.validate(1, true)
                val conditionExpr = parseILangExpression(stm.parameter.expectAt(0),usedRegSets)
                val cases = ArrayList<SwitchStatement.Case>()
                for (c in stm.children) {
                    if (!c.isBlock || c.command != "case")
                        throw Exception("expected case")
                    cases.add(SwitchStatement.Case(c.parameter.children.map {
                        if (it is ILangTokenParameterAtom && it.atom.startsWith("0I"))
                            it.atom.drop(2).toULong(16)
                        else
                            throw Exception("expected int32 as case value")
                    }, parseILangStatements(c,usedRegSets)))
                }
                SwitchStatement(conditionExpr, cases)
            }
            else -> throw Exception("invalid ilang statement ${stm.command}")
        })
    }
    return result
}


fun parseFunctionBody(rootBody:ILangToken, usedRegSets:ILangMultiVariableSet):List<Statement> {
    return parseILangStatements(rootBody,usedRegSets)
}
