package ilang

import assembler.CpuAccessSize
import java.lang.StringBuilder

fun serializeILangStatement(stms:List<Statement>, indent:String, result:StringBuilder) {
    for (s in stms) {
        when(s) {
            is LetStatement -> result.append("${indent}let ${s.destVar.name} ${s.expr.toILang()}\n")
            is LetArrayStatement -> {
                val idxE = s.idx.toILang()
                val valueE = s.expr.toILang()
                val name = s.destArrayVar.name
                val byteOffset = "0I${s.byteOffset.toString(16)}"
                result.append(when(s.accessSize) {
                    CpuAccessSize.None,CpuAccessSize.QWord->throw Exception("not supported")
                    CpuAccessSize.DWord-> "${indent}letarr $name $idxE $valueE\n"
                    CpuAccessSize.Byte-> "${indent}letbytearr $name $idxE $valueE $byteOffset\n"
                    CpuAccessSize.Word-> "${indent}letwordarr $name $idxE $valueE $byteOffset\n"
                })
            }
            is SwitchStatement -> {
                val subIndent = "$indent        "
                when(s.expr.dataType) {
                    ILangDataType.Bool -> {
                        result.append("${indent}if ${s.expr.toILang()}:\n")
                        for (c in s.cases) {
                            if (c.values.size == 1 && c.values.first() == 0UL) {
                                if (c.body.isNotEmpty()) {
                                    result.append("$indent    false:\n")
                                    serializeILangStatement(c.body,subIndent,result)
                                }
                            }
                            else if (c.values.size == 1 && c.values.first() == 1UL) {
                                if (c.body.isNotEmpty()) {
                                    result.append("$indent    true:\n")
                                    serializeILangStatement(c.body,subIndent,result)
                                }
                            }
                            else
                                throw Exception("if error")
                        }
                    }
                    ILangDataType.Int32 -> {
                        result.append("${indent}switch ${s.expr.toILang()}:\n")
                        for (c in s.cases) {
                            result.append("$indent    case")
                            for (v in c.values)
                                result.append(" 0I${v.toString(16)}")
                            result.append(":\n")
                            serializeILangStatement(c.body,subIndent,result)
                        }
                    }
                    ILangDataType.Int64 -> throw Exception("switch error")
                }
            }
            is MacroStatement -> {
                if (s.doFunctionCall)
                    result.append("${indent}call (${s.macroName} ${s.inputs.map { it.name }.joinToString(" ")})\n")
                else
                    result.append("${indent}macro ${s.macroName}")
            }
        }
    }
}

fun serializeILangFunction(func:Function, indent:String, result:StringBuilder) {
    result.append("${indent}function ${func.name}:\n")
    for (v in func.registerVarSet().sets)
        result.append("$indent    useregset ${v.name}\n")
    for (v in func.localVarSet().getVariableList()) {
        result.append("$indent    ")
        result.append(when(v.direction) {
            ILangVariableDirection.Input -> "input "
            ILangVariableDirection.Output -> "output "
            ILangVariableDirection.InOut -> "inout "
            ILangVariableDirection.Local -> "local "
            ILangVariableDirection.Register -> "reg "
        })
        result.append(when(v.dataType) {
            ILangDataType.Bool -> "B"
            ILangDataType.Int32 -> "I"
            ILangDataType.Int64 -> "L"
        })
        result.append(" ${v.name}\n")
    }
    result.append("$indent    body:\n")
    serializeILangStatement(func.body,"$indent        ",result)
}

fun serializeILang(func: Function):String {
    val result = StringBuilder()
    serializeILangFunction(func, "", result)
    return result.toString()
}

fun serializeILangFuncs(funcs: List<Function>):String {
    val result = StringBuilder()
    for (f in funcs)
        serializeILangFunction(f, "", result)
    return result.toString()
}

