package ilang

import special.affectedVars
import special.dependendVars
import java.lang.Exception

//Was hier passiert:
//Let Statements, die exakt einmal verwendet (und danach nicht durch seiteneffekt verändert) werden entfernt und im Ziel eingesetzt

class LetForwardAnalyserVarState {
    class State(val letStatements:HashSet<LetStatement>, val dependsOnVariable: HashSet<ILangVariable>, var statementsStillValid:Boolean) {
        constructor(stm:LetStatement):this(hashSetOf(stm),stm.expr.getUsedVariables().toHashSet(),true)
        constructor():this(HashSet(),HashSet(),true)
        fun mergeWith(other:State) {
            letStatements.addAll(other.letStatements)
            dependsOnVariable.addAll(other.dependsOnVariable)
            if (!other.statementsStillValid)
                statementsStillValid = false
        }
    }
    private val map: HashMap<ILangVariable, State>
    val forwardLet:HashSet<LetStatement>
    val noForwardLet:HashSet<LetStatement>

    constructor(parent: LetForwardAnalyserVarState) {
        map = HashMap(parent.map)
        forwardLet = parent.forwardLet //wird nicht kopiert
        noForwardLet = parent.noForwardLet //wird nicht kopiert
    }

    constructor() {
        map = HashMap()
        forwardLet = HashSet()
        noForwardLet = HashSet()
    }

    fun setLetStatement(s:LetStatement) {
        map.remove(s.destVar)
        //alle variablen durchgehen, ob die entsprechenden statements jetzt obsolet geworden sind
        for (v in map.values)
            if (v.statementsStillValid && v.dependsOnVariable.contains(s.destVar))
                v.statementsStillValid = false
        map[s.destVar] = State(s)
    }
    fun setUsedInExpression(v:ILangVariable) {
        val state = map[v] ?: return
        if (state.letStatements.size == 1) {
            //es gibt nur ein einziges statement (bei switch/if kein fwd durchführen)
            val s = state.letStatements.first()
            if (forwardLet.contains(s) || !state.statementsStillValid) {
                forwardLet.remove(s)
                noForwardLet.add(s)
            }
            else if (!noForwardLet.contains(s))
                forwardLet.add(s)
        }
        else {
            //es gibt mehrere statements (aus switch/if), nie forwarden
            for (s in state.letStatements) {
                forwardLet.remove(s)
                noForwardLet.add(s)
            }
        }
    }
    fun setUsedInExpression(e:Expression) {
        for (v in e.getUsedVariables())
            setUsedInExpression(v)
    }
    fun setUsedOutsideFunction(v:ILangVariable) {
        val removeFromFwdLet = ArrayList<LetStatement>()
        for (f in forwardLet)
            if (f.destVar === v)
                removeFromFwdLet.add(f)
        forwardLet.removeAll(removeFromFwdLet)
        noForwardLet.addAll(removeFromFwdLet)

    }
    fun joinFromChildren(children:List<LetForwardAnalyserVarState>) {
        if (children.isEmpty())
            return
        val affectedVarsInChildren = HashSet<ILangVariable>() //welche variablen wurden in den zweigen modifiziert?
        for (c in children)
            affectedVarsInChildren.addAll(c.map.keys)
        for (v in affectedVarsInChildren) {
            val outerState = map[v] ?: State()
            var state = State()
            for (c in children)
                state.mergeWith(c.map[v] ?: outerState)
            map[v] = state
        }
    }
}

class LetForwardExecuteVarState {
    private val map: HashMap<ILangVariable, LetStatement>
    private val dbg:Boolean
    constructor(debug:Boolean) {
        map = HashMap()
        dbg = debug
    }
    constructor(parent:LetForwardExecuteVarState) {
        map = HashMap(parent.map)
        dbg = parent.dbg
    }
    fun setFwdLet(stm:LetStatement) {
        if (dbg)
            println("F> $stm")
        map[stm.destVar] = stm
    }
    fun setNonFwdLet(stm:LetStatement) {
        if (dbg && map.containsKey(stm.destVar))
            println("R> $stm")
        map.remove(stm.destVar)
    }
    fun replaceByFwdLet(origExpr:VariableExpression):LetStatement? {
        return map[origExpr.variable]
    }
    fun joinFromChildren(children:List<LetForwardExecuteVarState>) {
        if (children.isEmpty())
            return
        val affectedVarsInChildren = HashSet<ILangVariable>() //welche variablen wurden in den zweigen modifiziert?
        for (c in children)
            affectedVarsInChildren.addAll(c.map.keys)
        for (v in affectedVarsInChildren) {
            var vRes:LetStatement? = map[v] ?: continue
            for (c in children)
                if (c.map[v] !== vRes)
                    vRes = null
            if (vRes == null) {
                if (dbg)
                    println("X> $v")
                map.remove(v)
            }
        }
    }
}

fun forwardLetStatements(function:Function,neverForwardVars:Set<ILangVariable>,debug: Boolean):Function {
    fun countUsage(stmList:List<Statement>, varState:LetForwardAnalyserVarState) {
        for (stm in stmList) {
            when (stm) {
                is LetStatement -> {
                    varState.setUsedInExpression(stm.expr)
                    varState.setLetStatement(stm)
                }
                is LetArrayStatement -> {
                    varState.setUsedInExpression(stm.idx)
                    varState.setUsedInExpression(stm.expr)
                }
                is SwitchStatement -> {
                    varState.setUsedInExpression(stm.expr)
                    val childStates = ArrayList<LetForwardAnalyserVarState>()
                    for (c in stm.cases) {
                        val childState = LetForwardAnalyserVarState(varState)
                        childStates.add(childState)
                        countUsage(c.body, childState)
                    }
                    varState.joinFromChildren(childStates)
                }
                is MacroStatement -> throw Exception("Macro Statement not allowed here")
            }
        }
    }

    fun markRegistersAndOutputAsUsed(varState:LetForwardAnalyserVarState) {
        //Register
        for (set in function.registerVarSet().sets) {
            for (v in set.getVariableList()) {
                varState.setUsedOutsideFunction(v)
            }
        }
        //Outputs
        for (v in function.localVarSet().getVariableList())
            if (v.direction == ILangVariableDirection.Output || v.direction == ILangVariableDirection.InOut)
                varState.setUsedOutsideFunction(v)
    }

    //1. benutzung der let statements zählen
    val usageVarState = LetForwardAnalyserVarState()
    countUsage(function.body, usageVarState)
    markRegistersAndOutputAsUsed(usageVarState)
    //2. analysieren, welche der potentiellen fwd let stms wirklich durchgeführt werden

    val affectedArrays = affectedVars(function.body).arrs
    val potentialFwdStms = ArrayList(usageVarState.forwardLet)
    while(potentialFwdStms.isNotEmpty() && (neverForwardVars.contains(potentialFwdStms.first().destVar) || dependendVars(potentialFwdStms.first()).arrs.intersect(affectedArrays).isNotEmpty()))
        potentialFwdStms.removeAt(0)
    if (potentialFwdStms.isEmpty()) //nichts zu tun
        return function

    val fwdLetStm = potentialFwdStms.first()
    var fwdDone = false
    if (debug)
        println("Planned FWD Let Stm: $fwdLetStm")
    fun doForwardStep(rootExpr:Expression, varState: LetForwardExecuteVarState):Expression = rootExpr.map {
        if (it !is VariableExpression) it
        else {
            val replByLetStm = varState.replaceByFwdLet(it)
            if (replByLetStm == null)
                it
            else {
                if (fwdDone)
                    throw Exception("let statement $replByLetStm replaced twice")
                fwdDone = true
                replByLetStm.expr
            }
        }
    }
    fun doForwardStep(stmList:List<Statement>, varState: LetForwardExecuteVarState):List<Statement> {
        val result = ArrayList<Statement>()
        for (stm in stmList) {
            when (stm) {
                is LetStatement -> {
                    val newStm = stm.modified(doForwardStep(stm.expr, varState)) //muss hier gemacht werden, damit auch bei verworfenen stms geforwarded wird
                    if (fwdLetStm === stm)
                        varState.setFwdLet(stm)
                    else {
                        result.add(newStm)
                        varState.setNonFwdLet(newStm)
                    }
                }
                is LetArrayStatement -> result.add(stm.modified(doForwardStep(stm.idx, varState), doForwardStep(stm.expr, varState)))
                is SwitchStatement -> {
                    val newExpr = doForwardStep(stm.expr, varState)
                    val newCases = ArrayList<SwitchStatement.Case>()
                    val childStates = ArrayList<LetForwardExecuteVarState>()
                    for (c in stm.cases) {
                        val childVarState = LetForwardExecuteVarState(varState)
                        childStates.add(childVarState)
                        newCases.add(c.modified(doForwardStep(c.body, childVarState)))
                    }
                    varState.joinFromChildren(childStates)
                    result.add(stm.modified(newExpr, newCases))
                }
                is MacroStatement -> throw Exception("Macro Statement not allowed here")
            }
        }
        return useNewStatements(stmList, result)
    }
    val resultFun = function.modified(doForwardStep(function.body,LetForwardExecuteVarState(debug)))
    //überprüfen, dass alle repl durchgeführt wurden
    if (!fwdDone)
        throw Exception("let stm not forwarded: $fwdLetStm")
    return resultFun
}