package ilang

fun replaceVariablesInExpression(rootExpr:Expression, varMap:HashMap<ILangVariable,ILangVariable>):Expression = rootExpr.map {
    if (it is VariableExpression) it.modified(varMap[it.variable] ?: it.variable) else it
}
fun replaceVariablesInStatement(stm:Statement, varMap:HashMap<ILangVariable,ILangVariable>):Statement {
    return when(stm) {
        is LetStatement -> stm.modified(varMap[stm.destVar] ?: stm.destVar).modified(replaceVariablesInExpression(stm.expr, varMap))
        is LetArrayStatement -> stm.modified(replaceVariablesInExpression(stm.idx, varMap), replaceVariablesInExpression(stm.expr, varMap))
        is SwitchStatement -> stm.modified(replaceVariablesInExpression(stm.expr, varMap),stm.cases.map { it.modified(replaceVariablesInStatements(it.body, varMap)) })
        is MacroStatement -> throw Exception("Macro Statement not allowed here")
    }
}
fun replaceVariablesInStatements(rootStm:List<Statement>, varMap:HashMap<ILangVariable,ILangVariable>):List<Statement> {
    val result = ArrayList<Statement>()
    for (stm in rootStm)
        result.add(replaceVariablesInStatement(stm, varMap))
    return useNewStatements(rootStm, result)
}
fun replaceVariables(function:Function, varMap:HashMap<ILangVariable,ILangVariable>):Function = function.modified(replaceVariablesInStatements(function.body, varMap))


fun replaceVariablesByExpressionInExpression(rootExpr:Expression, varMap:HashMap<ILangVariable,Expression>):Expression = rootExpr.map {
    if (it is VariableExpression) varMap[it.variable] ?: it else it
}
fun replaceVariablesByExpressionInStatement(rootStm:List<Statement>, varMap:HashMap<ILangVariable,Expression>):List<Statement> {
    val result = ArrayList<Statement>()
    for (stm in rootStm)
        result.add(when(stm) {
            is LetStatement -> {
                if (varMap.containsKey(stm.destVar))
                    throw Exception("variable ${stm.destVar} cannot be replaced by expression, was assigned by let statement")
                stm.modified(replaceVariablesByExpressionInExpression(stm.expr, varMap))
            }
            is LetArrayStatement -> stm.modified(replaceVariablesByExpressionInExpression(stm.idx, varMap), replaceVariablesByExpressionInExpression(stm.expr, varMap))
            is SwitchStatement -> stm.modified(replaceVariablesByExpressionInExpression(stm.expr, varMap),stm.cases.map { it.modified(replaceVariablesByExpressionInStatement(it.body, varMap)) })
            is MacroStatement -> throw Exception("Macro Statement not allowed here")
        })
    return useNewStatements(rootStm, result)
}
fun replaceVariablesByExpression(function: Function, varMap: HashMap<ILangVariable, Expression>):Function = function.modified(replaceVariablesByExpressionInStatement(function.body, varMap))

fun removeOutputs(function: Function, outputs:List<String>):Function {
    val locals = function.localVarSet()
    val newLocals = ILangVariableSet(locals)
    val varMap = HashMap<ILangVariable,ILangVariable>()
    for (o in outputs) {
        val oldVar = newLocals.getVariable(o) ?: throw Exception("no output $o found")
        val newVar = ILangVariable("${o}_dummy",oldVar.dataType,ILangVariableDirection.Local)
        varMap[oldVar] = newVar
        newLocals.addVariable(newVar)
    }
    return replaceVariables(function, varMap).modified(newLocals)
}

fun replaceVarUntilChange(function: Function, oldRegVar:ILangVariable, newLocalVar:ILangVariable):Function {
    fun replaceInExpr(rootExpr: Expression):Expression = rootExpr.map { if (it is VariableExpression && it.variable === oldRegVar) VariableExpression(newLocalVar) else it }
    fun replaceInStms(stms:List<Statement>):Pair<List<Statement>,Boolean> { //true = modified -> abort
        val resultStm = ArrayList<Statement>(stms)
        var abort = false
        for ((i,stm) in resultStm.withIndex()) {
            resultStm[i] = when(stm) {
                is LetStatement -> {
                    if (stm.destVar === oldRegVar)
                        abort = true
                    stm.modified(replaceInExpr(stm.expr))
                }
                is LetArrayStatement -> stm.modified(replaceInExpr(stm.idx),replaceInExpr(stm.expr))
                is SwitchStatement -> {
                    stm.modified(replaceInExpr(stm.expr),stm.cases.map {
                        val (subStm,subAbort) = replaceInStms(it.body)
                        if (subAbort)
                            abort = true
                        it.modified(subStm)
                    })
                }
                is MacroStatement -> throw Exception("macros not supported")
            }
            if (abort)
                break
        }
        return Pair(useNewStatements(stms,resultStm),abort)
    }
    return function.modified(listOf(LetStatement(newLocalVar,VariableExpression(oldRegVar)))+replaceInStms(function.body).first)
}