package special

import ilang.*
import ilang.Function

enum class CogBusSelVersion {StoppedOtherState,StoppedState5,RunningOtherState,RunningState5AndHubOp;
    fun toStateCode():ULong = when(this) {
        StoppedOtherState->0
        StoppedState5->1
        RunningOtherState->2
        RunningState5AndHubOp->3
    }.toULong()
    fun toLabelName():String = when(this) {
        StoppedOtherState->"cogBusSel_SX"
        StoppedState5->"cogBusSel_SH"
        RunningOtherState->"cogBusSel_RX"
        RunningState5AndHubOp->"cogBusSel_RH"
    }
}

class CompoundStateWithTransitionResult(val zMacros:List<Function>, val state4:List<Function>, val otherState:List<Function>)

fun analyseCompoundStateTransitions(manager:CompoundStateManager, zMacros:List<MacroStatement>, cogRegs:ILangVariableSet):CompoundStateWithTransitionResult {
    data class ReducedCompoundState(val running: CompoundState.Running, val baseState:Int, val state5OpCodeType:CompoundState.State5OpCodeType)
    val allStates = manager.getAllStates()
    val compoundStateVar = cogRegs.getVariable("compound_state")!!
    val compoundStateVarX = ILangVariable("compound_statex",ILangDataType.Int32,ILangVariableDirection.Register)
    val busSelState = ILangVariable("bus_sel_state",ILangDataType.Int32,ILangVariableDirection.Register)
    val zMacrosResult = ArrayList<Function>()
    cogRegs.addVariable(compoundStateVarX)
    cogRegs.addVariable(busSelState)
    val macrosCalledFromRunningState = HashSet<MacroStatement>()
    val macrosCalledFromStoppedState = HashSet<MacroStatement>()
    fun followZMacroOfState4(f:Function, running: CompoundState.Running) {
        iterateStmList(f.body) {
            stm ->
            if (stm is LetStatement && stm.destVar === compoundStateVar)
                throw Exception("illegal compound state assignment in state 4")
            if (stm is MacroStatement) {
                if (!stm.macroName.startsWith("macroCogState4ZF"))
                    throw Exception("illegal macro call in state4: ${stm.macroName}")
                if (!zMacros.contains(stm))
                    throw Exception("illegal macro call in state4: ${stm.macroName} internal error, not found in list")
                when(running) {
                    CompoundState.Running.Stopped -> macrosCalledFromStoppedState.add(stm)
                    CompoundState.Running.Running -> macrosCalledFromRunningState.add(stm)
                }
            }
        }
    }
    fun getReducedStateOfStateId(id:ULong):ReducedCompoundState {
        val st = allStates[id.toInt()].state
        return ReducedCompoundState(st.running,st.baseState,st.state5OpCodeType)
    }
    fun getJumpDestination(id:ULong):String = allStates[id.toInt()].f!!.name
    fun getNextReducedStateFromExpression(expr:Expression):ReducedCompoundState {
        if (expr is ConstantExpression)
            return getReducedStateOfStateId(expr.value)
        if (expr !is LookupExpression)
            throw Exception("illegal expr: $expr")
        val st = getReducedStateOfStateId(expr.values.firstOrNull() ?: throw Exception("lookup is empty"))
        for (i in 1 until expr.values.size) {
            val st2=getReducedStateOfStateId(expr.values[i])
            if (st != st2)
                throw Exception("reduced state must be the same in all lookup values $st vs $st2")
        }
        return st
    }
    fun transformExpressionToJumpExpressions(expr:Expression):Expression {
        if (expr is ConstantExpression)
            return ConstantJumpDestinationExpression(expr.value,getJumpDestination(expr.value))
        if (expr !is LookupExpression)
            throw Exception("illegal expr: $expr")
        return JumpDestinationLookupExpression(expr.index,expr.values.map { JumpDestinationLookupExpression.Entry(it, getJumpDestination(it)) })
    }
    fun handleStateTransition(currentState:ReducedCompoundState, nextState:ReducedCompoundState):CogBusSelVersion? {
        if (currentState.baseState == 4) {
            //hier muss auf state1 gewechselt werden, es kann running/nicht running sein
            if (nextState.baseState != 1)
                throw Exception("illegal state transition ${currentState.baseState} -> $nextState")
            if (currentState.running != nextState.running) {
                if (nextState.running != CompoundState.Running.Running)
                    throw Exception("only transition not running -> running allowed")
                return CogBusSelVersion.RunningOtherState
            }
            return null
        }
        if (currentState.running != nextState.running) //running wechsel nur in state 4 erlaubt
            throw Exception("invalid running state transition: $currentState -> $nextState")
        if (currentState.baseState != 5 && nextState.baseState == 5) {
            if (currentState.baseState != 3)
                throw Exception("transition to state 5 only allowed from state 5")
            if (nextState.running == CompoundState.Running.Stopped) {
                if (nextState.state5OpCodeType != CompoundState.State5OpCodeType.Unknown)
                    throw Exception("illegal to state5 transition $nextState")
                return CogBusSelVersion.StoppedState5
            }
            return when(nextState.state5OpCodeType) {
                CompoundState.State5OpCodeType.Unknown -> throw Exception("illegal to state5 transition: $nextState")
                CompoundState.State5OpCodeType.HubOperation -> CogBusSelVersion.RunningState5AndHubOp
                CompoundState.State5OpCodeType.NoHubOperation -> null
            }
        }
        if (currentState.baseState == 5 && nextState.baseState != 5) { //von state5 -> state4
            if (nextState.baseState != 4)
                throw Exception("invalid running state transition: $currentState -> $nextState")
            return if (nextState.running == CompoundState.Running.Stopped)
                    CogBusSelVersion.StoppedOtherState
                else when(currentState.state5OpCodeType) {
                    CompoundState.State5OpCodeType.Unknown -> throw Exception("illegal to state3 transition: $currentState")
                    CompoundState.State5OpCodeType.HubOperation -> CogBusSelVersion.RunningOtherState
                    CompoundState.State5OpCodeType.NoHubOperation -> null
                }
        }
        return null
    }
    fun handleCompoundStateAssignment(stm:LetStatement, currentState:ReducedCompoundState):List<Statement> {
        fun buildResultStm(newExpr:Expression, transition:CogBusSelVersion?):List<Statement> {
            val newCompoundStateSetStm = stm.modified(compoundStateVarX).modified(newExpr)
            return if (transition == null)
                    listOf(newCompoundStateSetStm)
                else
                    listOf(newCompoundStateSetStm,
                        LetStatement(busSelState,ConstantJumpDestinationExpression(transition.toStateCode(),transition.toLabelName()))
                    )
        }
        if (stm.expr !is MuxExpression) {
            val nextRedState = getNextReducedStateFromExpression(stm.expr)
            val transition = handleStateTransition(currentState,nextRedState)
            return buildResultStm(transformExpressionToJumpExpressions(stm.expr),transition)
        }
        val nextRedStateTrue = getNextReducedStateFromExpression(stm.expr.trueExpr)
        val nextRedStateFalse = getNextReducedStateFromExpression(stm.expr.falseExpr)
        val newExprTrue = transformExpressionToJumpExpressions(stm.expr.trueExpr)
        val newExprFalse = transformExpressionToJumpExpressions(stm.expr.falseExpr)
        if (nextRedStateTrue == nextRedStateFalse) {
            //pseudo mux
            val transition = handleStateTransition(currentState, nextRedStateTrue)
            return buildResultStm(stm.expr.modified(stm.expr.condition,newExprTrue,newExprFalse),transition)
        }
        val transitionTrue = handleStateTransition(currentState, nextRedStateTrue)
        val transitionFalse = handleStateTransition(currentState, nextRedStateFalse)
        if (transitionTrue == transitionFalse)
            return buildResultStm(stm.expr.modified(stm.expr.condition,newExprTrue,newExprFalse),transitionTrue)
        return listOf(SwitchStatement(stm.expr.condition,listOf(
            SwitchStatement.Case(listOf(1UL),buildResultStm(newExprTrue,transitionTrue)),
            SwitchStatement.Case(listOf(0UL),buildResultStm(newExprFalse,transitionFalse))
        )))
    }
    fun mapStmsForCompoundStateAssignment(stms:List<Statement>, currentState:ReducedCompoundState):List<Statement> {
        return mapStmMultiList(stms) {
            if (it is LetStatement && it.destVar === compoundStateVar)
                handleCompoundStateAssignment(it,currentState)
            else
                listOf(it)
        }
    }
    fun analyseZMacro(macro:MacroStatement):Function {
        val running = if (macrosCalledFromRunningState.contains(macro)) {
            if (macrosCalledFromStoppedState.contains(macro))
                throw Exception("macro ${macro.macroName} called from running and stopped")
            else
                CompoundState.Running.Running
        } else if (macrosCalledFromStoppedState.contains(macro))
            CompoundState.Running.Stopped
        else
            throw Exception("macro ${macro.macroName} never called")
        val newStms = mapStmsForCompoundStateAssignment(macro.body, ReducedCompoundState(running, 4,CompoundState.State5OpCodeType.Unknown))
        return macro.modified(newStms).toFunction()
    }
    val state4Funcs = ArrayList<Function>()
    val stateOtherFuncs = ArrayList<Function>()
    for (st in allStates) {
        val f = st.f!!
        if (st.state.baseState == 4) {
            state4Funcs.add(f)
            followZMacroOfState4(f,st.state.running)
        }
        else {
            val newF = f.modified(
                mapStmsForCompoundStateAssignment(
                    f.body,
                    ReducedCompoundState(st.state.running, st.state.baseState, st.state.state5OpCodeType)
                )
            )
            st.f = newF
            stateOtherFuncs.add(newF)
        }
    }
    for (m in zMacros)
        zMacrosResult.add(analyseZMacro(m))
    return CompoundStateWithTransitionResult(zMacrosResult,state4Funcs,stateOtherFuncs)
}