package special

import ilang.*
import ilang.Function

//Der CompoundState ist ein neuer State zusammengefügt aus den Variablen state, running, cancel und match

class CompoundState(val baseState:Int, val running:Running, val cancel:Cancel, val match:Match, val state5OpCodeType:State5OpCodeType) {
    fun isCompatible(other:CompoundState):Boolean = baseState == other.baseState && running == other.running &&
            (cancel == other.cancel || cancel == Cancel.Cancel || other.cancel == Cancel.Cancel) &&
            match == other.match && state5OpCodeType == state5OpCodeType
    enum class Running {Running,Stopped;
        override fun toString(): String = when(this) {
            Running -> "R"
            Stopped -> "S"
        }
        fun toBool() = when(this) {
            Running -> true
            Stopped -> false
        }
    }
    enum class Match {Match,NoMatch,Undefined;
        override fun toString(): String = when(this) {
            Match -> "M"
            NoMatch -> "N"
            Undefined -> "U"
        }
    }
    enum class Cancel {Cancel,NoCancel,Undefined;
        override fun toString(): String = when(this) {
            Cancel -> "C"
            NoCancel -> "X"
            Undefined -> "U"
        }
    }
    enum class State5OpCodeType {Unknown,HubOperation,NoHubOperation;
        override fun toString(): String = when(this) {
            Unknown -> "U"
            HubOperation->"H"
            NoHubOperation->"P"
        }
    }
    override fun toString(): String = "$baseState$running$cancel$match$state5OpCodeType"
}

class CompoundStateManager {
    companion object {
        fun getExtraCodeName(state:CompoundState, extraStateName:String):String = "$state$extraStateName"
        fun getBaseCodeName(codeNumber:Int):String = "cogClk${codeNumber.toString(16)}_"
    }
    class StateCode(var f:Function?, val state:CompoundState, val code:Int) {
        fun setFunction(newF:Function, extraStateName: String) {
            if (f!=null)
                throw Exception("statecode set twice")
            f = newF.rename(getBaseCodeName(code)+getExtraCodeName(state,extraStateName))
        }
    }
    private val stateCodes = ArrayList<StateCode>()
    fun getAllStates():List<StateCode> = stateCodes
    fun reserveStateCode(state:CompoundState):StateCode {
        val newCode = StateCode(null,state,stateCodes.size)
        stateCodes.add(newCode)
        return newCode
    }
    fun addFunction(f:Function, state:CompoundState, extraStateName:String):StateCode {
        //suchen, ob es schon es die funktion schon gibt
        for (s in stateCodes) {
            val sF = s.f ?: continue
            if (!isSameStatementList(sF.body,f.body))
                continue
            if (!s.state.isCompatible(state))
                throw Exception("not matching states ${s.state} vs $state")
            s.f = sF.rename("${sF.name}_${getExtraCodeName(state,extraStateName)}")
            return s
        }
        //nicht gefunden, neuen state hinzufügen
        val codeNumber = stateCodes.size
        val newCode = StateCode(f.rename(getBaseCodeName(codeNumber)+getExtraCodeName(state,extraStateName)),state,codeNumber)
        stateCodes.add(newCode)
        return newCode
    }
}


fun extractCompoundState(f:Function, initialCompoundState:CompoundState, callback:(st:CompoundState)->Expression):Function {
    val compoundVars = listOf("state","running","cancel","match").map { f.getVariableByName(it) }
    fun belongsToCompoundStatement(stm:Statement):Boolean = affectedVars(stm).vars.intersect(compoundVars).isNotEmpty()
    fun affectedByCompoundStatement(stm:Statement):Boolean = dependendVars(stm).vars.intersect(compoundVars).isNotEmpty()

    //prüfen, das nichts von den compound vars abhängt (ist einfacher)
    for (stm in f.body)
        if (affectedByCompoundStatement(stm))
            throw Exception("function ${f.name} is affected by compound state variables (forbidden)")

    //vor und nach den compoundstatements kommen evtl andere stms
    val origStms = ArrayList(f.body)
    val stmBefore = ArrayList<Statement>()
    while (origStms.isNotEmpty() && !belongsToCompoundStatement(origStms.first())) {
        stmBefore.add(origStms.first())
        origStms.removeFirst()
    }
    //jetzt kommen die compound stms
    val compoundStms = ArrayList<Statement>()
    while (origStms.isNotEmpty() && belongsToCompoundStatement(origStms.first())) {
        compoundStms.add(origStms.first())
        origStms.removeFirst()
    }
    //jetzt kommen evtl nicht compound statemens danach
    val stmAfter = ArrayList<Statement>()
    while (origStms.isNotEmpty()) {
        if (belongsToCompoundStatement(origStms.first()))
            throw Exception("compound state statements not contingous in ${f.name}")
        stmAfter.add(origStms.first())
        origStms.removeFirst()
    }

    fun traverseCompoundStms(idx:Int, currentState:CompoundState):Expression {
        if (idx >= compoundStms.size)
            return callback(currentState)
        fun traverseStm(stm:Statement):Expression {
            if (stm is SwitchStatement) {
                if (stm.expr.dataType != ILangDataType.Bool)
                    throw Exception("compound state error, no switch stm supported, only ifs")
                val tc = stm.matchingCase(1UL)
                val fc = stm.matchingCase(0UL)
                fun handleIfBranch(cseBody:List<Statement>):Expression {
                    if (cseBody.isEmpty())
                        return traverseCompoundStms(idx+1, currentState)
                    if (cseBody.size>1)
                        throw Exception("compound state error, multi if stm not supported")
                    return traverseStm(cseBody.first())
                }
                return MuxExpression(ILangDataType.Int32,stm.expr,
                    handleIfBranch(tc.body),
                    handleIfBranch(fc.body)
                )
            }
            if (stm !is LetStatement)
                throw Exception("invalid statement in compound stm: $stm")
            fun handleConstant(c:ConstantExpression):Expression {
                if (c.dataType != stm.destVar.dataType)
                    throw Exception("internal datatype error")
                return traverseCompoundStms(idx+1,when(stm.destVar.name) {
                    "state" -> CompoundState(c.value.toInt(),currentState.running,currentState.cancel,currentState.match,currentState.state5OpCodeType)
                    "running" -> CompoundState(currentState.baseState,if (c.value == 1UL) CompoundState.Running.Running else CompoundState.Running.Stopped,currentState.cancel,currentState.match,currentState.state5OpCodeType)
                    "cancel" -> CompoundState(currentState.baseState,currentState.running,if(c.value == 1UL) CompoundState.Cancel.Cancel else CompoundState.Cancel.NoCancel,currentState.match,currentState.state5OpCodeType)
                    "match" -> CompoundState(currentState.baseState,currentState.running,currentState.cancel,if (c.value == 1UL) CompoundState.Match.Match else CompoundState.Match.NoMatch,currentState.state5OpCodeType)
                    else -> throw Exception("invalid let statement in compound stms: $stm")
                })
            }
            fun traverseExpr(expr:Expression):Expression {
                if (expr is MuxExpression)
                    return MuxExpression(ILangDataType.Int32,expr.condition,
                        traverseExpr(expr.trueExpr),traverseExpr(expr.falseExpr))
                if (expr is ConstantExpression)
                    return handleConstant(expr)
                if (stm.destVar.dataType == ILangDataType.Bool)
                    return MuxExpression(ILangDataType.Int32,expr,handleConstant(ConstantExpression(1UL,ILangDataType.Bool)),handleConstant(ConstantExpression(0UL,ILangDataType.Bool)))
                throw Exception("unsupported expression in compound stm $stm")
            }
            return traverseExpr(stm.expr)
        }
        return traverseStm(compoundStms[idx])
    }
    val newCompoundStm = LetStatement(f.getVariableByName("compound_state"),reduceExpression(traverseCompoundStms(0,initialCompoundState),true))
    return f.modified(stmBefore+listOf(newCompoundStm)+stmAfter)
}